/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/////////////////////////////////////////////////////////////////////////
//
// DiagnosticsImpl.cpp
// -----------------
//
//    This is the implementation file for the DiagnosticsImpl class
//
//=========================================================================

#include "stdio.h"

//
//    Header files
//
#include "Diagnostics_I.h"


/////////////////////////////////////////////////////////////////////////////
// CECAboutDlg dialog


DiagnosticsImpl::DiagnosticsImpl ()
{}


DiagnosticsImpl::~DiagnosticsImpl ()
{}

//
// @mfunc AddClass
//        ========
//
//     This function adds a diagnostic class to the registry.
//
// @parm The name of the executable to add the diagnostic class for
// @parm The name of the node on which the excutable must be running
// @parm The username under which the executable must be running
// @parm The class of diagnostics to be added
// @parm The level below or equal to which all diagnostics of this class are to be logged
//

long    DiagnosticsImpl::AddClass (const char* Executable ,const char* User
                                   ,const char* Node ,const char* Class ,long Level)
{
    HKEY    hRegKey;
    char    szRegKey[256];
    char    szNewKey[256];
    LPDWORD lpdwDisposition = 0;
    char    lpszValueName[32];

    //
    //    Attempt to open the key for the specified executable, node and user
    //
    hRegKey = OpenExecutableKey (Executable ,User ,Node ,szRegKey);

    //
    //    if we failed to open the key then create a new one.
    //
    if (hRegKey == NULL)
    {
        (void)sprintf (szRegKey ,"%s\\%s|%s|%s" ,DIAG_REGKEY ,Executable ,Node ,User ,DIAG_CLASSKEY);
        if (RegCreateKeyEx (HKEY_CURRENT_USER, szRegKey, 0, NULL, REG_OPTION_NON_VOLATILE
                            ,KEY_WRITE, NULL, &hRegKey, lpdwDisposition) != ERROR_SUCCESS)
        {
            return -1;
        }
		RegCloseKey(hRegKey);
    }
	else
		RegCloseKey(hRegKey);

    //
    //    Add a new sub-key to CLASSES for the required diagnostic class.
    //
    (void)sprintf (szNewKey ,"%s\\%s" ,szRegKey ,Class);
    RegCreateKeyEx (HKEY_CURRENT_USER, szNewKey, 0, NULL, REG_OPTION_NON_VOLATILE
                    ,KEY_WRITE, NULL, &hRegKey, lpdwDisposition);

//    RegOpenKeyEx (HKEY_CURRENT_USER, szNewKey, 0, KEY_ALL_ACCESS, &hRegKey);

    //
    //    Set the LEVEL value for the key
    //
    (void) strcpy (lpszValueName ,DIAG_LEVELKEY);
    RegSetValueEx (hRegKey, lpszValueName, NULL, REG_DWORD, (unsigned char *) &Level ,sizeof(Level));

	RegCloseKey(hRegKey);
    //
    //    ...and finally return
    //
    return 0;
}


//
// @mfunc DeleteClass
//        ===========
//
//     This function deletes a diagnostic class from the registry.
//
// @parm The name of the executable to delete the diagnostic class from
// @parm The name of the node on which the excutable must be running
// @parm The username under which the executable must be running
// @parm The class of diagnostics to be added
//
long    DiagnosticsImpl::DeleteClass (const char* Executable ,const char* User
                                      ,const char* Node       ,const char* Class)
{
    HKEY    hRegKey;
    char    szRegKey[256];
    char    szNewKey[256];

    //
    //    Attempt to open the key for the specified executable, node and user
    //
    hRegKey = OpenExecutableKey (Executable ,User ,Node ,szRegKey);

    //
    //    if we failed to open the key then we cannot delete any classes for it as it does not exist
    //
    if (hRegKey == NULL) return 0;

    //
    //    Format the name of the specific class sub-key beneath the CLASSES key and delete it.
    //
    (void)sprintf (szNewKey ,"%s\\%s" ,szRegKey ,Class);
    return RegDeleteKey(hRegKey, szNewKey);
}


//
// @mfunc DeleteAllClasses
//        ================
//
//     This function deletes all diagnostic classes beneath the specified executable from the registry.
//
// @parm The name of the executable to delete the diagnostic classes from
// @parm The name of the node on which the excutable must be running
// @parm The username under which the executable must be running
//
long    DiagnosticsImpl::DeleteAllClasses (const char* Executable ,const char* User ,const char* Node)
{
    HKEY    hRegKey;
    char    szRegKey[256];

    //
    //    Attempt to open the key for the specified executable, node and user
    //
    hRegKey = OpenExecutableKey (Executable ,User ,Node ,szRegKey);

    //
    //    if we failed to open the key then we cannot delete any classes for it as it does not exist
    //
    if (hRegKey == NULL) return 0;

    //
    //    Now iterate through each of the classes beneath this key deleting them.
    //
    TCHAR       szBuffer[256];
    DWORD       dwSize = 256;
    FILETIME    time;
    DWORD       dwIndex = 0;
    LONG        lRes;

    while (RegEnumKeyEx (hRegKey, dwIndex, szBuffer, &dwSize, NULL, NULL, NULL, &time) == ERROR_SUCCESS)
    {
        lRes = DeleteClass (Executable ,User ,Node ,szBuffer);
        if (lRes != ERROR_SUCCESS)
            return lRes;
        dwSize = 256;
    }

    return 0;
}


//
// @mfunc IsTraced
//        ========
//
//     This function determines whether the specified executable on the specified node runing under
//     the specified username is being traced or not.
//
// @parm The name of the executable to check for tracing of
// @parm The name of the node on which the excutable must be running
// @parm The username under which the executable must be running
//
long    DiagnosticsImpl::IsTraced   (const char* Executable ,const char* User ,const char* Node)
{
    HKEY    hRegKey;
    char    szRegKey[256];

    //
    //    Attempt to open the key for the specified executable, node and user
    //
    hRegKey = OpenExecutableKey (Executable ,User ,Node ,szRegKey);
	RegCloseKey(hRegKey);
    return (hRegKey == NULL) ? -1 : 0;
}


//
// @mfunc GetClass
//        ========
//
//     This function returns class details to the user.
//
// @parm The name of the executable to get class details for.
// @parm The name of the node on which the excutable must be running
// @parm The username under which the executable must be running
// @parm Buffer to return the class name into
// @parm Length of the above buffer
// @parm Pointer to long to hold the level specified for the returned class.
// @rdesc -1 if no such key
//          0 if all classes/levels recovered
//          1 if not enough sace in the buffer for all classes and/or levels
//
long    DiagnosticsImpl::GetClasses (const char* Executable ,const char* User ,const char* Node
                                     ,char* &ClassBuff ,unsigned long RetBufLen
                                     ,char*& LevelBuf  ,unsigned long LevelBufLen)
{
    HKEY    hRegKey;
    char    szRegKey[256];

    //
    //    Attempt to open the key for the specified executable, node and user
    //
    hRegKey = OpenExecutableKey (Executable ,User ,Node ,szRegKey);
    if (hRegKey == NULL) return -1;

    //
    //    Enumerate through the subkeys of the CLASSES key (which gives us the classes), adding their names to
    //    the buffer so long as their is enough space and returning the Levels in a secondary buffer. We separate
    //    the individual classes and levels using the pipe ('|') character as a delimiter
    //
    char        cBuffer[10];
    TCHAR       szBuffer[256];
    DWORD       dwSize = 256;
    FILETIME    time;
    DWORD       dwIndex = 0;
    LONG        Level;
    unsigned long LevelLen = sizeof(Level);

    char*       pClassBuff = ClassBuff;
    char*       pLevelBuff = LevelBuf;

    while (RegEnumKeyEx (hRegKey, dwIndex, szBuffer, &dwSize, NULL, NULL, NULL, &time) == ERROR_SUCCESS)
    {

        //
        //    Picked up a new class. Add it to the buffer if sufficient space, if not return a state of 1
        //
        if (RetBufLen < (strlen (szBuffer) + 2)) return 1;

        //
        //    Add a separator if not the first entry
        //
        if (pClassBuff != ClassBuff)
            *pClassBuff++ = '|';

        (void)strcpy(pClassBuff ,szBuffer);
        pClassBuff += strlen(szBuffer);

        //
        //    Now locate the LEVEL velue for this key and add it to the levels buffer
        //
        RegQueryValueEx (hRegKey, DIAG_LEVELKEY, 0, NULL, (unsigned char *) &Level ,&LevelLen);
        itoa(Level, cBuffer, 10);

        //
        //    Picked up a new level . Add it to the buffer if sufficient space, if not return a state of 1
        //
        if (LevelBufLen < (strlen (cBuffer) + 2)) return 1;

        //
        //    Add a separator if not the first entry
        //
        if (pLevelBuff != LevelBuf)
            *pLevelBuff++ = '|';

        (void)strcpy(pLevelBuff ,cBuffer);
        pLevelBuff += strlen(cBuffer);

        //
        //    Update the context and loop around for another shot.
        //
        dwIndex++;
    }

    return 0;
}

//
// @mfunc OpenExecutableKey
//        =================
//
//     This function attempts to open the key for the specified executable running on the specified node
//     under the specified user. Note that either NODE or USER may be specified as ANY. We actually open
//     the CLASSES key beneath the executable as this is where we will work.
//
// @parm The name of the executable to open the key of.
// @parm The name of the node on which the excutable must be running
// @parm The username under which the executable must be running
// @parm Returned name of the key actually opened
// @rdesc Returns the handle to the executable key or -1 if no key found
//
HKEY    DiagnosticsImpl::OpenExecutableKey (const char* Executable ,const char* User ,const char* Node
        ,char* BaseKeyOpened)
{
    HKEY    hRegKey;

    //
    //    We have four chances for openeing this key
    //      1> The key may include the executable, node and user    Executable|Node|User
    //      2> The key may only specifiy the executable and node    Executable|Node|ALL
    //      3> The key may only specify the executable and user     Executable|ALL|User
    //      4> the key may only specify the executable              Executable|ALL|ALL
    //
    //    type 1
    //
    (void)sprintf (BaseKeyOpened ,"%s\\%s|%s|%s\\%s" ,DIAG_REGKEY ,Executable ,Node ,User ,DIAG_CLASSKEY);
    if (RegOpenKeyEx (HKEY_CURRENT_USER, BaseKeyOpened, 0, KEY_READ|KEY_WRITE, &hRegKey) == ERROR_SUCCESS)
        return hRegKey;

    //
    //    Not type 1 key - try type 2.
    //
    (void)sprintf (BaseKeyOpened ,"%s\\%s|%s|ALL\\%s" ,DIAG_REGKEY ,Executable ,Node ,DIAG_CLASSKEY);
    if (RegOpenKeyEx (HKEY_CURRENT_USER, BaseKeyOpened, 0, KEY_READ|KEY_WRITE, &hRegKey) == ERROR_SUCCESS)
        return hRegKey;

    //
    //    Not type 2 key - try type 3.
    //
    (void)sprintf (BaseKeyOpened ,"%s\\%s|ALL|%s\\%s" ,DIAG_REGKEY ,Executable ,User ,DIAG_CLASSKEY);
    if (RegOpenKeyEx (HKEY_CURRENT_USER, BaseKeyOpened, 0, KEY_READ|KEY_WRITE, &hRegKey) == ERROR_SUCCESS)
        return hRegKey;

    //
    //    Not type 3 key - try type 4.
    //
    (void)sprintf (BaseKeyOpened ,"%s\\%s|ALL|ALL\\%s" ,DIAG_REGKEY ,Executable ,User ,DIAG_CLASSKEY);
    if (RegOpenKeyEx (HKEY_CURRENT_USER, BaseKeyOpened, 0, KEY_READ|KEY_WRITE, &hRegKey) == ERROR_SUCCESS)
        return hRegKey;

    return NULL;
}

