/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrRandom
// @module  Random Number Generator | This module contains the implementation
//          for a random number generating class.  This class generates random
//          numbers for various types within specified ranges.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Standard includes
#include <time.h>

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Utils/CyrRandom.h>

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The default constructor seeds the random number generator based on
//          the current system time.
CyrRandom::CyrRandom()
{
    time_t t = time( NULL );
    int x = t % 256;

    // Seed random number generator
    Seed( x, x, x );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor seeds the random number generator with values
//          specified by the calling program.
// @parm    First seed.
// @parm    Second seed.
// @parm    Third seed.
CyrRandom::CyrRandom( int x, int y, int z )
{
    // Seed random number generator
    Seed( x, y, z );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor seeds the random number using the standard short
//          used to seed the srand() function.
// @parm    The seed value.
CyrRandom::CyrRandom( short seed )
{
    int x = seed % 256;

    // Seed random number generator
    Seed( x, x, x );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method seeds the random number generator.
// @syntax  void Seed( int x, int y, int z )
// @syntax  void Seed( short seed )
void CyrRandom::Seed( int x, int y, int z )
{
    if (( x <= 0 ) || ( x > 255 ))
        _x = 1;
    else
        _x = x;
    if (( y <= 0 ) || ( y > 255 ))
        _y = 1;
    else
        _y = y;
    if (( z <= 0 ) || ( z > 255 ))
        _z = 1;
}

void CyrRandom::Seed( short seed )
{
    int x = seed % 256;
    Seed( x, x, x );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method generates a pseudo-random number in the range 0.0 to
//          1.0.
// @rdesc   The random number.
double CyrRandom::Random()
{
    _x = ( 171 * _x ) % 30269;
    _y = ( 172 * _y ) % 30307;
    _z = ( 170 * _z ) % 30323;

    double result = _x / 30269.0;
    result += ( _y / 30307.0 );
    result += ( _y / 30323.0 );

    result -= (int)result;

    return result;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc
// @rdesc
long CyrRandom::RandInt(long min, long max )
{
    double result = Random() * (max + 1 - min);
    return min + (long)result;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc
// @rdesc
short CyrRandom::RandInt(short min, short max )
{
    double result = Random() * (max + 1 - min);
    return min + (short)result;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc
// @rdesc
int CyrRandom::RandInt(int min, int max )
{
    double result = Random() * (max + 1 - min);
    return min + (int)result;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc
// @rdesc
float CyrRandom::Uniform( float min, float max )
{
    float result = min + ( max - min ) * (float)Random();
    return result;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc
// @rdesc
double CyrRandom::Uniform( double min, double max )
{
    double result = min + ( max - min ) * (double)Random();
    return result;
}

