/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __SERVER__
#  include <gserver.hpp>
#endif  // __SERVER__

#include <algorithm>
//==========================================================< Constant data >
const char* CTServer::azOption_[CTServer::OPT_LAST] =
#ifdef  WIN32
    {
        "-console",
        "-service",
        "-install",
        "-uninstall",
        "-stop",
        "-pause",
        "-continue",
        "-port",
        "-help"
    };
#endif  // WIN32
#ifdef  UNIX
{
    "-console",
    "-daemon",
    "-install",
    "-uninstall",
    "-stop",
    "-pause",
    "-continue",
    "-port",
    "-help"
};
#endif  // UNIX


//---------------------------------------------------------------------------
// CTServer::CTServer
//---------------------------------------------------------------------------
// Constructor.
//---------------------------------------------------------------------------
// Arguments : serviceDisplayName : Service name.
//             outputStream : output stream for messages.
//             argc : Number of arguments.
//             argv : Arguments array.
//             pLogFileList : list of log output streams.
//             tPort : server's port #.
//             bBlocking : if CTRUE, uses the blocking model on selects.
//             bVerbose : if CTRUE, displays info messages (if not service/daemon).
//             secondsTimeout : when bBlocking is CFALSE, this is the number
//                              of seconds of timeout for the selects.
//             millisecondsTimeout : when bBlocking is CFALSE, this is the number
//                                   of milliseconds of timeout for the selects.
// Return    : none.
//---------------------------------------------------------------------------
CTServer::CTServer(
    const string& serviceDisplayName,
    ostream& outputStream,
    int argc, char *argv[],
    CTOLogFileList *pLogFileList,
    port_t tPort,
    bool bBlocking,
    bool bVerbose,
    long secondsTimeout,
    long millisecondsTimeout,
    bool bReuse_port)
        : CTService (argv[0], serviceDisplayName),
        serviceDisplayName_ (serviceDisplayName),
        //server_ (CTSocket::STREAM, bReuse_port),
        eCommand_ (SRV_START),
        outputStream_(outputStream),
        pMainThread_(NULL),
        pListenThread_(NULL),
        bVerbose_(bVerbose),
        bBlocking_(bBlocking),
        iClients_(0),
        pLogFileList_(pLogFileList),
        continue_(mutexContinue_)
{
    timeOut_.tv_sec = secondsTimeout;
    timeOut_.tv_usec = millisecondsTimeout;

    unsigned int uiPort;
    parse(uiPort, argc, argv);
    if(uiPort)
        tPort = (port_t)uiPort;

    server_.reuseAddress (bReuse_port);
    server_.bindToPort(tPort);
}

//---------------------------------------------------------------------------
// CTServer::~CTServer
//---------------------------------------------------------------------------
// Destructor.
//---------------------------------------------------------------------------
// Arguments : none.
// Return    : none.
//---------------------------------------------------------------------------
CTServer::~CTServer()
{
    if(pMainThread_ != NULL)
        delete pMainThread_;

    if(pListenThread_ != NULL)
        delete pListenThread_;
}

//---------------------------------------------------------------------------
// CTServer::parse
//---------------------------------------------------------------------------
// Parse the command line arguments.
//---------------------------------------------------------------------------
// Arguments :
//             uiPort (IN/OUT): port number.
//             argc : Number of arguments.
//             argv : Arguments array.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::parse(unsigned int& uiPort, int argc, char* argv[])
{
    int         i;
    string  arg;

    uiPort = 0;
    programName_ = argv[0];
    programName_.erase (0, programName_.find_last_of('\\') + 1);
    programName_.erase (0, programName_.find_last_of('/') + 1);

    for (i = 1; i < argc; i++)
    {
        arg = argv[i];
	    enumOption eOption;
        for(eOption = OPT_CONSOLE; eOption < OPT_LAST; eOption=getNextOption(eOption))
            if(arg == azOption_[eOption])
                break;

        switch(eOption)
        {
        case OPT_CONSOLE:
            eCommand_ = SRV_CONSOLE;
            break;
        case OPT_SERVICE:
            eCommand_ = SRV_SERVICE;
            break;
        case OPT_INSTALL:
            eCommand_ = SRV_INSTALL;
            break;
        case OPT_UNINSTALL:
            eCommand_ = SRV_UNINSTALL;
            break;
        case OPT_STOP:
            eCommand_ = SRV_STOP;
            break;
        case OPT_PAUSE:
            eCommand_ = SRV_PAUSE;
            break;
        case OPT_CONTINUE:
            eCommand_ = SRV_CONTINUE;
            break;
        case OPT_PORT:
            if(++i == argc)
            {
                // Error from command line arguments.
                *pLogFileList_ << error << "parse: option " << arg << " : missing port #." << endl;
                throw CTException (string ("parse: option ") + arg + " : missing port #.", __FILE__, __LINE__);
            }
            sscanf(argv[i], "%u", &uiPort);
            break;
        case OPT_HELP:
            eCommand_ = SRV_HELP;
            break;
        default:
            // Error from command line arguments.
            *pLogFileList_ << error << "parse: option " << arg << " : bad argument." << endl;
            throw CTException (string ("parse: option ") + arg + " : bad argument.", __FILE__, __LINE__);
            break;
        }
    }
}

//---------------------------------------------------------------------------
// CTServer::rollMyChicken
//---------------------------------------------------------------------------
// Execute the service command.
//---------------------------------------------------------------------------
// Arguments : none.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::rollMyChicken (void)
{
    switch (eCommand_)
    {
    case SRV_START :
        *pLogFileList_ << serviceDisplayName_ << " on " << server_ << " is starting..." << endl;
        start () ;
        break ;
    case SRV_CONSOLE :
        *pLogFileList_ << serviceDisplayName_ << " on " << server_ << " is starting..." << endl;
        start (START_CONSOLE) ;
        break ;
    case SRV_SERVICE :
        *pLogFileList_ << serviceDisplayName_ << " on " << server_ << " is starting..." << endl;
        start (START_SERVICE) ;
        break ;
    case SRV_INSTALL :
        install () ;
        break ;
    case SRV_UNINSTALL :
        uninstall () ;
        break ;
    case SRV_STOP :
        control (SRVCTRL_STOP) ;
        break ;
    case SRV_PAUSE :
        control (SRVCTRL_PAUSE) ;
        break ;
    case SRV_CONTINUE :
        control (SRVCTRL_CONTINUE) ;
        break ;
    case SRV_HELP :
        help () ;
        break ;
    default :
        // Error from eCommand_.
        *pLogFileList_ << error << "rollMyChicken: eCommand_ is invalid." << endl;
        throw CTException ("rollMyChicken: eCommand_ is invalid.", __FILE__, __LINE__) ;
    }
}

//---------------------------------------------------------------------------
// CTServer::help
//---------------------------------------------------------------------------
// Displays help on the standard output.
//---------------------------------------------------------------------------
// Arguments : none.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::help (void)
{
    if (eCommand_ == SRV_HELP)
    {
        cout << serviceDisplayName_ << endl ;
        cout << "Usage : " << programName_ << " " << azOption_[OPT_HELP] << endl ;
        cout << "         " << azOption_[OPT_HELP] << ": this help." << endl ;
        cout << "        " << programName_ << " " << azOption_[OPT_INSTALL] << "|" << azOption_[OPT_UNINSTALL] << endl ;
        cout << "         " << azOption_[OPT_INSTALL] << ": install the service." << endl ;
        cout << "         " << azOption_[OPT_UNINSTALL] << ": uninstall the service." << endl ;
        cout << "        " << programName_ << " [" << azOption_[OPT_CONSOLE] << "|" << azOption_[OPT_SERVICE] << "] [" << azOption_[OPT_PORT] << " <port#>]" << endl ;
        cout << "         " << azOption_[OPT_CONSOLE] << ": run as a console." << endl ;
#ifdef  WIN32
        cout << "         " << azOption_[OPT_SERVICE] << ": run as a service." << endl ;
#endif  // WIN32
#ifdef  UNIX
        cout << "         " << azOption_[OPT_SERVICE] << ": run as a daemon." << endl ;
#endif  // UNIX
        cout << "         " << azOption_[OPT_PORT] << ": specify a port number." << endl ;
        cout << "        " << programName_ << " " << azOption_[OPT_STOP] << "|" << azOption_[OPT_PAUSE] << "|" << azOption_[OPT_CONTINUE] << endl ;
        cout << "         " << azOption_[OPT_STOP] << ": stop the service." << endl ;
        cout << "         " << azOption_[OPT_PAUSE] << ": suspend the service." << endl ;
        cout << "         " << azOption_[OPT_CONTINUE] << ": continue the service." << endl ;
    }
    else
    {   // Error from eCommand_.
        *pLogFileList_ << error << "help: eCommand_ is invalid." << endl;
        throw CTException ("help: eCommand_ is invalid.", __FILE__, __LINE__) ;
    }
}

//---------------------------------------------------------------------------
// CTClientTask
//---------------------------------------------------------------------------
// Internal Task function, launching the user task function.
//---------------------------------------------------------------------------
// Arguments : pServer : pointer to the current server.
//             sync : synchronization object.
// Return    : none.
//---------------------------------------------------------------------------
void CTClientTask(CTServer* pServer, CTCondition& sync)
{
    CTSocket workingSocket(pServer->getSocketServer().accept());
    sync.signal();
    if(pServer->getVerbose())
        pServer->getOutputStream() << "Executing task for thread "
        << (long)CTCurrentThread::getThreadId() << endl;

    // Call the user's task function
    pServer->clientTask(workingSocket);

    // Remove the thread from the thread list.
    pServer->aClientIsGone();
}

//---------------------------------------------------------------------------
// CTListenTask
//---------------------------------------------------------------------------
// Internal Task function, managing the listening for new client connexions.
// This thread function spawns a new dedicated thread for each client connexion.
//---------------------------------------------------------------------------
// Arguments : pServer : pointer to the current server.
// Return    : none.
//---------------------------------------------------------------------------
void CTListenTask(CTServer* pServer)
{
    try
    {
        pServer->getSocketServer().listen();
        if (pServer->getVerbose())
            pServer->getOutputStream() << "Listening on " << pServer->getSocketServer() << endl;

        for(;;)
        {
            CTSocket::CTSelection event;

            if(pServer->getBlocking())
                event = pServer->getSocketServer().select(_S_RECEIVE_);
            else
                event = pServer->getSocketServer().select(pServer->getTimeOut(),_S_RECEIVE_);

            if (event == CTSocket::TIMEDOUT)
            {
                continue;
            }
            else if (event == CTSocket::RECEIVE)
            {
                CTCondition sync (pServer->getListenMutex());
                CTLockGuard<CTMutex> guard (pServer->getListenMutex());
                CTThread *pThread = new CTThread(makeFunctorA2(CTClientTask, CTServer*, pServer, CTCondition&, sync), THREAD_SUSPENDED);
                pServer->aClientIsHere(pThread);
                pThread->resume();

                // Wait until a synchronization signal is received from the newly created thread,
                // signaling that the "accept" was issued by that thread.
                sync.wait();
            }
            else
            {
                pServer->getLogFileList() << error << "CTServer::main: unexpected event received" << endl;
                pServer->getOutputStream() << "CTListenTask: unexpected event received" << endl;
            }
        }
    }
    catch (CTSocketException& exception)
    {
        pServer->getLogFileList() << error << "CTListenTask: <" << exception.what() << '>' << "\nListener thread aborting..." << endl;
        pServer->getOutputStream() << "CTListenTask: <" << exception.what() << '>' << "\nListener thread aborting..." << endl;
    }
    /*    catch (RWxmsg& exception)
        {
            pServer->getLogFileList() << error << "CTListenTask: " << exception.why() << "\nListener thread aborting..." << endl;
            pServer->getOutputStream() << "CTListenTask: " << exception.why() << "\nListener thread aborting..." << endl;
        }*/
    catch (const char *msg)
    {
        pServer->getLogFileList() << error << "CTListenTask: " << msg << "\nListener thread aborting..." << endl;
        pServer->getOutputStream() << "CTListenTask: " << msg << "\nListener thread aborting..." << endl;
    }
    catch (...)
    {
        pServer->getLogFileList() << error << "CTListenTask: " << "unknown exception caught; listener thread aborting..." << endl;
        pServer->getOutputStream() << "CTListenTask: " << "unknown exception caught; listener thread aborting..." << endl;
    }
}

//---------------------------------------------------------------------------
// CTServer::init
//---------------------------------------------------------------------------
// Initialization service function.
//---------------------------------------------------------------------------
// Arguments : none.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::init(void)
{
    try
    {
        mutexContinue_.lock();

        // Spawn the listener thread
        pListenThread_ = new CTThread(makeFunctorA1(CTListenTask, CTServer*, this));

        // Build a thread object for the current (main) thread.
        pMainThread_   = new CTThread(CTCurrentThread::getThreadId());
    }
    catch (CTSocketException& exception)
    {
        *pLogFileList_ << error << "CTServer::init <" << exception.what() << '>' << endl;
        outputStream_ << "CTServer::init <" << exception.what() << '>' << endl;
    }
    /*    catch (RWxmsg& exception)
        {
            *pLogFileList_ << error << "CTServer::init: " << exception.why() << endl;
            outputStream_ << "CTServer::init: " << exception.why() << endl;
        }*/
    catch (const char *msg)
    {
        *pLogFileList_ << error << "CTServer::init: " << msg << endl;
        outputStream_ << "CTServer::init: " << msg << endl;
    }
    catch (...)
    {
        *pLogFileList_ << error << "CTServer::init: unknown exception caught" << endl;
        outputStream_ << "CTServer::init: unknown exception caught" << endl;
    }
}

//---------------------------------------------------------------------------
// CTServer::main
//---------------------------------------------------------------------------
// Main service function.
//---------------------------------------------------------------------------
// Arguments : none.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::main(void)
{
    // Wait for termination condition
    continue_.wait();
}

//---------------------------------------------------------------------------
// CTServer::exit
//---------------------------------------------------------------------------
// Exit service function.
//---------------------------------------------------------------------------
// Arguments : none.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::exit(void)
{
    *pLogFileList_ << serviceDisplayName_ << " on " << server_ << " is stopping." << endl;
    outputStream_ << serviceDisplayName_ << " on " << server_ << " is stopping." << endl;
    joinAndTerminateAllThreads();   // Terminate client threads
    pListenThread_->kill();         // Terminate the listener thread.
    CTCurrentThread::join(*pListenThread_); // Wait until listener thread is stopped
    mutexContinue_.unlock();
}

//---------------------------------------------------------------------------
// CTServer::controlNotify
//---------------------------------------------------------------------------
// Notifications from the service manager.
//---------------------------------------------------------------------------
// Arguments : eServiceControl : action asked by the service program manager.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::controlNotify(enumServiceControl eServiceControl)
{
    switch (eServiceControl)
    {
    case SRVCTRL_STOP:
        continue_.signal();
        break;
    case SRVCTRL_PAUSE:
        pListenThread_->suspend();
        pMainThread_->suspend();
        pauseAllThreads();
        break;
    case SRVCTRL_CONTINUE:
        resumeAllThreads();
        pMainThread_->resume();
        pListenThread_->resume();
        break;
    }
}

//---------------------------------------------------------------------------
// CTServer::joinAndTerminateAllThreads
//---------------------------------------------------------------------------
// Terminates all client threads of the service, and waits for their end.
// The main thread (running in the CTServer::main function) is not affected.
// The listener thread is not either.
//---------------------------------------------------------------------------
// Arguments : none.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::joinAndTerminateAllThreads()
{
    CTThreadList::iterator iter;
    for(iter=threadList_.begin();iter!=threadList_.end();iter++)
    {
        (*iter)->kill();
    }
    for(iter=threadList_.begin();iter!=threadList_.end();iter++)
    {
        CTCurrentThread::join(**iter);
    }
    for(iter=threadList_.begin();iter!=threadList_.end();iter++)
    {
        delete *iter;
    }
    threadList_.clear();

}

//---------------------------------------------------------------------------
// CTServer::pauseAllThreads
//---------------------------------------------------------------------------
// Pauses all client threads of the server.
// The main thread (running in the CTServer::main function) is not affected.
//---------------------------------------------------------------------------
// Arguments : none.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::pauseAllThreads()
{
    CTThreadList::iterator iter;

    for(iter=threadList_.begin();iter!=threadList_.end();iter++)
    {
        (*iter)->suspend();
    }
}

//---------------------------------------------------------------------------
// CTServer::resumeAllThreads
//---------------------------------------------------------------------------
// Resumes all client threads of the server.
// The main thread (running in the CTServer::main function) is not affected.
//---------------------------------------------------------------------------
// Arguments : none.
// Return    : none.
//---------------------------------------------------------------------------
void CTServer::resumeAllThreads()
{
    CTThreadList::iterator iter;

    for(iter=threadList_.begin();iter!=threadList_.end();iter++)
    {
        (*iter)->resume();
    }
}

//---------------------------------------------------------------------------
// isSameThread
//---------------------------------------------------------------------------
// Returns 1 if a given thread has an identifier specified on input.
//---------------------------------------------------------------------------
// Arguments :  pThread : a pointer to a thread.
//              data : a thread id.
// Return    :  1 if a given thread has an identifier specified on input.
//---------------------------------------------------------------------------
bool isSameThread(CTThread *pThread, void *data)
{
    return pThread->getThreadId() == data;
}

//---------------------------------------------------------------------------
// CTServer::aClientIsGone()
//---------------------------------------------------------------------------
// Removes the object representing the current thread from the thread list.
// This function must be called from within the thread to be removed.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None.
//---------------------------------------------------------------------------
void CTServer::aClientIsGone()
{
    try
    {
        CTLockGuard<CTMutex> guard(mutexUser_);

        iClients_--;
        if (getVerbose())
            outputStream_ << server_ << ": " << iClients_ << " clients connected." << endl;
        CTThreadList::iterator pThread_iter;
        for(pThread_iter=threadList_.begin();pThread_iter!=threadList_.end();pThread_iter++)
        {
            if((*pThread_iter)->getThreadId()==CTCurrentThread::getThreadId())break;
        }
        if (pThread_iter  != threadList_.end())
        {
            threadList_.erase(pThread_iter);
            delete *pThread_iter;
        }
        else
        {
            *pLogFileList_ << warning << "CTServer::aClientIsGone: strange thread terminated" << endl;
            throw CTException("CTServer::aClientIsGone: strange thread terminated", __FILE__, __LINE__);
        }
    }
    catch (...)
    {
        *pLogFileList_ << error << "CTServer::aClientIsGone: unknown exception caught" << endl;
        outputStream_ << "CTServer::aClientIsGone: unknown exception caught" << endl;
    }
}

//---------------------------------------------------------------------------
// CTServer::aClientIsHere()
//---------------------------------------------------------------------------
// Enlists a thread object into the thread list.
//---------------------------------------------------------------------------
// Arguments :  pThread : a pointer to the thread to enlist.
// Return    :  None.
//---------------------------------------------------------------------------
void CTServer::aClientIsHere(CTThread *pThread)
{
    try
    {
        CTLockGuard<CTMutex> guard(mutexUser_);
        threadList_.push_back(pThread);

        iClients_++;
        if (getVerbose())
            outputStream_ << server_ << ": " << iClients_ << " clients connected." << endl;
    }
    catch (...)
    {
        *pLogFileList_ << error << "CTServer::aClientIsHere: unknown exception caught" << endl;
        outputStream_ << "CTServer::aClientIsHere: unknown exception caught" << endl;
    }
}
