/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __SERVER__
#define __SERVER__

#ifdef  WIN32
#  pragma warning(disable : 4710)
#endif  // WIN32


#ifndef __MTHREAD__
#  include <mthread.hpp>
#endif  // __MTHREAD__

#ifndef __MUTEX__
#  include <mutex.hpp>
#endif  // __MUTEX__

#ifndef __SERVICE__
#  include <service.hpp>
#endif  // __SERVICE__

#ifndef __LOGFILE__
#  include <logfile.hpp>
#endif  // __LOGFILE__

#ifndef __SOCKET__
#  include <socket.hpp>
#endif

#ifdef  WIN32
#  pragma warning(disable : 4512)
#endif  // WIN32



class CTSecureBoolean : public CTMonitor<CTMutex>
{
public:             // Constructor(s) and destructor.
CTSecureBoolean  () : bBoolean_ (true)
    {}

    ~CTSecureBoolean ()
    {}


public:             // Operator(s).
    CTSecureBoolean& operator= (const bool bBoolean)
    {
        LockGuard lock (monitor ());
        bBoolean_ = bBoolean;
        return *this;
    }
    bool             operator() ()
    {
        LockGuard lock (monitor ());
        return bBoolean_;
    }

    //-----------------------------------------------------------------------------
private:
    bool            bBoolean_ ;
};

// Cyrtools Server.
// This class groups the functionnality of a Windows NT service / Unix daemon,
// and a multithreaded server based on sockets.
class DLLSPEC_CYRTOOLS CTServer : public CTService
{
public:
    // Constructor / Destructor
    CTServer(const string& serviceDisplayName,
             ostream& outputStream,
             int argc, char* argv[],
             CTOLogFileList *pLogFileList = NULL,
             port_t tPort = 4490,
             bool bBlocking = true,
             bool bVerbose = false,
             long secondsTimeout = 0,
             long millisecondsTimeout = 0,
             bool bReuse_port=false);
    virtual ~CTServer();


    /////////////////////// Server features ///////////////////////////

public:
    virtual void clientTask(CTSocket& workingSocket) = 0;
    void aClientIsGone();
    void aClientIsHere(CTThread *pThread);
    void pauseAllThreads(void);
    void resumeAllThreads(void);
    void joinAndTerminateAllThreads(void);
    void setTimeOut(const timeval &timeOut);
    timeval& getTimeOut(void)
    {
        return timeOut_;
    }
    void setPollingOnSelect(void);
    void setBlocking(bool bBlocking)
    {
        bBlocking_ = bBlocking;
    }
    int  getBlocking(void) const
    {
        return bBlocking_;
    }
    void setVerbose(bool bVerbose)
    {
        bVerbose_ = bVerbose;
    }
    int  getVerbose(void)
    {
        return bVerbose_ && getStartMode() != START_SERVICE;
    }
    CTMutex& getListenMutex(void)
    {
        return mutexListen_;
    }
    CTOLogFileList& getLogFileList(void)
    {
        return *pLogFileList_;
    }
    CTSocket& getSocketServer(void)
    {
        return server_;
    }
    int getNbClients(void) const
    {
        return iClients_;
    }
    ostream&  getOutputStream(void)
    {
        return outputStream_;
    }



    ///////////////// Service / Daemon features //////////////////////////

public:
    enum enumServiceCommand
    {
        SRV_START = 0,
        SRV_CONSOLE,
        SRV_SERVICE,
        SRV_INSTALL,
        SRV_UNINSTALL,
        SRV_STOP,
        SRV_PAUSE,
        SRV_CONTINUE,
        SRV_HELP
    };
    enum enumOption
    {
        OPT_CONSOLE = 0,
        OPT_SERVICE,
        OPT_INSTALL,
        OPT_UNINSTALL,
        OPT_STOP,
        OPT_PAUSE,
        OPT_CONTINUE,
        OPT_PORT,
        OPT_HELP,
        OPT_LAST
    };

public:             // Control methods.
    void            parse (unsigned int& uiPort, int argc, char* argv[]);
    void            rollMyChicken (void);
    void            help (void);

public:             // User methods.
    void            init (void);
    void            main (void);
    void            exit (void);
    void            controlNotify (enumServiceControl);

protected:
    enumOption getNextOption(enumOption eOption)
    {
        switch(eOption)
        {
        case OPT_CONSOLE:
            eOption = OPT_SERVICE;
            break;
        case OPT_SERVICE:
            eOption = OPT_INSTALL;
            break;
        case OPT_INSTALL:
            eOption = OPT_UNINSTALL;
            break;
        case OPT_UNINSTALL:
            eOption = OPT_STOP;
            break;
        case OPT_STOP:
            eOption = OPT_PAUSE;
            break;
        case OPT_PAUSE:
            eOption = OPT_CONTINUE;
            break;
        case OPT_CONTINUE:
            eOption = OPT_PORT;
            break;
        case OPT_PORT:
            eOption = OPT_HELP;
            break;
        case OPT_HELP:
            eOption = OPT_LAST;
            break;
        default:
            break;
        }

        return eOption;
    }
    //void operator=(const CTServer& server);


    //-- Data Members -------------------------------------------------------------------

    /////////////////////// Service / Daemon data members ///////////////////////////

protected:
    string              programName_ ;
    string              serviceDisplayName_ ;
    enumServiceCommand      eCommand_ ;
    CTMutex                 mutexContinue_;
    CTCondition             continue_ ;
    static const char*      azOption_[OPT_LAST] ;


    /////////////////////// Server data members ///////////////////////////

protected:
    CTSocket              server_;
    int                   iClients_;
    timeval               timeOut_;
    CTMutex               mutexListen_;
    CTMutex               mutexUser_;
    bool                 bBlocking_;
    CTThread             *pMainThread_;
    CTThread             *pListenThread_;
    typedef list<CTThread*> CTThreadList;
    CTThreadList threadList_;
    ostream&              outputStream_;
    bool                 bVerbose_;
    CTOLogFileList       *pLogFileList_;
};

#endif  // __SERVER__
