/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
/*===========================================================================*/
/*									     */
/*			     **** C O M M O N ****   			     */
/*									     */
/*===========================================================================*/
/*									     */
/*			Module Name: PSL_DEVICE_RTNS_UNIX		     */
/*			============					     */
/*									     */
/* Description:								     */
/* ============								     */
/*									     */
/*	This module contains a set of routines for the current device	     */
/*	and are UNIX specific.						     */
/*									     */
/*									     */
/*===========================================================================*/

#include <psl_macros.h>
#ifdef UNIX
/*
 *  Standard headers
 */
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
/*
 *  winsize structure is in sys/ioctl.h in AIX
 */
#ifdef IBM
#include <sys/ioctl.h>
#include <stdlib.h>                                                   
#include <string.h>                                                   
#endif
/*
 *  winsize structure is in sys/ioctl.h in OSF
 */
#ifdef OSF
#include <sys/ioctl.h>
#endif

/*
 *  Function prototypes
 */
#include psl_common_file(psl_rtndef.h)

#ifdef IBM
static long GET_CHANNEL(char *dp);
static Boolean MATCH_CHANNELS(char *d1p, char *d2p);
#endif

/*
 *  Structure and parameter files
 */
#include psl_common_file(shrmsg.h)
#include psl_common_file(diagstkdef.h)
#include psl_common_file(psldiag_prmdef.h)


/*===========================================================================*/
/*									     */
/*	PSL_GET_WINSIZE - Get the current device depth/width		     */
/*	===============							     */
/*									     */
/*	Calling Format:							     */
/*									     */
/*	  iret = PSL_GET_WINSIZE(depthp, widthp, diag_ptr);		     */
/*									     */
/*	Where:								     */
/*									     */
/*	  iret		= PSL return status				     */
/*	  depthp	= Ptr to returned depth				     */
/*	  widthp	= Ptr to returned width				     */
/*	  diag_ptr	= Ptr to diagnostic stack			     */
/*									     */
/*	This function obtains the depth and width of the current terminal    */
/*	device.								     */
/*									     */
/*	NOTE:								     */
/*	     stdin MUST be open to use this function.			     */
/*									     */
/*===========================================================================*/
unsigned long PSL_GET_WINSIZE(unsigned short *depthp,
			      unsigned short *widthp,
			      struct diagstk *diag_ptr)
{
    struct winsize	win;			/* Window structure */
    long		sts;			/* Signed status */

/*
 *  Fill in our window structure.
 */
    sts = ioctl(0, TIOCGWINSZ, &win);
    if (sts == -1)
    {
/*
 *	ioctl(TIOCGWINSZ) fails on pseudo devices so we just return 0,0
 */
	*depthp = 0;
	*widthp = 0;
    }
    else
    {
	*depthp = win.ws_row;
	*widthp = win.ws_col;
    }	
    return PSL_NORMAL;
}


/*===========================================================================*/
/*									     */
/*	PSL_TEST_CTRLTRM - Test device to see if its our controlling terminal*/
/*	================					             */
/*									     */
/*	Calling Format:							     */
/*									     */
/*	  iret = PSL_TEST_CTRLTRM(devicenamep, matchp, diag_ptr);      	     */
/*									     */
/*	Where:								     */
/*									     */
/*	  iret		= PSL return status				     */
/*        devicenamep   = Ptr to null terminated device name                 */
/*        matchp        = Ptr to returned test state                         */
/*	  diag_ptr	= Ptr to diagnostic stack			     */
/*									     */
/*      This function tests the specified device name to see if its the      */
/*      controlling terminal for the current process.                        */
/*									     */
/*	NOTE:								     */
/*	     stdin MUST be open to use this function.			     */
/*									     */
/*===========================================================================*/
unsigned long PSL_TEST_CTRLTRM(char *devicenamep,
                               Boolean *matchp,
                               struct diagstk *diag_ptr)
{
    unsigned long	iret;			/* PSL return status */
    struct stat         stat1;                  /* stat buffer for stdin */
    struct stat         stat2;                  /* stat buffer for device */
    int                 sts;                    /* Signed status */
#ifdef IBM
    char                *indevnamp;             /* stdin device name*//*40X260*/
#endif

/*
 *  stat stdin to obtain the major/minor numbers for stdin.
 */
    sts = fstat(0, &stat1);                     /* Get info on stdin */
    if (sts != 0)
    {
        iret = PSL_DIAG_ERRNO(diag_ptr, "fstat(0,)");
        return iret;
    }

/*
 *  stat the specified device to obtain the major/minor numbers for it.
 */
    sts = stat(devicenamep, &stat2);
    if (sts != 0)
    {
        iret = PSL_DIAG_ERRNO(diag_ptr, "stat()");
        return iret;
    }

/*
 *  If the major/minor numbers for stdin and the specified device match then
 *  the specified device must be the controlling terminal (NOT IBM).
 *  On the IBM we also need to check the channel number.
 */
    if (stat1.st_rdev != stat2.st_rdev)
    {
        *matchp = PFALSE;
    }
    else
    {
#ifndef IBM
        *matchp = PTRUE;
#else /* IBM */

/*
 *      Get the stdin device name.
 */
        indevnamp = ttyname(0);
        if (indevnamp == NULL)
        {
            iret = PSL_DIAG_ERRNO(diag_ptr, "fstat(0,)");
            return iret;
        }

/*
 *      Match the two device channel numbers.
 */
        *matchp = MATCH_CHANNELS(indevnamp, devicenamep);
#endif    
    }
    return PSL_NORMAL;
}

#ifdef IBM
/*===========================================================================*/
/*                                                                           */
/*      MATCH_CHANNELS - see if two device channel # are the same            */
/*                                                                           */
/*      match = MATCH_CHANNELS(d1p, d2p);                                    */
/*                                                                           */
/*      where:                                                               */
/*                                                                           */
/*        match         = Boolean status of match                            */
/*        d1p           = Ptr to first device name                           */
/*        d2p           = Ptr to second device name                          */
/*                                                                           */
/*===========================================================================*/
static Boolean MATCH_CHANNELS(char *d1p,
                              char *d2p)
{
    long                 chan1;                 /* device 1 channel # */
    long                 chan2;                 /* device 2 channel # */
    
    chan1 = GET_CHANNEL(d1p);
    chan2 = GET_CHANNEL(d2p);
    if (chan1 == chan2)
    {
        return PTRUE;
    }
    return PFALSE;
}


/*===========================================================================*/
/*                                                                           */
/*      GET_CHANNEL - get the specified devices channel #                    */
/*                                                                           */
/*      channel = GET_CHANNEL(dp);                                           */
/*                                                                           */
/*      where:                                                               */
/*                                                                           */
/*        channel       = the returned channel number                        */
/*        dp            = Ptr to device name                                 */
/*                                                                           */
/*===========================================================================*/
static long GET_CHANNEL(char *dp)
{
    char                *cp;
    long                chan;
    
    cp = strrchr(dp, '/');                      /* Find last '/' */
    if (cp == NULL) return 0;                   /* not found return channel 0 */

    return atol(cp + 1);
}
#endif /*IBM*/
#endif /*UNIX*/

