/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
/*==========================================================================*/
/*									    */
/*		           **** C O M M O N ****			    */
/*									    */
/*==========================================================================*/
/*									    */
/*			Module Name: PSL_MSGF_RTNS_UNIX.C		    */
/*			============					    */
/*									    */
/* Description:								    */
/* ============								    */
/*									    */
/*	This module contains the functions required to use shared	    */
/*	message queues in UNIX.						    */
/*									    */
/*==========================================================================*/

/*
 *	The functions in this module should be all that is required to
 *	manipulate shared message queues. The functions are:
 *
 *	PSL_MSGF_INIT	- Initialize a message queue
 *	PSL_MSGF_SEND	- Send a message to a queue
 *	PSLMSQ_RECEIVE	- Receive a message from a queue
 *	PSLMSQ_REMOVE	- Remove a message queue
 *
 *	N.B. This module is paired with PSL_MSGF_RTNS_VMS.C which performs
 *	     similar functions in VMS.
 */

/*
 *  PSL macros.
 */
#include <psl_macros.h>

#ifdef UNIX				/* Only build in UNIX */

/*
 *  Standard header files
 */
#include <stdio.h>			/* Standard Input/Output */
#include <stdlib.h>			/* Standard library */

/*
 *  More header files. After the ifdef as some are not defined in VMS.
 */
#include <sys/types.h>			/* System types */
#include <sys/stat.h>			/* File structures */
#include <fcntl.h>
#include <errno.h>			/* Errors */
#include <sys/ipc.h>			/* Inter Process Communications */
#include <sys/msg.h>			/* Message facilities */
#include <unistd.h>	    		/* UNIX standard header */
#include <signal.h>			/* Signals */
#include <string.h>			/* String routines */
#include <ctype.h>			/* Functions for testing characters */
#include <pwd.h>			/* Password file structures */

/*
 *  Function prototypes
 */
#include <psl_rtndef.h>			/* PSL routines */
#include <psl_msgf_rtndef.h>		/* PSL MSGF routines */

static unsigned long GET_QUEUE(char *idstr_ptr,
			       long *queue_id_ptr,
			       long *id_ptr,
			       struct diagstk *diag_ptr);

static void GET_KEY_FILE(char *idstr_ptr,
			 char *keyfile_ptr,
			 unsigned long keyfile_size);

static unsigned long AWAIT_EMPTY_QUEUE(long qid,
				       struct diagstk *diag_ptr);

/*
 *  Parameter files
 */
#include <shrmsg.h>			/* Messages */
#include <psldiag_prmdef.h>		/* Diagnostic stack parameters */
#include <psl_msgf_prmdef.h>		/* Message facility access flags */
#include <psl_trc_prmdef.h>		/* PSL tracing symbols */
#include <pslerr_prmdef.h>		/* PSL error routine parameters */
#include <psl_ipcdef.h>			/* PSL IPC definitions */

/*
 *    A single piece of static data is required to hold a pointer to the
 *    buffer allocated in the PSL_MSGF_INIT routine to hold any messages
 *    read
 */
static struct msgbuf *msgf_buf_ptr;	/* Pointer to message buffer	*/


/************************************************************************/
/*									*/
/*	PSL_MSGF_INIT - Initialize a shared message queue		*/
/*									*/
/*	This routine is just a shell around msgget			*/
/*									*/
/*	iret = PSL_MSGF_INIT(msgf_idstr_ptr, msgf_maxmsg, msgf_maxbuf	*/
/*			     msgf_access, msgf_flags,			*/
/*			     queue_id_ptr, diag_ptr);			*/
/*									*/
/*	iret      	=  Return status				*/
/*	msgf_idstr_ptr	=  String used as an id for the message facility*/
/*	msgf_maxmsg	=  Maximum size of a single message		*/
/*	msgf_maxbuf	=  Maximum # bytes on a queue			*/
/*	msgf_access	=  Message facility protection mask   		*/
/*			   this is a mask of PSLPERM_xxx_yyy flags	*/
/*	msgf_flags	=  Message facility initialization flags	*/
/*	queue_id_ptr	=  Returned queue id				*/
/*	diag_ptr	=  Ptr to the diagnostics stack			*/
/*									*/
/*	This function is responsible for ensuring that a particular	*/
/*	message facility exists. A message queue is a system facility,	*/
/*	which, once created is available to all processes on the same	*/
/*	system. For this reason all message facilities in UNIX must be	*/
/*	identified by an unique id so that different processes do not	*/
/*	interfere with each others facilities. The SVR4 standard for	*/
/*	obtaining this unique id is by using the id of a newly created	*/
/*	file on the current system (as each file on a particular system */
/*	has a unique id).						*/
/*	IT IS ESSENTIAL THAT THIS METHOD IS ADHERED TO.			*/
/*									*/
/*	Note that having obtained a unique id via "msgf_idstr_ptr" an	*/
/*	existing message queue with this id will be used. This allows	*/
/*	the first process to create the facility and others to use it.	*/
/*									*/
/*	N.B. On the UNIX platform the parameter msgf_maxmsg will be	*/
/*	ignored. In UNIX the maximum size of a single message cannot be */
/*	changed from C - however the file /etc/system may be used to	*/
/*	set the default value.						*/
/*									*/
/************************************************************************/

unsigned long PSL_MSGF_INIT(char *msgf_idstr_ptr,
			    unsigned long msgf_maxmsg,
			    unsigned long msgf_maxbuf,
			    unsigned long msgf_access,
			    unsigned long msgf_flags,
			    long *queue_id_ptr,
			    struct diagstk *diag_ptr)
{

unsigned long	    iret;		/* PSL Return status */
	 long	    status;		/* Return status from System V fns */
	 Boolean    param_error;	/* Parameter problem? */
unsigned long	    diag_save;		/* Saved diagnostic context */
	 int	    msg_flags;		/* Permission & creation flags for */
					/* Message queue */
struct	 msqid_ds   msq_struct;		/* UNIX message queue structure */
	 long	    key;		/* Key to message queue */
unsigned long	    permissions;	/* Permissions for queue */
	 char	    key_file[256];	/* Name of the key file */
	 size_t	    msgbuf_len;		/* Length of message buffer */
struct	 stat	    stat_buf;		/* stat() returned buffer */
	 int	    keyfd;		/* Key file descriptor */    

/*
 *  Output diagnostic trace information
 */

#ifdef PSL_TRACING
    trace_flag = PSL_TEST_TRACE(PTRC_GEN_MISC);
			    long *id_ptr);
    if (trace_flag)
    {
	PSL_OUTPUT_TRACE(
	    "PSL_MSGF_INIT idstr_ptr=%p, maxmsg=%u, maxbuf=%u, access=%X",
		msgf_idstr_ptr, msgf_maxmsg, msgf_maxbuf, msgf_access);
	PSL_OUTPUT_TRACE(" flags=%X, queue_id_ptr=%p, diag_ptr=%p\n",
		msgf_flags, queue_id_ptr, diag_ptr);
    }

#endif	/* PSL_TRACING */

/*
 *  Push message onto diagnostics stack indicating the current function
 */
    diag_save = PSL_ADD_STATDIAG(diag_ptr, "During PSL_MSGF_INIT");

/*
 *  Assume no parameter problems and successful routine.
 */
    param_error = PFALSE;
    iret = PSL_NORMAL;

/*
 *  Check for a valid id string
 */
    if (msgf_idstr_ptr == NULL)
    {
	param_error = PTRUE;
	(void) PSL_ADD_STATDIAG(diag_ptr, "Invalid NULL value for id string");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
    }

/*
 *  Check for invalid permissions in msq_access
 */
    permissions = msgf_access;
    permissions &= ~(PSLPERM_OWNER_READ |
		     PSLPERM_OWNER_WRITE |
		     PSLPERM_OWNER_PHYSICAL |
		     PSLPERM_OWNER_LOGICAL |
		     PSLPERM_GROUP_READ |
		     PSLPERM_GROUP_WRITE |
		     PSLPERM_GROUP_PHYSICAL |
		     PSLPERM_GROUP_LOGICAL |
		     PSLPERM_WORLD_READ |
		     PSLPERM_WORLD_WRITE |
		     PSLPERM_WORLD_PHYSICAL |
		     PSLPERM_WORLD_LOGICAL |
		     PSLPERM_SYSTEM_READ |
		     PSLPERM_SYSTEM_WRITE |
		     PSLPERM_SYSTEM_PHYSICAL |
		     PSLPERM_SYSTEM_LOGICAL);

    if (permissions != 0)
    {
	param_error = PTRUE;
	(void) PSL_ADD_DIAGNOSTIC(diag_ptr, "%X is invalid protection mask",
			   msgf_access);
    }

/*
 *  If there has been a parameter problem, report it and exit.
 */
    if (param_error)
    {
	iret = PSL_BADPARAM;
	PSL_PUTDIAG(diag_ptr,
		    2,
		    PSL_SYSERVERR,
		    "PSL_MSGF_INIT",
		    iret);
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	goto end;
    }

/*
 *  Now set up the permission flag
 */
    permissions = 0;
    if ((msgf_access & PSLPERM_OWNER_WRITE) != 0)
	permissions |= 0200;
    if ((msgf_access & PSLPERM_OWNER_READ) != 0)
	permissions |= 0400;
    if ((msgf_access & PSLPERM_GROUP_WRITE) != 0)
	permissions |= 020;
    if ((msgf_access & PSLPERM_GROUP_READ) != 0)
	permissions |= 040;
    if ((msgf_access & PSLPERM_WORLD_WRITE) != 0)
	permissions |= 02;
    if ((msgf_access & PSLPERM_WORLD_READ) != 0)
	permissions |= 04;

/*
 *  If the user specified PSLMSGF_CREIDFIL then we should create the message
 *  facility key file. This is used to obtain a unique id for the message
 *  facility.
 */
    if ((msgf_flags & PSLMSGF_CREIDFIL) != 0)
    {
	GET_KEY_FILE(msgf_idstr_ptr, key_file, sizeof (key_file));
   	status = stat (key_file, &stat_buf);
	if (status != 0)		/* If file does not exist */
	{
	    PSL_CLEARERR();
	    keyfd = creat(key_file, 0666);			   
	    if (keyfd == -1)					     
	    {
		iret = PSL_OPENOUT;
		PSL_LOG_CIOERR(diag_ptr, iret , key_file);
		PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
		goto end;
	    }
/*
 *  Change permission to ensure world read/write access
 */
	    status = fchmod(keyfd, 0666);			 
	    if (status == -1)			
	    {
		iret = PSL_OPENOUT;
		PSL_LOG_CIOERR(diag_ptr, iret , key_file);
		PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
		goto end;
	    }
	}
    }

/*
 *  Now get a unique key based upon the msgf_idstr_ptr string. This must be the
 *  filename (no extension) of a file which already exists in the directory.
 */
    iret = GET_QUEUE(msgf_idstr_ptr,
		     &key,		/* Returned key */
		     NULL,		/* Do not require message queue id */
		     diag_ptr);
    if (!(psl_test_status(iret)))
    {
	goto end;
    }

/*
 *  Try to find an existing queue with this key.
 *  If the queue is not found then we create a new one.
 */
    errno = 0;				/* Reset the error number */
    *queue_id_ptr = msgget(key, 0);	/* Try to get queue id */

/*
 *  If the flags argument specified that the message queue was not to be
 *  created then just return the queue id.
 */
    if (msgf_flags & PSLMSGF_NOCREATE)
    {
/*
 *	If a queue does not exist then return an error.
 */
	if (*queue_id_ptr == -1)
	{
            iret = PSL_FACNOEXISTS;
	    PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	    goto end;
	}
    }
/*
 *  We now have a queue id or -1 for an error calling msgget()
 */
    if (*queue_id_ptr == -1)		/* Queue does not exist */
    {

/*
 *	If no message queue with this key exists then we must create one.
 */
	if (errno == ENOENT)		/* Queue does not exist */
	{
	    msg_flags = IPC_CREAT |	/* Create flag */
			IPC_EXCL |	/* Error if exists flag */
			permissions;	/* Queue permissions */

	    errno = 0;			/* Reset the error number */
	    *queue_id_ptr = msgget(key, msg_flags);

	    if (*queue_id_ptr == -1)	/* Failed to create queue */
	    {
		iret = PSL_DIAG_ERRNO(diag_ptr, "msgget()");
		PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
		goto end;
	    }				/* End of failed to create new queue */
	}				/* End of queue did not exist */
	else				/* Unexpected error */
	{
	    iret = PSL_DIAG_ERRNO(diag_ptr, "msgget()");
	    PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	    goto end;
	}
    }
    else				/* Queue already exists */
    {
/*
 *	A queue already exists that we have access to.
 */
	/*EMPTY*/;			/* Nothing to do */
    }

/*
 *  If msgf_maxbuf is non zero then the maximum number of bytes on the queue
 *  requires changing.
 */
    if (msgf_maxbuf != 0)		/* Change queue size? */
    {
/*
 *	Get the state of the queue - this is returned in a structure.
 */
	errno = 0;			/* Reset the error number */
	status = msgctl(*queue_id_ptr, IPC_STAT, &msq_struct);
	if (status == -1)		/* Failed to get queue status */
	{
	    iret = PSL_DIAG_ERRNO(diag_ptr, "msgctl()");
	    PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	    goto end;
	}				/* End of error getting queue state */

/*
 *	Change the maximum bytes on the queue. This involves modifying
 *	the queue parameter msg_qbytes and writing the structure back to
 *	the queue.
 */
	msq_struct.msg_qbytes = msgf_maxbuf;
	errno = 0;			/* Reset the error number */
	status = msgctl(*queue_id_ptr, IPC_SET, &msq_struct);
	if (status == -1)		/* Failed to set queue parameters */
	{
	    iret = PSL_DIAG_ERRNO(diag_ptr, "msgctl()");
	    PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	    goto end;
	}				/* End of error changing queue size */
    }					/* End of change queue size */

/*
 *    We shall allocate a buffer to be used when reading/writing to/from the
 *    message facility. We use the maximum message size parameter as a base but
 *    must allocate sufficient space for a header record.
 */
    msgbuf_len = sizeof(struct msgbuf) + msgf_maxmsg;
    msgf_buf_ptr = (struct msgbuf *) malloc (msgbuf_len);
    if (msgf_buf_ptr == NULL)		/* Failed to allocate memory */
    {
	iret = PSL_INSVIRMEM;
	(void) PSL_ADD_DIAGNOSTIC(diag_ptr
			          ,"Failed to allocate message buffer");
	PSL_PUTDIAG(diag_ptr, 2, PSL_SYSERVERR, "PSL_MSGF_INIT", iret);
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	goto end;
    }

/*
 *  Successful completion so return diagnostics stack to its pre-function state
 */
    PSL_REMOVE_DIAGNOSTICS(diag_ptr, diag_save);

/*
 *  This label is branched to whenever there is an error or is dropped into
 *  from above if this function executes successfully.
 */

end:
    return iret;
}



/************************************************************************/
/*									*/
/*	PSLOBS_MSGF_SEND - OBSOLETE Send message to a message queue     */
/*									*/
/*	iret = PSLOBS_MSGF_SEND(qid, dat_ptr, dat_len,			*/
/*			     msgf_flags, diag_ptr);			*/
/*									*/
/*	iret      	=  Return status				*/
/*	qid		=  Message queue id				*/
/*	dat_ptr		=  Ptr to data to be sent			*/
/*	dat_len		=  Length of data to be sent			*/
/*	msgf_flags	=  Flags					*/
/*			   PSLMSGF_WAITIOREQ - wait for sent msg to be	*/
/*					       unqueued			*/
/*			   PSLMSGF_WAITSPC - wait until msg may be sent */
/*					     e.g. if queue full		*/
/*	diag_ptr	=  Pointer to the diagnostics stack		*/
/*									*/
/*    This function is an obsolete version of PSL_MSGF_SEND as it 	*/
/*    does not allow a timeout to be specified.                     	*/
/*									*/
/*    This routine should NOT be called, use PSL_MSGF_SEND instead	*/
/*									*/
/************************************************************************/

unsigned long PSLOBS_MSGF_SEND(long qid,
			    void *dat_ptr,
			    unsigned long dat_len,
			    unsigned long msgf_flags,
			    struct diagstk *diag_ptr)
{
    return PSL_MSGF_SEND (qid ,dat_ptr ,dat_len ,msgf_flags ,0 ,diag_ptr);
}

/************************************************************************/
/*									*/
/*	PSL_MSGF_SEND - Send message to a message queue			*/
/*									*/
/*	iret = PSL_MSGF_SEND(qid, dat_ptr, dat_len,			*/
/*			     msgf_flags, timeout ,diag_ptr);		*/
/*									*/
/*	iret      	=  Return status				*/
/*	qid		=  Message queue id				*/
/*	dat_ptr		=  Ptr to data to be sent			*/
/*	dat_len		=  Length of data to be sent			*/
/*	msgf_flags	=  Flags					*/
/*			   PSLMSGF_WAITIOREQ - wait for sent msg to be	*/
/*					       unqueued			*/
/*			   PSLMSGF_WAITSPC - wait until msg may be sent */
/*					     e.g. if queue full		*/
/*	timeout         = Period to wait for message to be read		*/
/*	diag_ptr	=  Pointer to the diagnostics stack		*/
/*									*/
/*	This function is responsible for sending a message to the	*/
/*	message queue identified by "msgf_idstr_ptr". Note that in an	*/
/*	attempt to mimic the VMS IOS$_NOW flag for mailbox I/O the flag */
/*	PSLMSGF_WAITIOREQ waits after a message has been queued for the	*/
/*	queue to completely clear. This is not exactly what VMS does as	*/
/*	when IOS$_NOW is omitted (it waits for a read request on the	*/
/*	mailbox).							*/
/*									*/
/************************************************************************/
unsigned long PSL_MSGF_SEND(long qid,
			    void *dat_ptr,
			    unsigned long dat_len,
			    unsigned long msgf_flags,
			             long timeout,
			    struct diagstk *diag_ptr)
{
/*
 *  Local Data,
 */
unsigned long	    iret;		/* Return status */
	 long	    status;		/* Return status from System V fns */
unsigned long	    diag_save;		/* Saved diagnostic context */
	 Boolean    param_error;	/* Parameter problem? */
	 long	    flags;		/* Flags for msgsnd() */

/*
 *  Output diagnostic trace information
 */
#ifdef PSL_TRACING
    trace_flag = PSL_TEST_TRACE(PTRC_GEN_MISC);
    if (trace_flag)
    {
	PSL_OUTPUT_TRACE("PSL_MSGF_SEND qid=%d", qid);
	PSL_OUTPUT_TRACE(" dat_ptr=%p, dat_len=%u, flags=%x, diag_ptr=%p\n",
			 dat_ptr, dat_len, msgf_flags, diag_ptr);
    }

#endif	/* PSL_TRACING */

/*
 *  Push message onto diagnostics stack indicating the current function
 */
    diag_save = PSL_ADD_STATDIAG(diag_ptr, "During PSL_MSGF_SEND");

/*
 *  Assume no parameter problems and successful routine.
 */
    param_error = PFALSE;
    iret = PSL_NORMAL;

/*
 *  Check for null pointer values and zero lengths
 */
    if (qid < 0)			/* No queue id */
    {
	param_error = PTRUE;
	(void) PSL_ADD_DIAGNOSTIC(diag_ptr,
	    "Invalid NULL value for queue id");
    }

    if (dat_ptr == NULL)		/* No data to send */
    {
	param_error = PTRUE;
	(void) PSL_ADD_DIAGNOSTIC(diag_ptr,
			   "Invalid NULL value for data pointer");
    }

    if (dat_len == 0)			/* Zero data length */
    {
	param_error = PTRUE;
	(void) PSL_ADD_DIAGNOSTIC(diag_ptr, "Invalid zero data length");
    }

/*
 *  If there has been a parameter problem, report it and exit.
 */
    if (param_error)
    {
	iret = PSL_BADPARAM;
	PSL_PUTDIAG(diag_ptr,
		    2,
		    PSL_SYSERVERR,
		    "PSL_MSGF_SEND",
		    iret);
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	goto end;
    }

/*
 *  Setup the message type and copy the message text into the buffer.
 */
    msgf_buf_ptr->mtype = 1;					    
    (void) memcpy(msgf_buf_ptr->mtext, dat_ptr, dat_len);

/*
 *  Set up the flags.
 *  If the user did not ask for us to wait if the queue is full then we set
 *  IPC_NOWAIT.
 */
    flags = 0;
    if ((msgf_flags & PSLMSGF_WAITSPC) == 0)
    {
	flags |= IPC_NOWAIT;
    }

/*
 *  Send the message
 */
    errno = 0;				/* Reset error status */
    status = msgsnd(qid, msgf_buf_ptr, dat_len, flags);	
    if (status == -1)			/* Failed to send message */
    {
/*
 *	If the queue does not exist then we must return a special status
 *	so that calling utilities know what has happened.
 */
	if (errno == EINVAL)		/* Queue does not exist */
	{
	    iret = PSL_FACNOEXISTS;
	    (void) PSL_ADD_DIAGNOSTIC(diag_ptr, "Message queue not found");
	    PSL_PUTDIAG(diag_ptr, 1, iret);
	}
	else
	{
	    iret = PSL_DIAG_ERRNO(diag_ptr, "msgsnd()");
	}
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	goto end;
    }

/*
 *  If the user specified PSLMSGF_WAITIOREQ then we must wait until there
 *  the message we have just sent has been read. Affectively this means
 *  waiting until the queue is empty or it disappears.
 */
    if ((msgf_flags & PSLMSGF_WAITIOREQ) != 0)
    {
	iret = AWAIT_EMPTY_QUEUE(qid, diag_ptr);
	if (!(psl_test_status(iret))) goto end;
    }
/*
 *  Successful completion so return diagnostics stack to its pre-function state
 */
    PSL_REMOVE_DIAGNOSTICS(diag_ptr, diag_save);

/*
 *  This label is branched to when an error occurs.
 */

end:

    return iret;
}

/************************************************************************/
/*									*/
/*	PSLOBS_MSGF_RECEIVE - OBSOLETE Receive message from mailbox	*/
/*									*/
/*	iret = PSLOBS_MSGF_RECEIVE(qid, buf_ptr, buf_len,		*/
/*				USER_FN, user_ptr, ios_ptr,		*/
/*				msgf_flags, diag_ptr);			*/
/*									*/
/*	iret      	=  Return status				*/
/*	qid		=  Message queue id				*/
/*	buf_ptr		=  Ptr to buffer to receive message		*/
/*	buf_len		=  Length of above buffer			*/
/*	USER_FN		=  Function to be called when message received  */
/*	user_ptr	=  Ptr passed to above function			*/
/*	ios_ptr		=  Ptr to read status return quadword		*/
/*	msgf_flags	=  Flags					*/
/*			   Currently only PSLMSGF_WAITIOREQ		*/
/*	diag_ptr	=  Ptr to the diagnostics stack			*/
/*									*/
/*	This function is an obsolete version of the PSl_MSGF_RECEIVE 	*/
/*	routine which has been re-written to allow a timeout to be 	*/
/*	specified.							*/
/*									*/
/*    This routine should NOT be called, use PSL_MSGF_RECEIVE instead	*/
/*									*/
/************************************************************************/

unsigned long PSLOBS_MSGF_RECEIVE(long qid,
			       void *buf_ptr,
			       unsigned long buf_len,
			       unsigned long (*USER_FN)(void *),
			       void *user_ptr,
			       unsigned long *ios_ptr,
			       unsigned long msgf_flags,
			       struct diagstk *diag_ptr)

{
    return PSL_MSGF_RECEIVE (qid ,buf_ptr ,buf_len ,USER_FN ,user_ptr
			    ,ios_ptr ,msgf_flags ,0 ,diag_ptr);
}

/************************************************************************/
/*									*/
/*	PSL_MSGF_RECEIVE - Receive message from mailbox			*/
/*									*/
/*	iret = PSL_MSGF_RECEIVE(qid, buf_ptr, buf_len,			*/
/*				USER_FN, user_ptr, ios_ptr,		*/
/*				msgf_flags, timeout ,diag_ptr);		*/
/*									*/
/*	iret      	=  Return status				*/
/*	qid		=  Message queue id				*/
/*	buf_ptr		=  Ptr to buffer to receive message		*/
/*	buf_len		=  Length of above buffer			*/
/*	USER_FN		=  Function to be called when message received  */
/*	user_ptr	=  Ptr passed to above function			*/
/*	ios_ptr		=  Ptr to read status return quadword		*/
/*	msgf_flags	=  Flags					*/
/*			   Currently only PSLMSGF_WAITIOREQ		*/
/*	timeout		= Time to wait to read a message		*/
/*	diag_ptr	=  Ptr to the diagnostics stack			*/
/*									*/
/*	This function removes a message from a message queue. Note	*/
/*	that although the parameters seem to suggest that an		*/
/*	asynchronous read is possible (i.e. USER_FN and user_ptr)	*/
/*	only synchronous message facility reads are possible. The	*/
/*	two above mentioned parameters are included in the prototype	*/
/*	purely for reasons of compatibility of calling sequences with	*/
/*	VMS								*/
/*									*/
/*	Note.								*/
/*	This function tries to construct an IOSB as in VMS. The		*/
/*	read_status_ptr points to a quadword containing the word status,*/
/*	word # of bytes in message and longword 0.			*/
/*									*/
/************************************************************************/

unsigned long PSL_MSGF_RECEIVE(long qid,
			       void *buf_ptr,
			       unsigned long buf_len,
			       unsigned long (*USER_FN)(void *),
			       void *user_ptr,
			       unsigned long *ios_ptr,
			       unsigned long msgf_flags,
			                long timeout,
			       struct diagstk *diag_ptr)

{
/*
 *  Local Data,
 */
unsigned long	    iret;		/* Return status */
	 long	    status;		/* Return status from System V fns */
unsigned long	    diag_save;		/* Saved diagnostic context */
	 Boolean    param_error;	/* Parameter problem? */
	 int	    flags;		/* Flags passed to msgsnd */
unsigned short	    *word_ptr;		/* Useful ptr to word */

/*
 *  Output diagnostic trace information
 */
#ifdef PSL_TRACING
    trace_flag = PSL_TEST_TRACE(PTRC_GEN_MISC);
    if (trace_flag)
    {
	PSL_OUTPUT_TRACE("PSL_MSGF_RECEIVE qid=%d", qid);
	PSL_OUTPUT_TRACE(" buf_ptr=%p, buf_len=%u, USER_FN=%p, user_ptr=%p",
			 buf_ptr, buf_len, &USER_FN, user_ptr);
	PSL_OUTPUT_TRACE(" ios_ptr=%p, flags=%x, diag_ptr=%p\n",
			 ios_ptr, msgf_flags, diag_ptr);
    }

#endif	/* PSL_TRACING */

/*
 *  Push message onto diagnostics stack indicating the current function
 */
    diag_save = PSL_ADD_STATDIAG(diag_ptr, "During PSL_MSGF_RECEIVE");

/*
 *  Assume no parameter problems and successful routine.
 */
    param_error = PFALSE;
    iret = PSL_NORMAL;

/*
 *  Check for null pointer values and zero lengths
 */

    if (qid < 0)			/* No queue id */
    {
	param_error = PTRUE;
	(void) PSL_ADD_DIAGNOSTIC(diag_ptr,
	    "Invalid NULL value for queue id");
    }

    if (buf_ptr == NULL)		/* No receive buffer */
    {
	param_error = PTRUE;
	(void) PSL_ADD_DIAGNOSTIC(diag_ptr, "Invalid NULL value for buffer pointer");
    }

    if (buf_len == 0)			/* Zero buffer length */
    {
	param_error = PTRUE;
	(void) PSL_ADD_DIAGNOSTIC(diag_ptr, "Invalid zero buffer length");
    }

/*
 *  If there has been a parameter problem, report it and exit.
 */
    if (param_error)
    {
	iret = PSL_BADPARAM;
	PSL_PUTDIAG(diag_ptr,
		    2,
		    PSL_SYSERVERR,
		    "PSL_MSGF_RECEIVE",
		    iret);
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	goto end;
    }

/*
 *  Setup the flags passed to msgrcv()
 */
    flags = 0;
    if ((msgf_flags & PSLMSGF_WAITIOREQ) == 0)
    {
	flags = IPC_NOWAIT;
    }

/*
 *  Receive the message
 */
    status = msgrcv(qid,		/* Queue id */
		    msgf_buf_ptr,	/* Message buffer structure */
		    buf_len,		/* Available characters for msg */
		    0,			/* Get any type of message */
		    flags);		/* Flags */

    if (status == -1)			/* Failed to read message */
    {

/*
 *	If the user did not ask us to wait for a message and the queue was
 *	empty we do not signal an error.
 */
	if (((msgf_flags & PSLMSGF_WAITIOREQ) == 0) &&
	    (errno == ENOMSG))
	{
	    status = 0;			/* Set length read = 0 */
	}
	else
	{
	    iret = PSL_DIAG_ERRNO(diag_ptr, "msgrcv()");
	    PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	}
    }

/*
 *  Set up the IOSB. The first word is the status, the second the # of
 *  bytes in the message and the rest is 0.
 */
    word_ptr = (unsigned short *)ios_ptr;
    *word_ptr = 1;			/* Save Success status */
    word_ptr++;
    *word_ptr = (unsigned short) status;
					/* Save # of bytes in message */
    *(ios_ptr + 1) = 0;			/* Reset the rest of quadword */

/*
 *  Copy the read message into the users buffer
 */
    if (status != -1)			/* Read a message */
    {
	(void) memcpy(buf_ptr, msgf_buf_ptr->mtext, buf_len);         
    }

/*
 *  Successful completion so return diagnostics stack to its pre-function state
 */
    PSL_REMOVE_DIAGNOSTICS(diag_ptr, diag_save);

/*
 *  This label is branched to when an error occurs
 */

end:;
    return iret;
}



/************************************************************************/
/*									*/
/*	PSL_MSGF_REMOVE - Remove a message facility			*/
/*									*/
/*	iret = PSL_MSGF_REMOVE(qid, diag_ptr);				*/
/*									*/
/*	iret      	=  Return status				*/
/*	qid		=  Message queue id				*/
/*	diag_ptr	=  Ptr to the diagnostics stack			*/
/*									*/
/*	This function removes a message queue from the system.		*/
/*									*/
/************************************************************************/
unsigned long PSL_MSGF_REMOVE(long qid,
			       struct diagstk *diag_ptr)

{
/*
 *  Local Data,
 */
unsigned long	    iret;		/* Return status */
	 long	    status;		/* Return status from System V fns */
unsigned long	    diag_save;		/* Saved diagnostic context */
	 Boolean    param_error;	/* Parameter problem? */

/*
 *  Output diagnostic trace information
 */
#ifdef PSL_TRACING
    trace_flag = PSL_TEST_TRACE(PTRC_GEN_MISC);
    if (trace_flag)
    {
	PSL_OUTPUT_TRACE(
	  "PSL_MSGF_REMOVE qid=%d, diag_ptr=%p\n", qid, diag_ptr);
    }

#endif	/* PSL_TRACING */

/*
 *  Push message onto diagnostics stack indicating the current function
 */
    diag_save = PSL_ADD_STATDIAG(diag_ptr, "During PSL_MSGF_REMOVE");

/*
 *  Assume no parameter problems and successful routine.
 */
    param_error = PFALSE;
    iret = PSL_NORMAL;

/*
 *  Check for null pointer values and zero lengths
 */

    if (qid < 0)		/* No queue id */
    {
	param_error = PTRUE;
	(void) PSL_ADD_DIAGNOSTIC(diag_ptr,
	    "Invalid NULL value for mailbox device name pointer");
    }

/*
 *  If there has been a parameter problem, report it and exit.
 */
    if (param_error)
    {
	iret = PSL_BADPARAM;
	PSL_PUTDIAG(diag_ptr,
		    2,
		    PSL_SYSERVERR,
		    "PSL_MSGF_REMOVE",
		    iret);
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	goto end;
    }

/*
 *  Just call msgctl() to remove the facility
 */
    errno = 0;				/* Reset error status */
    status = msgctl(qid, IPC_RMID, NULL);
    if (status == -1)			/* Failed to remove queue */
    {
	iret = PSL_DIAG_ERRNO(diag_ptr, "msgctl()");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	goto end;
    }

/*
 *    ...and free up the buffer allocated to hold the messages
 */
    (void) free (msgf_buf_ptr);					    

/*
 *  Successful completion so return diagnostics stack to its pre-function state
 */
    PSL_REMOVE_DIAGNOSTICS(diag_ptr, diag_save);

/*
 *  This label is branched to when an error occurs.
 */

end:;
    return iret;
}



/************************************************************************/
/*									*/
/*	PSL_MSGF_CLOSE - Detach from message facility			*/
/*									*/
/*	iret = PSL_MSGF_CLOSE(qid, diag_ptr);				*/
/*									*/
/*	iret      	=  Return status				*/
/*	qid		=  Message queue id				*/
/*	diag_ptr	=  Ptr to the diagnostics stack			*/
/*									*/
/*	This function detaches the current process from the message	*/
/*	queue facility. This is a NOP in UNIX.				*/
/*									*/
/************************************************************************/
unsigned long PSL_MSGF_CLOSE(long qid,
			       struct diagstk *diag_ptr)

{
    return PSL_NORMAL;
}



/************************************************************************/
/*									*/
/*	GET_QUEUE - Calculate a key for a message queue {& queue id}	*/
/*									*/
/*	    iret = GET_QUEUE(idstr_ptr, key_ptr, id_ptr, diag_ptr)	*/
/*									*/
/*	iret      	=  Return status				*/
/*	idstr_ptr	=  Ptr to queue id string			*/
/*	key_ptr		=  Ptr to long to receive key			*/
/*	id_ptr		=  Ptr to returned queue id or NULL		*/
/*			   N.B. if this is non NULL then queue assumed	*/
/*			   to exist and id returned, therefore error if	*/
/*			   queue does not exist.			*/
/*	diag_ptr	=  Diagnostics stack pointer			*/
/*									*/
/*	This function uses the supplied idstr to formulate the name of  */
/*	a file (<pslnodedir>/<idstr>) which must exist. The unique id of*/
/*	this file becomes the 'key' to obtain the id of a message queue */
/*									*/
/************************************************************************/
static unsigned long GET_QUEUE(char *idstr_ptr,
			       long *key_ptr,
			       long *id_ptr,
			       struct diagstk *diag_ptr)
{
/*
 *  Local data
 */
unsigned long	    iret;		/* PSL Return status */
	 char	    key_file[256];	/* File used to obtain key */


    iret = PSL_NORMAL;			/* Assume successful */

/*
 *  Build the key filespec up from the supplied id string.
 */
    GET_KEY_FILE(idstr_ptr, key_file, sizeof (key_file));

/*
 *  Now get the key for the message queue from the keyfile
 */
    errno = 0;				/* Reset the error number */
    *key_ptr = ftok(key_file, (char) PSL_PROJECT_ID);	
    if(*key_ptr == -1)			/* Failed to get key */
    {
	iret = PSL_DIAG_ERRNO(diag_ptr, "ftok()");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	goto end;
    }

/*
 *  If the id_ptr is non NULL then we assume that the queue already exists
 *  and try to get the id.
 */
    errno = 0;				/* Reset the error number */
    if (id_ptr != NULL)
    {
	*id_ptr = msgget(*key_ptr, 0);
	if (*id_ptr == -1)		/* Error */
	{
	    iret = PSL_DIAG_ERRNO(diag_ptr, "msgget()");
	    PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	    goto end;
	}				/* End of error from msgget() */
    }					/* End of get queue id */

/*
 *  This label is branched to when an error occurs.
 */

end:
    return iret;
}



/************************************************************************/
/*									*/
/*	AWAIT_EMPTY_QUEUE - Wait for the message queue to be empty     	*/
/*									*/
/*	    iret = AWAIT_EMPTY_QUEUE(qid)				*/
/*									*/
/*	iret	= PSL return status					*/
/*		  PSL_NORMAL = we have an empty queue			*/
/*		  anything else is an error				*/
/*	qid	= Queue id						*/
/*									*/
/*	This function only returns when it spots that the queue is	*/
/*	empty or has gone away.						*/
/*									*/
/************************************************************************/
static unsigned long AWAIT_EMPTY_QUEUE(long qid,
				       struct diagstk *diag_ptr)
{
/*
 *  Local Data
 */
	 long	    status;		/* msgxxx() status */
struct	 msqid_ds   msg_struct;		/* Message queue structure */
unsigned long	    iret;		/* PSL return status */

    msg_struct.msg_qnum = 1;
/*
 *  Loop around waiting for an empty queue
 */
    while (msg_struct.msg_qnum != 0)
    {
	errno = 0;
	status = msgctl(qid, IPC_STAT, &msg_struct);
	if (status == -1)
	{
/*
 *	    If the queue does not exist then we must return a special status
 *	    so that calling utilities know what has happened.
 */
	    if (errno == EINVAL)	/* Queue does not exist */
	    {
	    	iret = PSL_FACNOEXISTS;
	    	(void) PSL_ADD_DIAGNOSTIC(diag_ptr, "Message queue not found");
								      /*40X004*/
	    	PSL_PUTDIAG(diag_ptr, 1, iret);
	    }
	    else
	    {
		iret = PSL_DIAG_ERRNO(diag_ptr, "msgsnd()");
	    }
	    PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_MSGF, iret);
	    goto end;
	}
	(void) sleep(1);		/* Sleep for 1 second */
    }
    iret = PSL_NORMAL;

/*
 *  This label is branched to with an error or is dropped into from above if
 *  the function executes successfully.
 */
end:
    return iret;
}


/************************************************************************/
/*									*/
/*	GET_KEY_FILE - Form the name of the key file from the id string	*/
/*									*/
/*	    GET_KEY_FILE(idstr_ptr, keyfile_ptr, keyfile_size)		*/
/*									*/
/*	idstr_ptr	=  Ptr to queue id string			*/
/*	keyfile_ptr	=  Ptr to returned key file name		*/
/*	keyfile_size	=  Size of key_file buffer       		*/
/*									*/
/*	This function forms the key file name from the id string.	*/
/*	The key file is used to obtain a unique id for the message	*/
/*	facility.							*/
/*									*/
/************************************************************************/
static void GET_KEY_FILE(char *idstr_ptr,
			 char *keyfile_ptr,
			 unsigned long keyfile_size)		   
{
    PSL_GET_NODEDIR (keyfile_ptr, keyfile_size);
    (void) strcat(keyfile_ptr, "/");			
    (void) strcat(keyfile_ptr, idstr_ptr);
    return;
}

#endif
