/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/* scl_enc_routines.c:
 * ===================
 *
 * This module contains routines used to encode data into an appropriate
 * format for output to an object file.
 */

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines     */
#include <trc_rtndef.h>         /* Tracing Routines     */
#include "scl_rtndef.h"         /* SCL Routines     */
#include "scl_def_tbl_rtndef.h"     /* SCL definitions routines */
#include "scl_env_tbl_rtndef.h"     /* SCL environment routines */
#include "scl_cod_tbl_rtndef.h"     /* SCL code  routines */
#include "scl_sub_tbl_rtndef.h"     /* SCL subexpression routines */

static      void    ENC_NULL_OPD (struct scl_tpa  *tpa_ptr);

static      void    ENC_NUMERIC_OPD (struct scl_tpa *tpa_ptr ,long number);

static      void    ENC_OPERAND     (struct   scl_tpa *tpa_ptr
                                     ,         Boolean remove_enttryn
                                     ,     char    *llb_pos_ptr
                                     ,     char    **ret_llb_pos_pptr);

static unsigned long    GET_OPD_COUNT   (struct scl_tpa *tpa_ptr);

static void GET_OPERAND_POS     (struct scl_tpa *tpa_ptr
                                 ,       long    number
                                 ,       char    **llb_pos_pptr);

static void FORWARD_ONE_OPERAND     (struct scl_tpa *tpa_ptr
                                     ,struct oph     **oph_pptr
                                     ,struct opd     **opd_pptr
                                     ,       char    **llb_pos_pptr);

/*
 *  Structure definitions.
 */
#include <shrmsg.h>         /* PSL Messages         */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block    */
#include <scpdef.h>         /* SCP Header record        */
#include <psl_trc_prmdef.h>     /* Tracing parameters       */
#include <psl_vm_prmdef.h>      /* Linked list parameters   */
#include <pslerr_prmdef.h>      /* Error routine parameters */
#include "scl_tbldef.h"         /* SCL Table Definitions    */
#include <vtest_prmdef.h>       /* VTEST Parameters         */
#include "scl_cmd.h"            /* Command Line Data Structure  */
#include "scl_comdef.h"         /* Common Data          */
#include "scl_limits.h"         /* SCL limits           */
#include "scl_tpacom.h"         /* Tparse common Data       */

/*======================================================================*/
/*                                  */

/*                                  */
/*    ENC_CASE - Encode CASE command                    */
/*    ===========================                   */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_CASE (argblk_ptr)                    */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*                                  */
/*                                      */
/*    This routine adds a CASE command to the code buffer if there have */
/*    been no previous errors in the current switch statement.      */
/*                                  */
/*    CASE  case_value, switch_variable, next_case_offset           */
/*                                 (If this case is not selected)   */
/*                                  */
/*    case_value -       last operand parsed.               */
/*                                  */
/*    switch_variable -  taken from the saved operand stack.            */
/*                                      */
/*    next_case_offset - encoded as NULL until it can be resolved       */
/*                       when the next case is parsed.                  */
/*                                  */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long   ENC_CASE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com    *com_ptr;       /* Common Data          */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE Data          */
    struct   swstack    *sw_ptr;            /* Switch stack                 */
    struct   scph       *header_ptr;    /* Pointer to operand header    */
    long       offset;     /* Current code offset          */
    char       *llb_sw_var_pos;     /* Linked list position of switch   */
    /* variable */
    char       *llb_case_var_pos;   /* Linked list position of case val*/
    char*      GotoOffset;     /* Case fal-through offset  */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    header_ptr = (*tpa_ptr).header_begps;
    sw_ptr = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl - 1 ;
    /* Point to current entry   */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a CASE Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif


    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Do not encode anything if we have had any error in a
     *    switch statement.
     */
    if ((*tpa_ptr).sw_error_fgl == 1 ) goto end;

    /*
     * If case count = 1 ie the first case at this level - there is no previous
     * case to assign an offset to. Therefore excluded from assigning offset
     * to previous case.
     */
    if  ((*sw_ptr).case_countl > 1)
    {

        /*
         * 
         *
         *    Before we encode this CASE we shall encode a jump around it to its code.
         *    The purpose of this is to allow the previous code to 'drop through' if 
         *    a break was not specified.
         */
        ENC_INTEGER (tpa_ptr, OPC_GOTO); /* Encode a GOTO opcode    */
        GotoOffset = (*tpa_ptr).code_ptrpc;
        /* ...save offset of jump offset */
        ENC_INTEGER (tpa_ptr,NULL_DISP);/* ...and allocate space for it  */

        /*
         * 
         *
         * Save current code offset  - this is the offset that the previous
         * case statement at the current switch level will jump to if switch variable
         * and case value don't match.
         */
        offset = (*tpa_ptr).code_ptrpc - (*tpa_ptr).code_begpc
                 + (*header_ptr).scph_codoff_l;

        /*
         * Encode this offset at the previous case "FALSE" offset
         */
        ENC_INTEGER_AT ((*sw_ptr).last_case_offpc ,offset);

        /*
         * Tracing information if required
         */

#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE ("\n1.ENC_CASE* offset last case (post) = %ld\n"
                              ,*(*sw_ptr).last_case_offpc);
            PSL_OUTPUT_TRACE ("2.ENC_CASE* current offset = %ld \n",offset);
        }
#endif
    }

    /*
     * NOW - for every case statement:
     * Encode the CASE opcode followed by operands and allocation of
     * space for false offset.
     */
    ENC_INTEGER (tpa_ptr, OPC_CASE);

    /*
     * Tracing information if required
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("3.ENC_CASE* Encoded integer OPC_CASE\n");
    }
#endif

    /*
     * Get the case value operand and save it at llb_case_var_pos
     */

    GET_OPERAND_POS(tpa_ptr,0,&llb_case_var_pos);

    /*
     * Tracing information if required
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("4.ENC_CASE*Got operand position of case value \n");
    }
#endif

    /*
     * Encode this saved operand
     */

    ENC_OPERAND (tpa_ptr,PTRUE,llb_case_var_pos,NULL);

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE("5.ENC_CASE*Encoded operand 1 \n");
    }
#endif

    /*
     * Encode operand - Switch variable
     * This is from the saved parsed operand LL.
     * Delete CURRENT parsed operand LL and revert to one saved in swstack.
     * This contains the switch variable
     */
    (*tpa_ptr).parsed_op_templlps = (*tpa_ptr).parsed_operand_llps;
    (*tpa_ptr).parsed_operand_llps = (*sw_ptr).sw_operand_llps;

    /*
     * Tracing information if required
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE("6.ENC_CASE*Swapped operand LList\n");
    }
#endif

    /*
     * Switch variable is first item in this list -
     * Get this operand.
     */
    GET_OPERAND_POS(tpa_ptr,0,&llb_sw_var_pos);

    /*
     * Tracing information if required
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE("7.ENC_CASE*Got operand position\n");
    }
#endif

    /*
     * Encode this saved operand
     */
    ENC_OPERAND (tpa_ptr,PFALSE,llb_sw_var_pos,NULL);

    /*
     * Tracing information if required
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE("8.ENC_CASE*Encode operand 2 \n");
    }
#endif

    /*
     * Restore linked list
     */
    (*tpa_ptr).parsed_operand_llps =  (*tpa_ptr).parsed_op_templlps;

    /*
     *Tracing information if required
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE("9.ENC_CASE*Restore linked list\n");
    }
#endif

    /*
     * Save address of current case to be resolved at next case for this level
     * or in default case.
     */
    (*sw_ptr).last_case_offpc = (*tpa_ptr).code_ptrpc;

    /*
     * Tracing information if required
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE("10.ENC_CASE*Case offset =%ld\n",(*sw_ptr).last_case_offpc);
    }
#endif

    /*
     * Allocate space for next case offset, this will be resolved either by next
     * case, the default case or the end_switch command.
     */
    ENC_INTEGER (tpa_ptr,NULL_DISP);

    /*
     * 
     *
     *    If this is the second or subsequent case we will have encoded part of a
     *    GOTO earlier in this routine to allow us to branch around the CASE
     *    statement itself to implement CASE fall-through. Now fill in the offset
     *    to which this jump should branch as we have encoded the CASE check.
     */
    if  ((*sw_ptr).case_countl > 1)
    {
        offset = (*tpa_ptr).code_ptrpc - (*tpa_ptr).code_begpc
                 + (*header_ptr).scph_codoff_l;
        ENC_INTEGER_AT (GotoOffset ,offset);
    }

    /*
     *  
     *
     *    Case command encoding complete
     */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*
 * 
 */


/*======================================================================*/
/*                                  */

/*                                  */
/*    ENC_BREAK - Encode BREAK command                  */
/*    =========                             */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_BREAK (argblk_ptr)                   */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*                                  */
/*                                                                      */
/*    The following is encoded only if there have been no previous      */
/*    errors in the current switch command.             */
/*                                  */
/*    There is no BREAK command - it is implemented as :                */
/*                                  */
/*      GOTO End Switch                                         */
/*                                  */
/*    When a break is parsed GOTO is encoded followed by space          */
/*    for END_SWITCH address. This space after the GOTO is used         */
/*    to store the last address where space has been allocated          */
/*    for the END_SWITCH address.                                       */
/*                                          */
/*    When END_SWITCH is parsed all the breaks requiring goto addresses */
/*    at the current level can be resolved.                             */
/*                                      */
/*    break = 1     GOTO   (offset1_addrl ->)  null                 */
/*                                                                      */
/*    break = 2         GOTO   (offset2_addrl ->)  offset1_addrl        */
/*        .                                                             */
/*        .                                                             */
/*        .                                                             */
/*    LAST BREAK:                               */
/*                                  */
/*    break = n         GOTO   (break_addrpl -> )  offsetn_addrl        */
/*                                                                      */
/*                                                                      */
/*    The contents of BREAK_ADDRPL = address of last break where        */
/*    END_SWITCH address needs to be encoded.               */
/*                                                                      */
/*                                  */
/*----------------------------------------------------------------------*/
unsigned long   ENC_BREAK (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com    *com_ptr;       /* Common Data          */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE Data          */
    struct   swstack    *sw_ptr;            /* Switch stack                 */
    void    *void_ptr;     /* Work pointer to void     */
    long       offset;     /* Current code offset          */
    /*
     *    Copy pointers for ease of reference.
     */

    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    sw_ptr = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl - 1 ;
    /* Point to current switch entry */

    /*
     *    Output tracing information if required.
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a BREAK Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif


    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Do not encode anything if we have had any error in the
     *    switch statement.
     */
    if ((*tpa_ptr).sw_error_fgl == 1 ) goto end;

    /*
     * Increment Break_count
     */

    (*sw_ptr).break_countl += 1;

    /*
     * For the first BREAK there is no previous break to assign an offset to.
     * There fore this case has its own assigning code.
     */

    if ((*sw_ptr).break_countl == 1)
    {
        ENC_INTEGER (tpa_ptr, OPC_GOTO);

        /*
         * Save current code pointer this will be used later
         * to resolve all the break addresses
         * Code pointer (character) has to be cast as a long as it will be
         * "saved" as an integer in ENC_INTEGER later.
         * Because of alignment warnings in the compiler it done via a
         * void_ptr.
         */
        void_ptr = (void *) (*tpa_ptr).code_ptrpc;

        (*sw_ptr).break_addrpl = (long *) void_ptr;

        /*
         * Tracing information if required                      
         */

#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE ("\n1.ENC_BREAK*Contents of %ld is%ld\n",
                              (*sw_ptr).break_addrpl,*(*sw_ptr).break_addrpl);
        }
#endif

        /*
         * Allocate space for first break goto address
         */

        ENC_INTEGER (tpa_ptr, NULL_DISP);

    }

    /*
     * If this is not the first break it is more complicated-
     * We have to keep track off all places where "break" addresses are required
     */
    else
    {

        /*
         * Encode GOTO opcode
         */

        ENC_INTEGER (tpa_ptr, OPC_GOTO);

        /*
         * Break_addrpl points to the address of the last break command,
         * this is stored in  offset and then be encoded with the current
         * break commands GOTO.
         */

        offset = (long) (*sw_ptr).break_addrpl;

        /*
         * Tracing inforamtion if required
         */

#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE("2.ENC_BREAK*Contents of %ld is %ld \n",
                             (*sw_ptr).break_addrpl,*(*sw_ptr).break_addrpl);
            PSL_OUTPUT_TRACE("3.ENC_BREAK*.Offset = %ld \n", offset);
            PSL_OUTPUT_TRACE("4.ENC_BREAK*.Break_addrpc = %ld \n",
                             (*sw_ptr).break_addrpl);
        }
#endif

        /*
         * Code point is reassigned the current code offset.
         * Break_addrpl is then assigned this value.
         * Again a cast has been used because break_addrpl will be stored as an
         * integer.
         * Because of warnings in compiler of alignment problems this is done
         * via a void_ptr.
         */
        void_ptr = (void *) (*tpa_ptr).code_ptrpc;

        (*sw_ptr).break_addrpl = (long *) void_ptr;

        /* Save current address     */

        /*
         * Space is allocated for the end_switch address (where GOTO goes to in break)
         */

        ENC_INTEGER(tpa_ptr,NULL_DISP);         /* Allocate space for next   */
        /* address to be stored      */
        /*
         * Encode last break address at this address - ie saving previous break
         * address. This will be over written when the end switch is parsed
         * break_addrpl has to be cast as a char pointer as that is the type
         * ENC_INTEGER_AT is expecting.
         */

        ENC_INTEGER_AT(((char *) (*sw_ptr).break_addrpl),offset);
        /* Encode last break address */
        /* at current address        */
        /*
         * Tracing information if required
         */

#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE ("5.ENC_BREAK*Contents of %ld is %ld \n",
                              (*sw_ptr).break_addrpl,*(*sw_ptr).break_addrpl);
        }
#endif

    }
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*
 * 
 */

/*======================================================================*/
/*                                  */

/*                                  */
/*    ENC_END_SW - Encode END SWITCH command                */
/*    ==========                            */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_END_SW (argblk_ptr)                  */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*                                  */
/*                                  */
/*    There is no END_SWITCH command - it is a label for BREAK to GOTO  */
/*                                  */
/*                                      */
/*    When a break is parsed GOTO is encoded followed by space          */
/*    for END_SWITCH address. This space after the GOTO is used         */
/*    to store the last address where space has been allocated          */
/*    for the END_SWITCH address.                                       */
/*                                          */
/*    When END_SWITCH is parsed all the breaks requiring goto addresses */
/*    at the current level can be resolved.                             */
/*                                      */
/*    break = 1     GOTO   (offset1_addrl ->)  null                 */
/*                                                                      */
/*    break = 2         GOTO   (offset2_addrl ->)  offset1_addrl        */
/*        .                                                             */
/*        .                                                             */
/*        .                                                             */
/*    LAST BREAK:                               */
/*                                  */
/*    break = n         GOTO   (break_addrpl -> )  offsetn_addrc        */
/*                                                                      */
/*                                                                      */
/*    The contents of BREAK_ADDRPL = address of last break where        */
/*    END_SWITCH address needs to be encoded.               */
/*                                                                      */
/*    The contents of BREAK_ADDRPL are saved  in a temporary variable.  */
/*    The END_SWITCH offset is then encoded at location BREAK_ADDRPL    */
/*    points to.                                                        */
/*                                                                      */
/*    BREAK_ADDRPL is then re-assigned the  address stored in the temp. */
/*    variable. And so starts the loop to encode all END_SWITCH offsets */
/*    where there have been BREAKS at the current switch level.         */
/*                                                                      */
/*----------------------------------------------------------------------*/

unsigned long   ENC_END_SW (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com    *com_ptr;       /* Common Data          */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE Data          */
    struct   swstack    *sw_ptr;            /* Switch stack                 */
    struct   scph       *header_ptr;    /* Pointer to operand header    */
    long       offset;     /* Current code offset          */
    long       temp_addr;          /* Temp holder of break address */
    int        isub;       /* loop counter                 */

    /*
     *    Copy pointers for ease of reference.
     */

    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    header_ptr = (*tpa_ptr).header_begps;
    sw_ptr = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl;
    /* Pointer to current switch entry */



    /*
     *    Output tracing information if required.
     */

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding an END SWITCH LABEL\n");
        PSL_OUTPUT_TRACE ("\nSwitch depth is %ld \n",(*tpa_ptr).sw_depthl);
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif


    /*
     *    Do not encode anything if we have had an error in the command.
     */

    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Do not encode anything if we have had any error in the
     *    switch statement.
     */

    if ((*tpa_ptr).sw_error_fgl == 1 ) goto end;

    /*
     * Step through all BREAK goto reference addresses replacing contents
     * with actual address of end_switch.
     */

    offset = (*tpa_ptr).code_ptrpc - (*tpa_ptr).code_begpc
             + (*header_ptr).scph_codoff_l;

    /*
     * Loop through all BREAK addresses - only go down to break_countl = 2.
     * The resolving of address for the last break is done separately as prior
     * to resolving the contents of the space allocated,its contents will be NULL
     * and will cause access violation error if used in the loop below.
     */

    if ((*sw_ptr).break_countl > 1)
    {
        for (isub=(*sw_ptr).break_countl;isub!=1;isub--)
            /* Only go as far as count = 2 */
        {
#ifdef PSL_TRACING
            if (PSL_TEST_TRACE (PTRC_USR_USER1))
            {
                PSL_OUTPUT_TRACE("\n1.ENC_END_SW*In while loop\n");
                PSL_OUTPUT_TRACE("2.ENC_END_SW* offset = %ld\n",offset);
            }
#endif

            temp_addr =  *(*sw_ptr).break_addrpl;
            /* Holds address of previous   */
            /* BREAK/GOTO                  */
#ifdef PSL_TRACING
            if (PSL_TEST_TRACE (PTRC_USR_USER1))
            {
                PSL_OUTPUT_TRACE("3.ENC_END_SW*0.Temp_addr = %u \n",temp_addr);
                PSL_OUTPUT_TRACE("4.ENC_END_SW*1.Break_addr = %u \n",(*sw_ptr).break_addrpl);
            }
#endif


            ENC_INTEGER_AT(((char *) (*sw_ptr).break_addrpl), offset);
            /* Encode offset at current*/
            /* break address           */
#ifdef PSL_TRACING
            if (PSL_TEST_TRACE (PTRC_USR_USER1))
            {
                PSL_OUTPUT_TRACE("5.ENC_END_SW*int encoded is %ld\n",offset);
                PSL_OUTPUT_TRACE("6.ENC_END_SW*Contents of Break_addrpc = %u\n",
                                 *(*sw_ptr).break_addrpl);
                PSL_OUTPUT_TRACE("7.ENC_END_SW* Contents of %ld is %ld \n",
                                 (*sw_ptr).break_addrpl,*(*sw_ptr).break_addrpl);
            }
#endif

            (*sw_ptr).break_addrpl =  (long *) temp_addr;
            /* Re-assign BREAK_ADDRPL to */
            /* point at previous BREAK   */
#ifdef PSL_TRACING
            if (PSL_TEST_TRACE (PTRC_USR_USER1))
            {

                PSL_OUTPUT_TRACE("8.ENC_END_SW*Contents of %ld is %ld \n",
                                 (*sw_ptr).break_addrpl,*(*sw_ptr).break_addrpl);

                PSL_OUTPUT_TRACE("9.ENC_END_SW*4.Contents temp_addr = %u \n",
                                 temp_addr);

                PSL_OUTPUT_TRACE("10.ENC_END_SW*5.Break_addr = %ld  \n",
                                 (*sw_ptr).break_addrpl);

                PSL_OUTPUT_TRACE("11.ENC_END_SW*6.Contents of Break_addr = %ld \n",
                                 *(*sw_ptr).break_addrpl);
            }
#endif

            (*sw_ptr).break_countl -= 1;
            /* Decrement break_countl    */
        }
    }
    /*
     * Separate case for last BREAK to be resolved as there is no (valid) previous
     * address stored before END_SWITCH encoded.
     */

    if ((*sw_ptr).break_countl == 1)
    {

        ENC_INTEGER_AT(((char *) (*sw_ptr).break_addrpl), offset);
        /* Encode offset at current*/
        /* break address           */
#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE("1.1 ENC_END_SW*Contents of %ld is %ld \n",
                             (*sw_ptr).break_addrpl,*(*sw_ptr).break_addrpl);
        }
#endif

        (*sw_ptr).break_countl -= 1;                   /* Decrement break count  */
    }

    /*
     * If there is no DEFAULT_CASE the offset of end_switch has to be encoded
     * in last case false_offset.
     */

    if ((*sw_ptr).def_case_fgl == 0)
    {
        ENC_INTEGER_AT ((*sw_ptr).last_case_offpc,offset);

#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER1))
        {
            PSL_OUTPUT_TRACE("1.2 ENC_END_SW*No default case \n");
        }
#endif

    }
end:
    ;

    return PSL_NORMAL;          /* ...and return success    */
}

/*
 * 
 */

/*======================================================================*/
/*                                  */

/*                                  */
/*    ENC_DEF_CASE - Encode DEFAULT CASE command            */
/*    ===========================                   */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_DEF_CASE (argblk_ptr)                */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*                                  */
/*    This does not encode anything but assigns default case offset     */
/*    to the offset in the last CASE                                    */
/*                                  */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_DEF_CASE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com    *com_ptr;       /* Common Data          */
    struct   scl_tpa    *tpa_ptr;       /* TPARSE Data          */
    struct   swstack    *sw_ptr;            /* Switch stack                 */
    struct   scph       *header_ptr;    /* Pointer to operand header    */
    long       offset;     /* Current code offset          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    header_ptr = (*tpa_ptr).header_begps;
    sw_ptr = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl - 1 ;
    /* Point to current entry   */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a DEF_CASE Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif


    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Do not encode anything if we have had any error in the
     *    switch statement.
     */
    if ((*tpa_ptr).sw_error_fgl == 1 ) goto end;

    /*
     * Current code offset
     */

    offset = (*tpa_ptr).code_ptrpc - (*tpa_ptr).code_begpc
             + (*header_ptr).scph_codoff_l;

    /*
     * Encode current offset at previous case offset
     */

    ENC_INTEGER_AT ((*sw_ptr).last_case_offpc ,offset);

#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE("1.ENC_DEF_CASE*Data type %ld \n",(*sw_ptr).sw_dtypl);
    }
#endif


    /*
     * Default case is just a GOTO next commmand
     */

    ENC_INTEGER (tpa_ptr, OPC_GOTO);

    /*
     * Current offset.
     */

    offset = (*tpa_ptr).code_ptrpc - (*tpa_ptr).code_begpc
             + (*header_ptr).scph_codoff_l;

    /*
     *  Next command offset, encoded.
     */
    offset += sizeof (long);

    ENC_INTEGER (tpa_ptr, offset);
end:
    ;

    return PSL_NORMAL;          /* ...and return success    */
}

/*
 * 
 */

/*======================================================================*/
/*                                  */
/*    ENC_DEFINE_INPUT - Encode the DEFINE INPUT RATE command       */
/*    =================                         */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_DEFINE_INPUT (argblk_ptr)                */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds an DEFINE INPUT RATE instruction to the object  */
/*    code buffer.                          */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_DEFINE_INPUT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a DEFINE INPUT RATE Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the DEFINE INPUT RATE opcode.
     */
    ENC_INTEGER (tpa_ptr, OPC_DEFINE);  /* ...and encode the opcode */

    /*
     *    This is followed by OPD1 which defines the rate.
     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    /* ...and encode rate operand   */

end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_DEFINE_OUTPUT - Encode the DEFINE OUTPUT RATE command     */
/*    =================                         */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_DEFINE_OUTPUT (argblk_ptr)               */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds an DEFINE OUTPUT RATE instruction to the object */
/*    code buffer.                          */
/*                                  */
/*    A Define Output rate command has the following format:        */
/*      OPC_DEFINEOUT  RATE_OPD                     */
/*                                      */
/*----------------------------------------------------------------------*/

unsigned long   ENC_DEFINE_OUTPUT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a DEFINE OUTPUT RATE Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the DEFINE OUTPUT RATE opcode.
     */
    ENC_INTEGER (tpa_ptr, OPC_DEFINEOUT);
    /* ...and encode the opcode */

    /*
     *    This is followed by OPD1 which defines the rate.
     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    /* ...and encode rate operand   */

end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_DISABLE - Encode the DISABLE command              */
/*    ============                          */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_DISABLE (argblk_ptr)                 */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds an DISABLE instruction to the code buffer.      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_DISABLE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    long    opcode;        /* Opcode of instruction    */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a DISABLE TYPEAHEAD\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Check to see which script option has been requested (currently only
     *    typeahead is supported) and set the opcode accordingly.
     */
    if (((*tpa_ptr).scl_optl & SCLOPT_TYPEAHEAD) != 0)
    {                   /* If typeahead option specified */
        opcode = OPC_DISTYPAHD;     /* Set opcode = typeahead   */
    }

    ENC_INTEGER (tpa_ptr, opcode);  /* ...and encode the opcode */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_ENABLE - Encode the ENABLE command                */
/*    ==========                            */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_ENABLE (argblk_ptr)                  */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds an ENABLE instruction to the code buffer.       */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_ENABLE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    long    opcode;        /* Opcode of instruction    */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("Encoding an ENABLE TYPEAHEAD Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Check to see which script option has been requested (currently only
     *    typeahead is supported) and set the opcode accordingly.
     */
    if (((*tpa_ptr).scl_optl & SCLOPT_TYPEAHEAD) != 0)
    {                   /* If typeahead option specified */
        opcode = OPC_ENTYPAHD;      /* Set opcode = typeahead   */
    }

    ENC_INTEGER (tpa_ptr, opcode);  /* ...and encode the opcode */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_HEADER_CODE - Encode the header record code.          */
/*    ================                          */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_HEADER_CODE (argblk_ptr)             */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   RANDOM_VAL_CNT = Random Value List/Range Count      */
/*          SCRIPT_VAL_CNT = Script variables with associated   */
/*               values count               */
/*          VAL_CNT = Value List/Range count            */
/*      REC_CNTL = Number of records in script          */
/*          VARIABLES_TABLE = Variables Symbol Table        */
/*    Outputs:  OBJ_CODE_PTR = Object code pointer updated      */
/*                                  */
/*    This routine adds instructions to the head of the code section    */
/*    of a script to initialise all value pointers, seed variables and  */
/*    value range data in the Local and Script Data Areas:      */
/*                                  */
/*    The initialization command consists of the OPC_INITIALIZATION     */
/*    opcode followed by 0 or more 'sub-commands' which perform the     */
/*    actual initialization, terminated by a zero sub-command code. */
/*                                      */
/*    There are 6 sub-commands which are as follows:            */
/*                                      */
/*  INI_STEPS - Initialize step checking counts             */
/*  INI_SEED - Initialize the seed for a {repeatable) random var.   */
/*  INI_RANGE - Initialize the range value for a value range    */
/*  INI_VALUE - Initialize the value list pointer for a value list  */
/*  INI_RECORD - Initialize record information          */
/*  INI_REC_ARRAY - Set up format defn. array for each record.  */
/*                                      */
/*    There can be multiple occurences of each and any of these sub-    */
/*    commands depending on the types and number of variables declared  */
/*                                      */
/*----------------------------------------------------------------------*/

unsigned long   ENC_HEADER_CODE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("Encoding Header/Initialization Commands\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the initialization opcode. This will potentially be followed
     *    by the sub-commands or an immediate integer of 0 to signify the end
     *    of the initialization.
     */
    ENC_INTEGER (tpa_ptr, OPC_INIT);    /* Encode the opcode        */

    /*
     *    If we have specified /CHECK=STEPS then encode the value of the maximum
     *    steps variable as a sub-command.
     */
    if ((*com_ptr).max_stepsl != 0)
    {               /* If /MAX_STEPS > 0 specified (or defaulted) */
        ENC_INTEGER (tpa_ptr, INI_STEPS);
        /* Encode STEPS sub-command opcode */
        ENC_INTEGER (tpa_ptr, (long) (*com_ptr).max_stepsl);
        /* ...followed by max steps value  */
    }

    /*
     *    If we value lists then we should add initialization commands for them
     */
    if ((*tpa_ptr).val_cntl != 0)   /* If we have some value lists  */
    {

        /*
         *    Initialize any value lists/ranges in the Local Data Area
         */
        ENC_INI_VALUES (com_ptr, ADDR_LOCAL);
        /* Initialize local data area   */

        if ((*tpa_ptr).script_val_cntl != 0)
        {               /* ...if some script value lists */

            /*
             *    Initialize any value lists/ranges in the Script Data Area
             */
            ENC_INI_VALUES (com_ptr, ADDR_SCRIPT);
            /* Initialize script data area */
        }
    }

    if ((*tpa_ptr).rec_cntl != 0)   /* If we have some records  */
    {
        /*
         * Store all necessary data in header
         *
         *  1. Store number of records (REC_CNTL)
         *  2. For each record:
         *      i.   Record index value.
         *      ii.  Number of items in record.
         *      iii. Null terminated format array
         */

        ENC_INTEGER (tpa_ptr, INI_RECORD);


        ENC_INTEGER ( tpa_ptr, (*tpa_ptr).rec_cntl);

        /*
         * Encode record index and format array.
         */

        ENC_REC_FORMAT ( com_ptr, (*tpa_ptr).rec_cntl);

    }

    /*
     *    We have finished all initialization - terminate the initialization
     *    command with a zero integer.
     */
    ENC_INTEGER (tpa_ptr, 0);       /* Encode a zero integer marker */

end:
    ;
    return PSL_NORMAL;
}

/*======================================================================*/
/*                                  */
/*    ENC_INI_VALUES - Add initialization for value lists       */
/*    ==============                            */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_INI_VALUES (argblk_ptr)              */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   area_type - unsigned long data area type        */
/*      com_ptr - Pointer to the SCL common data record     */
/*    Implicit Inputs:                          */
/*          VARIABLES_TABLE = Variables Symbol Table        */
/*    Outputs:  OBJ_CODE_PTR = Object code pointer updated      */
/*                                  */
/*    This routine is called from the script initilization routine to   */
/*    add sub-commands to the object code buffer to handle the initial- */
/*    ization of value lists/ranges of the specified addressing mode    */
/*                                  */
/*    The addressing mode may be LOCAL or SCRIPT.           */
/*                                  */
/*======================================================================*/

void    ENC_INI_VALUES  (struct   scl_com *com_ptr
                         ,unsigned long    area_type)

{
    /*
     *    Local data
     */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *symtbl_ptr;       /* Pointer variable symbol table */
    unsigned long    isub;          /* Loop control variable    */

    /*
     *    Copy pointers for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps;
    symtbl_ptr = (*tpa_ptr).variable_begps;
    /* Variables symbol table   */

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a Value List Initialization opcode\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Scan the variables symbol table looking for a variable of the correct
     *    addressingh mode and which has either a value list or a range associated
     *    with it.
     */
    for (isub=0; isub < (*tpa_ptr).var_cntl; isub++, symtbl_ptr++)
    {
        if ((*symtbl_ptr).var_addrl != area_type) goto loop;
        /* Skip this if not same addressing */

        /*
         *    Check to see if this variable has a value list and if so initialize the
         *    value list pointer for it.
         */
        if (((*symtbl_ptr).var_optl & VOPT_LIST) != 0)
        {               /* If we have a value list  */
            ENC_INTEGER (tpa_ptr, INI_VALUE);
            /* Encode the INI_VALUE sub-command */
            CREATE_VAR_OPD (tpa_ptr, symtbl_ptr);
            /* Create an operand for this var */
            ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
            /* ...and encode the operand    */
        }

        else if (((*symtbl_ptr).var_optl & VOPT_RANGE) != 0)
        {               /* ...else if we have a range   */
            ENC_INTEGER (tpa_ptr, INI_RANGE);
            /* Encode the INI_RANGE sub-command */
            CREATE_VAR_OPD (tpa_ptr, symtbl_ptr);
            /* Create an operand for this var */
            ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
            /* ...and encode the operand    */
        }

        else                /* If neither range or list     */
        {
            goto loop;          /* ...then branch to loop again */
        }

        /*
         *    Check for this being random or repeatable random and set the seed
         *    value as appropriate. RANDOM values have their seed set to 0. This
         *    will be sorted out later by LGR.
         */
        if (((*symtbl_ptr).var_optl & VOPT_RANDOM) != 0)
        {               /* If REPEATABLE RANDOM variable */
            if (((*symtbl_ptr).var_optl & VOPT_SEED) != 0)
            {               /* If SEED value specified  */
                ENC_INTEGER (tpa_ptr, (*symtbl_ptr).var_seedl);
                /* ...encode seed value     */
            }
            else            /* ...else if not specified */
            {
                ENC_INTEGER (tpa_ptr, -12345);
                /* ...then default it       */
            }
        }

        else                /* ...Else must be RANDOM then  */
        {
            ENC_INTEGER (tpa_ptr, 0);   /* ...so leave LGR to sort seed out */
        }
loop:
        ;
    }

end:
    ;
    return;             /* ...finished so return    */
}

/*======================================================================*/
/*                                  */
/*    ENC_INPUT_VAR - Encode the INPUT variable command         */
/*    ==============                            */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_INPUT_VAR (argblk_ptr)               */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds an INPUT variable instruction to code buffer.   */
/*                                  */
/*    The format of an INPUT variable command is as follows:        */
/*                                  */
/*      OPCODE        VARIABLE_OPD  INPUT_OPTIONS  RESPONSE_OPD     */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_INPUT_VAR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("Encoding an INPUT variable Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the input opcode followed by the operand descriptor
     *    of the variable which is to be input, and then the
     *    response timer operand.
     */
    ENC_INTEGER (tpa_ptr, OPC_INPUT);   /* Encode the opcode        */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);   /* ...and encode the input OPD  */
    ENC_INTEGER (tpa_ptr, (long) (*tpa_ptr).opt_optl);
    /* ...code the options bitmask */

    /*
     *    If we have specified the 'WITH RESPONSE' option then encode the operand
     *    for it, otherwise encode a null operand.
     */
    if (((*tpa_ptr).cmnd_optl & CMND_OPT_RESPONSE) != 0)
    {                   /* If 'returning status' specified */
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
        /* ...encode it now then       */
    }

    else                /* ...otherwise...         */
    {
        ENC_NULL_OPD (tpa_ptr);     /* ...encode null opd in its place */
    }

    /*
     *    ...and return success
     */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_LOAD_ACCOUNT - Encode the LOAD ACCOUNT command        */
/*    =================                         */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_LOAD_ACCOUNT (argblk_ptr)                */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds a LOAD ACCOUNTING instruction to code buffer.   */
/*                                  */
/*    The format of a LOAD ACCOUNTING command is as follows:        */
/*                                  */
/*      OPCODE        VARIABLE_OPD                                  */
/*    unsigned long    struct OPD                               */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_LOAD_ACCOUNT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("Encoding the LOAD ACCOUNTING Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the load accounting opcode followed by the operand descriptor
     *    of the variable into which the accounting information is to be loaded
     */
    ENC_INTEGER (tpa_ptr, OPC_LOADACCNT); /* Encode the opcode      */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    /* ...and encode the OPD    */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_LOAD_DEVICE - Encode the LOAD DEVICE command          */
/*    ================                          */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_LOAD_DEVICE (argblk_ptr)             */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds a LOAD DEVICE instruction to code buffer.   */
/*                                  */
/*    The format of a LOAD DEVICE command is as follows:        */
/*                                  */
/*      OPCODE        VARIABLE_OPD                                  */
/*    unsigned long    struct OPD                               */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_LOAD_DEVICE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;


    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a LOAD DEVICE Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the load accounting opcode followed by the operand descriptor
     *    of the variable into which the accounting information is to be loaded
     */
    ENC_INTEGER (tpa_ptr, OPC_LOADDEV); /* Encode the opcode        */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    /* ...and encode the OPD    */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */

/*                                  */
/*    ENC_MATCH_VALUE - Encode the MATCH value command          */
/*    ===============                           */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_MATCH_VALUE (argblk_ptr)             */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*      (*tpa_ptr).open_fididxl - File ID of file to read   */
/*      (*tpa_ptr).onerror_labelc16 - Name of label for errors  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds a MATCH instruction to the object       */
/*    code buffer.                          */
/*                                  */
/*    the format of the MATCH instruction is as follows :-      */
/*  OPC_MATCH                           */
/*  MATCH_TYPE (UL)                         */
/*  VALUE_OPD                           */
/*  LOCATION_OPD                            */
/*  STATUS_OPD                          */
/*  OPTIONS_MASK (UL)                       */
/*  { OPTION1_OPD ... }                     */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_MATCH_VALUE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    unsigned long    MatchOptions = 0;  /* Options bitmask      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a MATCH VALUE Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the MATCH opcode.
     */
    ENC_INTEGER (tpa_ptr, OPC_MATCH);   /* ...and encode the opcode */

    /*
     *    ...followed by the type of the data being matched.
     */
    if (((*tpa_ptr).cmnd_optl & CMND_OPT_MATCH_CURRENCY) != 0)
        /* If 'Match Currency' specified */
    {
        ENC_INTEGER (tpa_ptr, (long) MATCH_CURRENCY);
        /* Encode MATCH-CURRENCY flag   */
    }

    /*
     *    ...and then the VP$SCREEN operand to define the area of the screen to
     *    scan for the value to match against
     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);

    /*
     *    ...and then the vale that we are to match against
     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);

    /*
     *    ...and then the status value into which we will write the result from 
     *    the match operation. 
     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);

    /*
     *    ...and then the options bitmask. Note that we need to translate the 
     *    internal command values to external option values!
     */
    if (((*tpa_ptr).cmnd_optl & CMND_OPT_MATCH_RETURNVALUE1) != 0)
        MatchOptions |= MATCH_RETURNVALUE1;

    if (((*tpa_ptr).cmnd_optl & CMND_OPT_MATCH_RETURNVALUE2) != 0)
        MatchOptions |= MATCH_RETURNVALUE2;

    ENC_INTEGER (tpa_ptr, MatchOptions);

    /*
     *    ...and then encode the options themselves.
     */
    if ((MatchOptions & MATCH_RETURNVALUE1) != 0)
    {
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    }

    /*
     *    ...and then encode the options themselves.
     */
    if ((MatchOptions & MATCH_RETURNVALUE2) != 0)
    {
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    }

    /*
     *    That's All Folks!!!
     */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}


/*======================================================================*/
/*                                  */
/*    ENC_NOTE - Encode the NOTE command                */
/*    =========                             */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_NOTE (argblk_ptr)                    */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds a NOTE instruction to the code buffer.      */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_NOTE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a NOTE Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    ...add a NOTE opcode to the code buffer.
     */
    ENC_INTEGER (tpa_ptr, OPC_NOTE);    /* ...and encode the opcode */

end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_NULL_OPD - Encode A Null Operand Header           */
/*    =============                         */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_NULL_OPD (tpa_ptr)                   */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   tpa_ptr - Pointer to the SCL_TPA data           */
/*    Outputs:  None                            */
/*                                  */
/*    This routine adds a NULL operand header to the object code buffer */
/*    to signify that this argument has not been specified.     */
/*                                  */
/*----------------------------------------------------------------------*/

static void     ENC_NULL_OPD    (struct scl_tpa *tpa_ptr)

{
    /*
     *    Local data
     */
    struct   oph      oph;          /* Null operand header      */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a NULL operand\n");
    }
#endif

    /*
     *    Initialize the operand to NULL
     */
    oph.addrl = ADDR_NONE;  /* Clear addressing mode.   */
    oph.dtypl = 0;      /* Clear data type      */
    oph.sizel = 0;      /* Clear data size      */
    oph.optl = 0;       /* Clear operand options    */

    /*
     *    ...and write it to the code buffer.
     */
    (void) memcpy ((*tpa_ptr).code_ptrpc, &oph, sizeof (struct oph));
    /* Copy operand header to code  */
    (*tpa_ptr).code_ptrpc += sizeof (struct oph);
    /* ...and update code pointer   */

    return;             /* ...and return */
}

/*======================================================================*/
/*                                  */

/*                                  */
/*    ENC_NUMERIC_OPD - Encode A Numeric Operand Header         */
/*    ===============                           */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_NUMERIC_OPD (tpa_ptr)                */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   tpa_ptr - Pointer to the SCL_TPA data           */
/*      number - UL number for integer value.           */
/*    Outputs:  None                            */
/*                                  */
/*    This routine adds an integer literal operand to the object code   */
/*    buffer with the value supplied.                               */
/*                                  */
/*----------------------------------------------------------------------*/

static void     ENC_NUMERIC_OPD (struct scl_tpa *tpa_ptr ,long number)

{
    /*
     *    Local data
     */
    struct   oph      oph;          /* Operand header       */
    struct   opd      opd;          /* Operand descriptor.      */

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a Numeric Operand value %ld\n",number);
    }
#endif

    /*
     *    Initialize the operand to NULL
     */
    oph.addrl = ADDR_IMMEDIATE;     /* Immediate addressing mode    */
    oph.dtypl = DTYP_INTEGER;       /* Integer data type        */
    oph.sizel = sizeof (long);      /* Size = (long) bytes      */
    opd.uopd.imi.vall = number;     /* Value as supplied to us  */
    oph.optl = 0;                       /* Clear options mask           */


    /*
     *    ...and write it to the code buffer.
     */
    (void) memcpy ((*tpa_ptr).code_ptrpc, &oph, sizeof (struct oph));
    /* Copy operand header to code  */
    (*tpa_ptr).code_ptrpc += sizeof (struct oph);
    /* ...and update code pointer   */

    (void) memcpy ((*tpa_ptr).code_ptrpc, &opd, sizeof (struct imi));
    /* ...add IMMEDIATE INTEGER OPD */
    (*tpa_ptr).code_ptrpc += sizeof (struct imi);
    /* ...and update code pointer   */

    return;             /* ...and return */
}
/*
 * 
 */

/*======================================================================*/
/*                                  */
/*                                  */
/*    ENC_OPERAND - Encode an operand descriptor.           */
/*    ============                          */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_OPERAND (argblk_ptr ,remove_entryn ,llb_pos_ptr  */
/*             ,ret_pos_pptr                */
/*                                  */
/*    Type: Encode routine                      */
/*    Status:   None                            */
/*    Pointers: Unchanged                       */
/*    Inputs:   tpa_ptr = Pointer to TPARSE data record         */
/*      llb_pos = Pointer to position first operand to encode   */
/*      remove_entry = Boolean to flag delete entry or not. */
/*                                  */
/*    Outputs:  New llb_pos returned in *ret_pos_pptr if not NULL   */
/*                                  */
/*    This routine encodes an operand descriptor.                       */
/*                                  */
/*    The operands to be encoded are held in the linked list of parsed  */
/*    operands. We start at the beginning of the list, deleting each    */
/*    entry as we go and encoding the operand into the script object    */
/*    code.                             */
/*                                  */
/*    This routine may be called recursively, in order that we may  */
/*    encode operands which require further operands to define      */
/*    themselves completely. For example arrays or VP$SCREEN references */
/*                                  */
/*    If the operand descriptor is for VP$OUTPUT then it is immediately */
/*    followed by a second operand descriptor for the modifier value.   */
/*                                  */
/*    If the operand descriptor is for VP$SCREEN then it is immediately */
/*    followed by the 4 operand descriptors for the modifier values.    */
/*                                  */
/*----------------------------------------------------------------------*/

static  void    ENC_OPERAND     (struct   scl_tpa *tpa_ptr
                                 ,         Boolean remove_entryn
                                 ,         char    *llb_pos_ptr
                                 ,     char    **ret_llb_pos_pptr)

{
    /*
     *    Local data
     */
    unsigned long   iret;           /* Status return variable.  */
    long   isub;           /* Loop control         */
    struct   oph    *oph_ptr;       /* Operand header pointer   */
    struct   opd    *opd_ptr;       /* Operand descriptor pointer   */
    struct   oph    *array_oph_ptr;     /* Array Operand header pointer */
    struct   symtbl *symtbl_ptr;        /* Pointer variable symbol table */
    unsigned long   opd_size;       /* Size of OPD section.     */
    char   *llb_ctx_ptr;       /* Linked list context      */
    char   mode;           /* Mode required.       */
    Boolean   array_end;      /* End of array expression  */
    struct   swstack    *sw_ptr;            /* Switch stack */

#if defined(UNIX) || defined(WIN32)
    unsigned long alignment_fault;      /* Alignment fault count (UNIX Only) */
#endif

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE("Encoding an OPERAND (pos = %ld)\n",llb_pos_ptr);
        PSL_OUTPUT_TRACE("Operand Count = %ld\n",GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Pick up the first OPH from the linked list remembering that we may be
     *    deleting it as we recover it.
     *
     *    The entry is actually a pointer to the OPH. We will be returned a pointer
     *    to the following item which we can use if this item is an array or
     *    requires further operands to clarify it.
     */
    if (remove_entryn)
    {

        /*
         *    Set up the mode given that if the position is NULL we shall use a
         *    mode of FIRST otherwise we shall use SPECIFIC
         */
        if (llb_pos_ptr == NULL)
        {
            mode = PSL_VM_POS_BEG;
        }
        else
        {
            mode = PSL_VM_POS_SPEC;
        }

        iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                              ,llb_pos_ptr ,&llb_ctx_ptr ,mode);
        /* Remove the OPH pointer first */
        if  (!(psl_test_status (iret))) /* ...report any errors     */
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"PSL_VM_REMOVE (in ENC_OPERAND Remove OPH)", iret);
        }

        llb_pos_ptr = llb_ctx_ptr;  /* update position to FOLLOWING item */
        iret = PSL_VM_REMOVE ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                              ,llb_pos_ptr ,&llb_ctx_ptr ,PSL_VM_POS_SPEC);
        /* ...then pick up the OPD ptr  */
        if  (!(psl_test_status (iret))) /* ...report any errors     */
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"PSL_VM_REMOVE (in ENC_OPERAND Remove OPD)", iret);
        }
    }
    else
    {

        /*
         *    Set up the mode given that if the position is NULL we shall use a
         *    mode of FIRST otherwise we shall use SPECIFIC
         */
        if (llb_pos_ptr == NULL)
        {
            mode = PSL_VM_MODE_FIRST;
        }
        else
        {
            mode = PSL_VM_MODE_SPEC;
        }

        iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &oph_ptr
                            ,llb_pos_ptr ,&llb_ctx_ptr ,mode);
        /* ...copy the OPH pointer  */
        if  (!(psl_test_status (iret))) /* ...report any errors     */
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"PSL_VM_COPY (in ENC_OPERAND Remove OPH)", iret);
        }

        llb_pos_ptr = llb_ctx_ptr;  /* ...and update the position   */
        iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps ,(char *) &opd_ptr
                            ,llb_pos_ptr ,&llb_ctx_ptr ,PSL_VM_MODE_NEXT);
        /* ...copy the OPD pointer  */
        if  (!(psl_test_status (iret))) /* ...report any errors     */
        {
            PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                         ,"PSL_VM_COPY (in ENC_OPERAND Remove OPD)", iret);
        }
    }

    llb_pos_ptr = llb_ctx_ptr;      /* ...and update the position   */

    /*
     *    ...and copy operand header to the object code buffer.
     */
    (void) memcpy ((*tpa_ptr).code_ptrpc, oph_ptr, sizeof (struct oph));
    /* Copy operand header to code  */
    (*tpa_ptr).code_ptrpc += sizeof (struct oph);
    /* ...and update code pointer   */

    /*
     *    We write the operand descriptor determined by the type of operand
     */
    if ((*oph_ptr).addrl == ADDR_TIMER) /* If a timer operand       */
    {
        opd_size = sizeof (struct tim); /* Size of TIMER OPD        */
    }

    else if ((*oph_ptr).addrl == ADDR_IMMEDIATE)
        /* If an immediate  operand */
    {
        if ((*oph_ptr).dtypl == DTYP_INTEGER)
            /* ...and an integer...     */
        {
            opd_size = sizeof (struct imi);
            /* Size of IMMEDIATE INTEGER    */
        }

        else
        {
            opd_size = (*oph_ptr).sizel; /* Size = size of character value */

            /*
             *    IF THIS IS A UNIX VERSION OF THE SOFTWARE WE MUST ENSURE THAT THE END OF
             *    THE CHARACTER STREAM IS LONGWORD ALIGNED. INCREMENT THE OPD_SIZE TO
             *    ACHIEVE THIS END
             */
#if defined(UNIX) || defined(WIN32)
            alignment_fault = (opd_size % sizeof (long));
            /* Get any "remainder"      */
            alignment_fault = ((sizeof(long) - alignment_fault)
                               % sizeof(long));
            /* Convert to "shortfall"   */
            opd_size += alignment_fault;
#endif
        }
    }

    else if (((*oph_ptr).addrl == ADDR_LOCAL)
             ||       ((*oph_ptr).addrl == ADDR_THREAD)
             ||       ((*oph_ptr).addrl == ADDR_GLOBAL)
             ||       ((*oph_ptr).addrl == ADDR_SCRIPT))
    {                   /* ...else if global/script/local */
        opd_size = sizeof (struct var); /* Size = size of VAR OPD   */
    }

    else if ((*oph_ptr).addrl == ADDR_OPERATOR)
    {                   /* If an operator operand   */
        opd_size = sizeof (struct opr); /* Size of OPERATOR OPD     */
    }

    else
    {
        opd_size = 0;           /* Operand requires no OPD  */
    }

    /*
     *    ...and write this much data to the code buffer.
     */
    (void) memcpy ((*tpa_ptr).code_ptrpc, opd_ptr, opd_size);
    /* Copy operand descriptor to code */
    (*tpa_ptr).code_ptrpc += opd_size;  /* ...and update code pointer   */

    /*
     *    Some operand descriptors are always followed by further information
     *    and/or operands. These are as follows:
     *
     *     VP$OUTPUT        : 1 integer modifier
     *     VP$SCREEN        : 4 Further operands
     *
     */
    if ((*oph_ptr).addrl == ADDR_VP_OUTPUT)
    {                   /* If VP$OUTPUT operand     */
        ENC_INTEGER (tpa_ptr, (long) (*tpa_ptr).dec_integerl);
        /* ...encode the modifier value */
    }

    else if ((*oph_ptr).addrl == ADDR_VP_SCREEN)
    {                   /* ...else if VP$SCREEN operand */

        /*
         *    VP$SCREEN may potentially be followed by upto 4 furtehr operands to
         *    define the screen co-ordinates to use. There are special rules to how we
         *    encode VP$SCREEN operands as some may not be specified.
         */
        if (((*tpa_ptr).cmnd_optl & CMND_OPT_VPS_LINE1) != 0)
        {               /* If line1 specified       */

            /*
             *    If LINE2 has NOT been specified then we need to encode LINE1 twice, 1 for
             *    LINE1 and 1 for LINE2.
             */
            if (((*tpa_ptr).cmnd_optl & CMND_OPT_VPS_LINE2) != 0)
            {               /* If line2 specified       */
                ENC_OPERAND (tpa_ptr, remove_entryn
                             ,llb_pos_ptr ,&llb_pos_ptr);
                /* ...then encode LINE1     */
                ENC_OPERAND (tpa_ptr, remove_entryn
                             ,llb_pos_ptr ,&llb_pos_ptr);
                /* ...then encode LINE2     */
            }
            else            /* ...otherwise...      */
            {
                ENC_OPERAND (tpa_ptr, PFALSE ,llb_pos_ptr ,NULL);
                /* ...encode LINE1 without deleting */
                ENC_OPERAND (tpa_ptr, remove_entryn
                             ,llb_pos_ptr ,&llb_pos_ptr);
                /* ...and LINE1 again as LINE2  */
            }
        }
        else                /* ...otherwise...      */
        {

            /*
             *    Line 1 has not been specified so encode a NULL operand in its place
             */
            ENC_NULL_OPD (tpa_ptr); /* ...encode null opd in its place */

            /*
             *    If LINE2 has been specified then encode it otherwise code a NULL instead.
             */
            if (((*tpa_ptr).cmnd_optl & CMND_OPT_VPS_LINE2) != 0)
            {               /* If line2 specified       */
                ENC_OPERAND (tpa_ptr, remove_entryn
                             ,llb_pos_ptr ,&llb_pos_ptr);
            }               /* ...then encode LINE2     */
            else            /* ...otherwise...      */
            {
                ENC_NULL_OPD (tpa_ptr); /* ...encode null operand   */
            }
        }

        /*
         *    Now deal with the columns.
         */
        if (((*tpa_ptr).cmnd_optl & CMND_OPT_VPS_COL1) != 0)
        {               /* If COL1 specified        */

            /*
             *    If COL2 has NOT been specified then we need to encode COL1 twice, 1 for
             *    COL1 and 1 for COL2.
             */
            if (((*tpa_ptr).cmnd_optl & CMND_OPT_VPS_COL2) != 0)
            {               /* If line2 specified       */
                ENC_OPERAND (tpa_ptr, remove_entryn
                             ,llb_pos_ptr ,&llb_pos_ptr);
                /* ...then encode COL1      */
                ENC_OPERAND (tpa_ptr, remove_entryn
                             ,llb_pos_ptr ,&llb_pos_ptr);
                /* ...then encode COL2      */
            }
            else            /* ...otherwise...      */
            {
                ENC_OPERAND (tpa_ptr, PFALSE ,llb_pos_ptr ,NULL);
                /* ...encode COL1 without deleting */
                ENC_OPERAND (tpa_ptr, remove_entryn
                             ,llb_pos_ptr ,&llb_pos_ptr);
                /* ...before restoring pos ptr  */

                /* ...and COL1 again as COL2    */
            }
        }
        else                /* ...otherwise...      */
        {

            /*
             *    Line 1 has not been specified so encode a NULL operand in its place
             */
            ENC_NULL_OPD (tpa_ptr); /* ...encode null opd in its place */

            /*
             *    If COL2 has been specified then encode it otherwise code a NULL instead.
             */
            if (((*tpa_ptr).cmnd_optl & CMND_OPT_VPS_COL2) != 0)
            {               /* If line2 specified       */
                ENC_OPERAND (tpa_ptr, remove_entryn
                             ,llb_pos_ptr ,&llb_pos_ptr);
            }               /* ...then encode COL2      */
            else            /* ...otherwise...      */
            {
                ENC_NULL_OPD (tpa_ptr); /* ...encode null operand   */
            }
        }
    }

    /*
     * 
     *
     *    Arrays really are special !!!
     *
     *    The OPERAND for the array itself is followed by:
     *
     *  DIMENSION 1 LOW VALUE
     *  DIMENSION 1 HIGH VALUE
     *  DIMENSION n LOW VALUE
     *  DIMENSION n HIGH VALUE
     *  ARRAY INDICES OPERANDS
     */
    if (((*oph_ptr).optl & VOPT_ARRAY) != 0)
    {                   /* If variable is an array  */

        /*
         *    ...encode the array boundary value(s).
         */
        symtbl_ptr = (*opd_ptr).uopd.var.symtbl_ptrps;
        /* Variable's symbol table entry */
        for (isub = 0; isub < (*opd_ptr).uopd.var.dimcntl; isub++)
        {
            ENC_INTEGER (tpa_ptr
                         ,(long) (*symtbl_ptr).var_array_bounds1l10[isub]);
            ENC_INTEGER (tpa_ptr
                         ,(long) (*symtbl_ptr).var_array_bounds2l10[isub]);
        }

        /*
         *    ...followed by the array indecis themselves. These consist of an
         *    expression for each array index terminated by a null operand.
         */
        for (isub = 0; isub < (*opd_ptr).uopd.var.dimcntl; isub++)
        {
            array_end = PFALSE;     /* ...flag not end of expression */
            do
            {
                /*
                 *      The following section of code is very inelegant. It
                 *      is here to nudge the operand ll forwards by one entry, as when 
                 *      a CASE instruction is being encoded and the switch condition is 
                 *      an array variable, the existing code doesn't work. 
                 *      It fails because the list still points to the previous entry when 
                 *      we get to this point (which happens to be the OPD).
                 *      However, it works fine for everything else (such as SET) but not
                 *      at all for CASE statements.  It would seem fairly safe to only
                 *      advance the list pointer if this is a CASE statement.  Paoula...
                 */
                if (tpa_ptr->sw_depthl > 0)/* If a switch statement active */
                {
                    sw_ptr = tpa_ptr->swstack_begps + tpa_ptr->sw_depthl - 1;
                    if (tpa_ptr->parsed_operand_llps == sw_ptr->sw_operand_llps)
                    {           /* If CASE statement handling on */
                        iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps, NULL,
                                            llb_pos_ptr ,&llb_ctx_ptr ,PSL_VM_MODE_NEXT);
                        llb_pos_ptr = llb_ctx_ptr;
                        /* Advance on entry in the list */
                    }
                }

                iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                                    ,(char *) &array_oph_ptr
                                    ,llb_pos_ptr ,NULL ,PSL_VM_MODE_SPEC);
                /* ...copy the OPH pointer  */
                if ((*array_oph_ptr).addrl == ADDR_NONE) array_end = PTRUE;
                /* ...flag end of expression    */
                ENC_OPERAND (tpa_ptr, remove_entryn
                             ,llb_pos_ptr ,&llb_pos_ptr);
                /* Encode array index 1     */
            }
            while (!array_end);
        }
    }

    /*
     * 
     *
     *    If we have REMOVED the operand from the linked list then we should ensure
     *    that we deallocate the operand header and descriptor allocated when we
     *    parsed the operand. If we simply copied the operand then we should NOT
     *    deallocate the data structures as they are still in use!!!
     */
    if (remove_entryn)
    {
        FREE (oph_ptr);

#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER5))
        {
            PSL_OUTPUT_TRACE ("FREE: oph - '%p' - ENC_OPERAND\n", oph_ptr);
        }
#endif  /* PSL_TRACING */

        FREE (opd_ptr);

#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER5))
        {
            PSL_OUTPUT_TRACE ("FREE: opd - '%p' - ENC_OPERAND\n", opd_ptr);
        }
#endif  /* PSL_TRACING */

    }

    /*
     *    ...and then return.
     *    If we have specified a return position pointer then return the position
     *    of the item which follows this one.
     */
end:
    ;
    if (ret_llb_pos_pptr != NULL)
    {
        *ret_llb_pos_pptr = llb_pos_ptr;
    }
    return;             /* ...and return        */
}

/*======================================================================*/
/*                                  */
/*    ENC_OPERATOR - Encode an operator.                */
/*    =============                         */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_OPERATOR (argblk_ptr)                */
/*    Type: Encode routine                      */
/*    Status:   None                            */
/*    Pointers: Unchanged                       */
/*    Inputs:   OPERATOR - long operand code                */
/*      tpa_ptr - pointer to tparse data            */
/*    Outputs:  None                            */
/*                                  */
/*    This routine encodes an operator. This is prefixed by an opcode   */
/*    to identify this an an operator.                  */
/*                                  */
/*----------------------------------------------------------------------*/

void    ENC_OPERATOR (struct scl_tpa *tpa_ptr, long operator)

{

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("Encoding an OPERATOR\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the opcode for an operator.
     */
    ENC_INTEGER (tpa_ptr, OPC_OPERATOR); /* Encode the operator opcode  */
    ENC_INTEGER (tpa_ptr, operator);    /* Followed by the operator */

end:
    ;
    return;             /* ...and return        */
}

/*======================================================================*/
/*                                  */
/*    ENC_PUSHIF_VPS - Finish off a VP$SCREEN simple condition      */
/*    ===============                           */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_PUSHIF_VPS (argblk_ptr)              */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*      ATTR_OPTL = Attributes bitmask              */
/*                                  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine completes the encoding of a simple condition if the  */
/*    first operand in the condition is VP$SCREEN           */
/*                                  */
/*    There are two formats for this routine depending on the operator. */
/*    For the MATCHES and NOT_MATCHES operators the following is coded: */
/*                                  */
/*  PUSHIF   (Screen name operand)                  */
/*  PUSHIF   (File Name operand)                    */
/*  VPS_OP   MATCHES/NOT_MATCHES                    */
/*       options_bitmask                    */
/*                                  */
/*    For all other operators the following is encoded          */
/*  PUSHIF   (Second operand)                   */
/*  VPS_OP   operator                               */
/*       options_bitmask                    */
/*                                  */
/*    The special opcode VPS_OP is used so that we can differentiate    */
/*    between the format of a standard condition expresssion and a  */
/*    VP$SCREEN one.                            */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_PUSHIF_VPS  (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    unsigned long    *if_ptr;       /* If command pointer       */
    unsigned long    *cond_ptr;     /* Pointer to IF stack */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a PUSHIF VP$SCREEN Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    cond_ptr = (*tpa_ptr).cond_stackl10 + (*tpa_ptr).cond_levell;
    /* Point to the current entry */

    /*
     *    Firstly encode the PUSHIF opcode
     */
    ENC_INTEGER (tpa_ptr, OPC_PUSHIF);

    /*
     *    Check to see if the operator is MATCHES or NOT_MATCHES. If it is
     *    then we will need to encode TWO operands. One for the screen name and
     *    a second for the file name. Note that the screen name may not have been
     *    specified in which case we will encode a NULL operand.
     */
    if ((*cond_ptr == OPR_MATCHES)
            ||  (*cond_ptr == OPR_NOT_MATCHES))
    {                   /* If MATCHES or NOT_MATCHES    */
        if (((*tpa_ptr).cmnd_optl & CMND_OPT_SCREEN_NAME) != 0)
        {               /* If 'Screen Name' specified   */
            ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
        }               /* ...encode it now then       */
        else                /* ...otherwise...         */
        {
            ENC_NULL_OPD (tpa_ptr); /* ...encode null opd in its place */
        }

        /*
         *    ...followed by the file name operand which will always have been specified
         *    if the command was valid.
         */
        ENC_INTEGER (tpa_ptr, OPC_PUSHIF);
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    }
    else
    {

        /*
         *    All other operators are followed by a single operand so encode it now.
         */
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    }

    /*
     *    ...this is followed by the operator. We use a special Opcode for this
     *    operator because it is used with VP$SCREEN which has a different format
     *    to that used for normal conditional expression evaluations.
     */
    ENC_INTEGER (tpa_ptr, OPC_VPS_OPERATOR);
    /* ...code the VP$SCREEN operator */
    ENC_INTEGER (tpa_ptr, (long) *cond_ptr);
    /* ...followed by the operator */

    /*
     *    ...and code the attributes bitmask.
     */
    ENC_INTEGER (tpa_ptr, (long) (*tpa_ptr).attr_optl);
    /* ...code the attributes bitmask */

    /*
     *    We must ENSURE that we CLEAR the operator on the current level as we will
     *    have problems later if we do not !!!
     */
    if_ptr = (*tpa_ptr).cond_stackl10 + (*tpa_ptr).cond_levell;
    /* Point to current level   */
    *if_ptr = OPR_NONE;         /* ...and remove operator   */

end:
    ;
    return PSL_NORMAL;
}

/*======================================================================*/
/*                                  */
/*    ENC_REC_FORMAT - Add initialization for record defn. array        */
/*    ==============                            */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_REC_FORMAT (argblk_ptr)              */
/*                                  */
/*    Type: Service routine                     */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   com_ptr - Pointer to the SCL common data record     */
/*      record_count - number of records in script      */
/*    Implicit Inputs:                          */
/*          VARIABLES_TABLE = Variables Symbol Table        */
/*    Outputs:  OBJ_CODE_PTR = Object code pointer updated      */
/*                                  */
/*    This routine is called from the script initilization routine to   */
/*    add sub-commands to the object code buffer to handle the initial- */
/*    ization of record definition arrays               */
/*                                  */
/*                                  */
/*======================================================================*/

void    ENC_REC_FORMAT  (struct   scl_com *com_ptr
                         ,unsigned long   rec_count)

{
    /*
     *    Local data
     */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *symtbl_ptr;       /* Pointer variable symbol table */
    unsigned long    isub;          /* Loop control variable    */
    unsigned long    loop_array;        /* Loop array control variable  */
    unsigned long    loop_limit;        /* Loop limit control variable  */
    long    *save_array;       /* Save array pointer       */
    /*
     *    Copy pointers for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps;
    symtbl_ptr = (*tpa_ptr).variable_begps;
    /* Variables symbol table   */

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a record def. array init. opcode\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Scan the variables symbol table looking for a variable of the correct
     *    data type (DTYP_RECORD).
     */
    for (isub=0; isub < (*tpa_ptr).var_cntl; isub++, symtbl_ptr++)
    {
        if ((*symtbl_ptr).var_typel != DTYP_RECORD) goto loop;
        /* Skip this if not correct datatype*/

        /*
         *    If it is a record:
         *    Encode INI_REC_ARRAY operand
         */
        ENC_INTEGER (tpa_ptr, INI_REC_ARRAY);

        /*
         *    Encode records index and number of items in the record
         */

        ENC_INTEGER (tpa_ptr,(*symtbl_ptr).var_indexl);

        /*
         * Check for any errors - record will have index of zero
         * if there were problems defining it.
         */
        if ((*symtbl_ptr).var_indexl <= 0 ) goto end;

        ENC_INTEGER (tpa_ptr,(*symtbl_ptr).var_rec_iteml);
        /*
         * Save the array:
         * Number of items in array is :
         *
         *  number of record items +1 for NULL terminator
         */
        loop_limit = (*symtbl_ptr).var_rec_iteml + 1;

        /*
         * Save start address of the array
         */

        save_array = (*symtbl_ptr).var_start_rec_arraypl;

        /*
         * Encode the array
         */

        for (loop_array = 0; loop_array < loop_limit; loop_array++, save_array++)
        {
            ENC_INTEGER (tpa_ptr ,*save_array);
        }


#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER5))
        {
            PSL_OUTPUT_TRACE ("FREE: recarray - '%p' - ENC_REC_FORMAT\n",
                              (*symtbl_ptr).var_start_rec_arraypl);
        }
#endif  /* PSL_TRACING */

loop:
        ;
    }

end:
    ;
    return;             /* ...finished so return    */
}

/*======================================================================*/
/*                                          */
/*    ENC_RESUME_REC - Encode a RESUME RECORDING command        */
/*    ===============                           */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_RESUME_REC (argblk_ptr)              */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*      ONERROR_LABELC16                    */
/*    Outputs:  None                            */
/*                                  */
/*    This routine adds a RESUME RECORDING instruction to the object    */
/*    code buffer.                          */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_RESUMEREC ONERROR_LABEL                 */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_RESUME_REC (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a RESUME RECORD Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the RESUMEREC opcode
     */
    ENC_INTEGER (tpa_ptr, OPC_RESUMEREC);

    /*
     *    ...then encode the ON ERROR clause operand. This is in effect the offset
     *    of the label in the code or -1 if the label has not been specified.
     */
    if (strlen ((*tpa_ptr).onerror_labelc16) == 0)
        /* If no error label        */
    {
        ENC_INTEGER (tpa_ptr, NULL_DISP);
        /* Encode null displacement */
    }

    else                /* ...else if label specified   */
    {
        /* ...then resolve the label    */
    }

    /*
     *    ...finally return success.
     */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}


/*======================================================================*/

/*    ENC_RESUME_SCR - Encode a RESUME SCREEN MODE command      */
/*    ===============                           */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_RESUME_SCR (argblk_ptr)              */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*      ONERROR_LABELC16                    */
/*    Outputs:  None                            */
/*                                  */
/*    This routine adds a RESUME SCREEN MODE instruction to the object  */
/*    code buffer.                          */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_RESUME_SCR ONERROR_LABEL                    */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_RESUME_SCR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a RESUME SCREEN MODE Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the RESUME_SCR opcode
     */
    ENC_INTEGER (tpa_ptr, OPC_RESUME_SCR);

    /*
     *    ...then encode the ON ERROR clause operand. This is in effect the offset
     *    of the label in the code or -1 if the label has not been specified.
     */
    if (strlen ((*tpa_ptr).onerror_labelc16) == 0)
        /* If no error label        */
    {
        ENC_INTEGER (tpa_ptr, NULL_DISP);
        /* Encode null displacement */
    }

    else                /* ...else if label specified   */
    {
        /* ...then resolve the label    */
    }

    /*
     *    ...finally return success.
     */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}


/*======================================================================*/
/*                                      */

/*    ENC_UPDATE_SCR - Encode the UPDATE SCREEN MODE command        */
/*    ==============                            */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_UPDATE_SCR (argblk_ptr)              */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds a UPDATE SCREEN MODE instruction to code buffer */
/*                                  */
/*    The format of a UPDATE SCREEN MODE command is as follows:     */
/*                                  */
/*      OPCODE        VARIABLE_OPD                                  */
/*    unsigned long    struct OPD                               */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_UPDATE_SCR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a UPDATE SCREEN MODE Command\n");
    }
#endif

    /*
     *    Encode the update screen mode opcode
     */
    ENC_INTEGER (tpa_ptr, OPC_UPDATE_SCR); /* Encode the opcode     */
end:
    ;
    /*
     *    ...then encode the ON ERROR clause operand. This is in effect the offset
     *    of the label in the code or -1 if the label has not been specified.
     */
    if (strlen ((*tpa_ptr).onerror_labelc16) == 0)
        /* If no error label        */
    {
        ENC_INTEGER (tpa_ptr, NULL_DISP);
        /* Encode null displacement */
    }

    else                /* ...else if label specified   */
    {
        /* ...then resolve the label    */
    }
    return PSL_NORMAL;          /* ...and return success    */
}
/*
 * 
 */

/*======================================================================*/
/*                                  */
/*    ENC_SNAP_VP$SCREEN - Encode a SNAP Operand            */
/*    ===================                       */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_SNAP_VP$SCREEN (argblk_ptr)              */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*      snap_maskl - SNAP options mask.             */
/*      onerror_label - ON ERROR label address          */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds a SNAP command to the code buffer.      */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_SNAP SNAP_OPD FILE_OPD SCR_OPD  DSC_OPD  SNAP_MASK  LAB_OFF */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_SNAP_VP_SCREEN (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a SNAP Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Add the SNAP opcode to the object code buffer
     */
    ENC_INTEGER (tpa_ptr, OPC_SNAP);    /* ...and encode the opcode */

    /*
     *    ...followed by the SNAP variables operand, File name operand, screen
     *    name operand (if specified), description operand (if specified)
     *    and ONERROR label offset.
     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    /* ...VP$SCREEN operand     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    /* ...FILE-NAME operand     */

    /*
     *    Either or both of the screen name and description operands may have been
     *    omitted. Check the options bitmask and encode a null operand for any
     *    which have been omitted.
     */
    if (((*tpa_ptr).cmnd_optl & CMND_OPT_SNAP_SCR) != 0)
    {                   /* If 'Screen' specified    */
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
        /* ...encode it now then    */
    }

    else                /* ...otherwise...      */
    {
        ENC_NULL_OPD (tpa_ptr);     /* ...encode null opd in its place */
    }

    if (((*tpa_ptr).cmnd_optl & CMND_OPT_SNAP_DSC) != 0)
    {                   /* If 'Description' specified   */
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
        /* ...encode it now then    */
    }

    else                /* ...otherwise...      */
    {
        ENC_NULL_OPD (tpa_ptr);     /* ...encode null opd in its place */
    }

    /*
     *    ...followed by the SNAP options mask.
     */
    ENC_INTEGER (tpa_ptr, (long) (*tpa_ptr).snap_maskl);

    /*
     *    ...then encode the ON ERROR clause operand. This is in effect the offset
     */
    if (strlen ((*tpa_ptr).onerror_labelc16) == 0)
        /* If no error label        */
    {
        ENC_INTEGER (tpa_ptr, NULL_DISP);
        /* Encode null displacement */
    }

    else                /* ...else if label specified   */
    {
        /* ...then resolve the label    */
    }

end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_SUSPEND_REC - Encode a SUSPEND RECORDING command      */
/*    ================                          */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_SUSPEND_REC (argblk_ptr)             */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*      ONERROR_LABELC16                    */
/*    Outputs:  None                            */
/*                                  */
/*    This routine adds a SUSPEND RECORDING instruction to the object   */
/*    code buffer.                          */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_SUSPENDREC ONERROR_LABEL                    */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_SUSPEND_REC (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a SUSPEND RECORDING Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the SUSPENDREC opcode
     */
    ENC_INTEGER (tpa_ptr, OPC_SUSPENDREC);

    /*
     *    ...then encode the ON ERROR clause operand. This is in effect the offset
     *    of the label in the code or -1 if the label has not been specified.
     */
    if (strlen ((*tpa_ptr).onerror_labelc16) == 0)
        /* If no error label        */
    {
        ENC_INTEGER (tpa_ptr, NULL_DISP);
        /* Encode null displacement */
    }

    else                /* ...else if label specified   */
    {
        /* ...then resolve the label    */
    }

    /*
     *    ...finally return success.
     */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}
/*
 * 
 */

/*======================================================================*/

/*    ENC_SUSPEND_SCRNMODE - Encode a SUSPEND SCREEN MODE command   */
/*    ====================                          */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_SUSPEND_SCRNMODE (argblk_ptr)                */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*      ONERROR_LABELC16                    */
/*    Outputs:  None                            */
/*                                  */
/*    This routine adds a SUSPEND SCREEN MODE instruction to the object */
/*    code buffer.                          */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_SUSPENDSCRMD ONERROR_LABEL                  */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_SUSPEND_SCRNMODE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a SUSPEND SCREEN MODE Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Encode the SUSPENDSCRMD opcode
     */
    ENC_INTEGER (tpa_ptr, OPC_SUSPEND_SCR);

    /*
     *    ...then encode the ON ERROR clause operand. This is in effect the offset
     *    of the label in the code or -1 if the label has not been specified.
     */
    if (strlen ((*tpa_ptr).onerror_labelc16) == 0)
        /* If no error label        */
    {
        ENC_INTEGER (tpa_ptr, NULL_DISP);
        /* Encode null displacement */
    }

    else                /* ...else if label specified   */
    {
        /* ...then resolve the label    */
    }

    /*
     *    ...finally return success.
     */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}
/*
 * 
 */

/*======================================================================*/
/*                                  */
/*    ENC_WAIT_TIMEOUT - Encode the WAIT n FOR TIMEOUT command      */
/*    =================                         */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_WAIT_TIMEOUT (argblk_ptr)                */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds a WAIT n FOR TIMEOUT instruction to the code    */
/*    buffer.                               */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_WAITTIM  PERIOD_OPD  ONTIMEOUT_OFF              */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_WAIT_TIMEOUT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a WAIT FOR TIMEOUT Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    This command consists solely of the opcode and the wait interval operand
     *    followed by the ontimeout label offset - so encode it.
     */
    ENC_INTEGER (tpa_ptr, OPC_WAITTIM);/* ...and encode the opcode  */

    /*
     *    ...followed by the WAIT period. this may or may not have been specified.
     */
    if (((*tpa_ptr).cmnd_optl & CMND_OPT_WAIT_PRD) != 0)
    {                   /* If 'Wait period' specified   */
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
        /* ...encode it now then    */
    }

    else                /* ...otherwise...      */
    {
        ENC_NULL_OPD (tpa_ptr);     /* ...encode null opd in its place */
    }

    /*
     *    ...then encode the ON TIMEOUT clause operand. This is in effect the offset
     *    of the label in the code or -1 if the label has not been specified.
     */
    if (strlen ((*tpa_ptr).ontimeout_labelc20) == 0)
        /* If no error label        */
    {
        ENC_INTEGER (tpa_ptr, NULL_DISP);
        /* Encode null displacement */
    }

    else                /* ...else if label specified   */
    {
        /* ...then resolve the label    */
    }
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_WSTR_END - Complete encoding of a WAIT n FOR STRING command   */
/*    =============                         */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_WSTR_END (argblk_ptr)                */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*          ONERROR_LABEL = ONERROR clause label            */
/*      (*tpa_ptr).onerror_Addrpc - Address of space for ERROR  */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine completes the encoding of a WAIT FOR STRING command  */
/*                                  */
/*    It terminates the variable list with a null operand, encodes the  */
/*    operand descriptor for the variable in which to search for the    */
/*    strings and adds the ONERROR label offset to the buffer.      */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_WAITSTR  PERIOD_OPD  OPD ... NOPD  IN_OPD  ONTIMEOUT_LABEL  */
/*                                  */
/*    where OPD ... NOPD is a parameter list terminated by a null       */
/*    operand, IN_OPD is the variable to search for the strings.    */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_WSTR_END (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   oph     *oph_ptr;      /* OPH pointer          */
    struct   opd     *opd_ptr;      /* OPD pointer          */
    char    *sav_code_ptr;     /* Saved code pointer       */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("Encoding the end of a WAIT FOR STRING Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    Terminate the list of the operands to search for with a null operand.
     */
    ENC_NULL_OPD (tpa_ptr);     /* Terminate the operand list   */

    /*
     *    ...then encode the IN buffer operand - If this has NOT been specified
     *    then we shall default in to all of VP$OUTPUT
     */

    if (((*tpa_ptr).cmnd_optl & CMND_OPT_WAIT_INOPD) == 0)
    {                   /* If 'Wait in' operand NOT specified */

        /*
         *    Default WAIT IN operand to be VP$OUTPUT
         */
        ALLOCATE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr);
        /* Allocate an operand on LL    */
        (*oph_ptr).addrl = ADDR_VP_OUTPUT; /* Set VP$OUTPUT addressing mode */
        (*oph_ptr).dtypl = DTYP_CHARACTER; /* Character data type   */
        (*oph_ptr).sizel = 512;     /* 512 data size        */
        (*tpa_ptr).dec_integerl = 512;  /* Set modifier size = 512  */
    }

    /*
     *    ...and encode the operand.
     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    /* ...encode it now then    */

    /*
     *    ...then encode the ON TIMEOUT clause operand. This is in effect the offset
     *    of the label in the code or -1 if the label has not been specified.
     *
     *    We allocated space for it earlier on in the processing of the command
     *    and it is in this space that we must encode the offset.
     */
    if (strlen ((*tpa_ptr).ontimeout_labelc20) != 0)
        /* If timeout label     */
    {
        sav_code_ptr = (*tpa_ptr).code_ptrpc;
        /* Save code pointer        */
        (*tpa_ptr).code_ptrpc = (*tpa_ptr).onerror_addrpc;
        /* Set pointer = ONERROR offset */
        /* ...then resolve the label    */
        (*tpa_ptr).code_ptrpc = sav_code_ptr;
        /* Restore code pointer     */
    }

    /*
     *    ...finally return success.
     */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_WSTR_OPD - Encode a WAIT n FOR STRING operand         */
/*    =============                         */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_WSTR_OPD (argblk_ptr)                */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*      OPD2 = Wait string operand              */
/*    Outputs:  Code pointer updated                    */
/*                                  */
/*    This routine adds an operand to the object code buffer. This      */
/*    operand defines a string to wait for.                             */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_WAITSTR  PERIOD_OPD  OPD ... NOPD  IN_OPD  ONTIMEOUT_LABEL  */
/*                                  */
/*    where OPD ... NOPD is a parameter list terminated by a null       */
/*    operand, IN_OPD is the variable to search for the strings.    */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_WSTR_OPD (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("\nEncoding a WAIT FOR STRING operand\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    The operand is in OPD2 - Encode it and then clear the operand descriptor
     *    sdo that it may be re-used.
     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
    /* Encode the operand       */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}

/*======================================================================*/
/*                                  */
/*    ENC_WSTR_START - Encode the start of a WAIT n FOR STRING command  */
/*    ===============                           */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_WSTR_START (argblk_ptr)              */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*      (*tpa_ptr).onerror_addrpc - Address of space for ON */
/*          TIMEOUT label                   */
/*                                  */
/*    This routine starts adding a WAIT n FOR STRING instruction to the */
/*    code buffer. This routine adds the opcode and wait period to      */
/*    the buffer so that the WAIT parameter list can be encoded     */
/*    directly into the code buffer as the operands are parsed.     */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_WAITSTR  PERIOD_OPD   ONTIMEOUT_LABEL  OPD ... NOPD  IN_OPD */
/*                                  */
/*    where OPD ... NOPD is a parameter list terminated by a null       */
/*    operand, IN_OPD is the variable to search for the strings.    */
/*                                  */
/*    This routine must allocate space for the ON TIMEOUT label offset  */
/*    which will then be filled in later.               */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_WSTR_START (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("Encoding the start of a WAIT FOR STRING Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    This command starts with the opcode and the wait interval operand
     *    - so encode it.
     */
    ENC_INTEGER (tpa_ptr, OPC_WAITSTR);/* ...and encode the opcode  */

    /*
     *    If we have specified the WAIT period then encode it otherwise encode
     *    a null period.
     */

    if (((*tpa_ptr).cmnd_optl & CMND_OPT_WAIT_PRD) != 0)
    {                   /* If 'Wait period' specified   */
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
        /* ...encode it now then    */
    }

    else                /* ...otherwise...      */
    {
        ENC_NULL_OPD (tpa_ptr);     /* ...encode null opd in its place */
    }

    /*
     *    Allocate space for the ON TIMEOUT label offset as NULL. This will
     *    be filled in later if we find an ON TIMEOUT label.
     */
    (*tpa_ptr).onerror_addrpc = (*tpa_ptr).code_ptrpc;
    /* Save current code pointer */
    ENC_INTEGER (tpa_ptr, NULL_DISP);   /* Encode this as ON TIMEOUT offset */
end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}


/*======================================================================*/
/*                                  */
/*    ENC_WAIT_OUT - Encode the start of a WAIT n FOR STRING command    */
/*    ===============                           */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_WSTR_START (argblk_ptr)              */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*      (*tpa_ptr).onerror_addrpc - Address of space for ON */
/*          TIMEOUT label                   */
/*                                  */
/*    This routine starts adding a WAIT n FOR OUTPUT instruction to the */
/*    code buffer. This routine adds the opcode and wait period to      */
/*    the buffer.                               */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_WAITOUT  PERIOD_OPD   ONTIMEOUT_LABEL               */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_WAIT_OUT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("Encoding the start of a WAIT FOR OUTPUT Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    This command starts with the opcode and the wait interval operand
     *    - so encode it.
     */
    ENC_INTEGER (tpa_ptr, OPC_WAITOUT);/* ...and encode the opcode  */

    /*
     *    If we have specified the WAIT period then encode it otherwise encode
     *    a null period.
     */

    if (((*tpa_ptr).cmnd_optl & CMND_OPT_WAIT_PRD) != 0)
    {                   /* If 'Wait period' specified   */
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
        /* ...encode it now then    */
    }
    else                /* ...otherwise...      */
    {
        ENC_NULL_OPD (tpa_ptr);     /* ...encode null opd in its place */
    }

    /*
     *    ...then encode the ON TIMEOUT clause operand. This is in effect the offset
     *    of the label in the code or -1 if the label has not been specified.
     */
    if (strlen ((*tpa_ptr).ontimeout_labelc20) == 0)
        /* If no timeout label      */
    {
        ENC_INTEGER (tpa_ptr, NULL_DISP);
        /* Encode null displacement */
    }

    else                /* ...else if label specified   */
    {
        /* ...then resolve the label    */
    }

end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}



/*======================================================================*/
/*                                  */

/*                                  */
/*    ENC_WAIT_CURSOR - Encode a WAIT FOR CURSOR command        */
/*    ===============                           */
/*                                  */
/*    Calling Format:                           */
/*  iret = ENC_WAIT_CURSOR (argblk_ptr)             */
/*                                  */
/*    Type: TPARSE action routine                   */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   argblk_ptr - Pointer to the tparse argument block.  */
/*    Outputs:  Code pointer updated                    */
/*      (*tpa_ptr).onerror_addrpc - Address of space for ON */
/*          TIMEOUT label                   */
/*                                  */
/*    This routine encodes a WAIT FOR CURSOR instruction to the     */
/*    code buffer.                              */
/*                                  */
/*    The format of this instruction is as follows:         */
/*  OPC_WAITPOS  PERIOD_OPD  VP$SCREEN  ONTIMEOUT_LABEL     */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   ENC_WAIT_CURSOR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Do not encode anything if we have had an error in the command.
     */
    if ((*tpa_ptr).error_in_commandn) goto end;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER1))
    {
        PSL_OUTPUT_TRACE ("Encoding the start of a WAIT FOR CURSOR Command\n");
        PSL_OUTPUT_TRACE ("Operand Count = %ld\n" ,GET_OPD_COUNT (tpa_ptr));
    }
#endif

    /*
     *    This command starts with the opcode and the wait interval operand
     *    - so encode it.
     */
    ENC_INTEGER (tpa_ptr, OPC_WAITPOS);/* ...and encode the opcode  */

    /*
     *    If we have specified the WAIT period then encode it otherwise encode
     *    a null period.
     */

    if (((*tpa_ptr).cmnd_optl & CMND_OPT_WAIT_PRD) != 0)
    {                   /* If 'Wait period' specified   */
        ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);
        /* ...encode it now then    */
    }
    else                /* ...otherwise...      */
    {
        ENC_NULL_OPD (tpa_ptr);     /* ...encode null opd in its place */
    }

    /*
     *    ...then encode the VP$SCREEN operand and its supporting operands.
     */
    ENC_OPERAND (tpa_ptr ,PTRUE ,NULL ,NULL);

    /*
     *    ...then encode the ON TIMEOUT clause operand. This is in effect the offset
     *    of the label in the code or -1 if the label has not been specified.
     */
    if (strlen ((*tpa_ptr).ontimeout_labelc20) == 0)
        /* If no timeout label      */
    {
        ENC_INTEGER (tpa_ptr, NULL_DISP);
        /* Encode null displacement */
    }

    else                /* ...else if label specified   */
    {
        /* Resolve the label    */
    }

end:
    ;
    return PSL_NORMAL;          /* ...and return success    */
}


/*======================================================================*/
/*                                  */

/*                                  */
/*    FORWARD_ONE_OPERAND - Skip LL past next 1 OPH/OPD pair        */
/*    ===================                       */
/*                                  */
/*    Calling Format:                           */
/*  FORWARD_ONE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr ,&llb_pos) */
/*                                  */
/*    Where:                                */
/*  tpa_ptr = pointer to the Tparse data                */
/*  oph_ptr = Address of a pointer to the OPH (returned)        */
/*  opd_ptr = Address of a pointer to the OPD (returned)        */
/*  llb_ptr = Address of a pointer to char LL position (updated)    */
/*                                  */
/*    This routine scans forward through the linked list of parsed  */
/*    operands to skip over the next OPH/OPD pair returning a pointer   */
/*    to the OPH read and the updating the position pointer to point    */
/*    to the following OPH.                     */
/*                                  */
/*    It must be noted that if the operand we locate is itself an array */
/*    then we will skip over the array elements as well.        */
/*                                  */
/*----------------------------------------------------------------------*/

static void FORWARD_ONE_OPERAND     (struct scl_tpa *tpa_ptr
                                     ,struct oph     **oph_pptr
                                     ,struct opd     **opd_pptr
                                     ,       char    **llb_pos_pptr)

{
    /*
     *    Local data
     */
    struct   oph   *oph_ptr;        /* Work pointer to CURRENT OPH  */
    struct   opd   *opd_ptr;        /* Work poinetr to OPD      */
    unsigned long  iret;            /* Status return        */
    unsigned long  isub;            /* Loop control variable.   */
    unsigned long  array_dimensions;    /* Dimensions of an array   */
    char  *llb_pos;        /* Linked list position     */
    char  *llb_ctx;        /* Linked list context      */
    char  mode;            /* Mode for copy        */
    Boolean  array_expression_end;    /* End of expression flag   */

    llb_pos = *llb_pos_pptr;        /* Set to required position     */

    /*
     *    Set up the mode given that if the position is NULL we shall use a
     *    mode of FIRST otherwise we shall use SPECIFIC
     */
    if (llb_pos == NULL)
    {
        mode = PSL_VM_MODE_FIRST;
    }
    else
    {
        mode = PSL_VM_MODE_SPEC;
    }

    /*
     *    We read the current OPH and OPD and return pointers to them both.
     *    This will however only give us the context for this item whereas we
     *    require the context of the following OPH. We will therefore need to read
     *    the following OPH simply to pick up the context.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                        ,(char *) &oph_ptr ,llb_pos ,&llb_ctx ,mode);
    /* Read the OPH at CURRENT position */
    if  (!(psl_test_status (iret))) /* ...signalling any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY (in FORWARD_ONE_OPERAND)", iret);
    }

    llb_pos = llb_ctx;          /* ...and update position   */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                        ,(char *) &opd_ptr ,llb_pos ,&llb_ctx ,PSL_VM_MODE_NEXT);
    /* ...and pick up OPD pointer   */
    if  (!(psl_test_status (iret))) /* ...signalling any errors.    */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY (in FORWARD_ONE_OPERAND)", iret);
    }

    llb_pos = llb_ctx;          /* ...and update position   */

    /*
     *    ...and read the next OPH record simply so that we can recover its LL
     *    pointer.
     */
    iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                        ,NULL ,llb_pos ,&llb_ctx ,PSL_VM_MODE_NEXT);
    /* Read the OPH at CURRENT position */
    if  (!(psl_test_status (iret))) /* ...signalling any errors     */
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR
                     ,"PSL_VM_COPY (in FORWARD_ONE_OPERAND)", iret);
    }
    llb_pos = llb_ctx;          /* ...and update position   */

    /*
     *    If the operand was an array then we must skip forward past the 'n'
     *    expressions for the array indecis where 'n' is the number of dimensions
     *    and where each expression (index) is terminated by a NULL operand.
     */
    if (((*oph_ptr).optl & VOPT_ARRAY) != 0)
    {                   /* If variable is an array  */

        /*
         *    We must save the array dimension count for the operand we are working on
         *    at the current time. This is because we will be corrupting the OPD_PTR
         *    as we call ourselves recursively to return the pointer to the OPH/OPD
         *    which we are looking for not the one we are working on.
         */
        array_dimensions = (*opd_ptr).uopd.var.dimcntl;
        /* Save the array dimensions as */
        /* ...we over-write it calling  */
        for (isub = 0; isub < array_dimensions; isub++)
        {               /* ...for each dimension defined */
            array_expression_end = PFALSE;
            /* ...flag not end of expression */

            /*
             *    Skip through this expression, terminating on a NONE operand
             */
            do              /* ...loop through the LL    */
            {
                iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                                    ,(char *) &oph_ptr ,llb_pos
                                    ,NULL ,PSL_VM_MODE_SPEC);
                /* ...copy the OPH pointer  */
                if ((*oph_ptr).addrl == ADDR_NONE) array_expression_end = PTRUE;
                /* ...flag end of expression    */
                FORWARD_ONE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr ,&llb_pos);
            }               /* Skip 1 OPH/OPD pair      */
            while (!array_expression_end);
            /* ...until the end of the expression */
        }
    }

    /*
     *    Set up thereturn OPH/OPD pointers to point to the last OPH/OPD which we
     *    have processed.
     */
    *oph_pptr = oph_ptr;        /* Return OPH pointer       */
    *opd_pptr = opd_ptr;        /* Return OPD pointer       */

    /*
     *    The position pointer at this point is what we must return to the caller.
     */
    *llb_pos_pptr = llb_pos;
    return;
}
/*
 * 
 */

/*======================================================================*/
/*                                  */

/*                                  */
/*    GET_OPD_COUNT - Recover and return count of operands In LL    */
/*    =============                         */
/*                                  */
/*    Calling Format:                           */
/*  count = GET_OPD_COUNT (tpa_ptr)                 */
/*                                  */
/*    Where:                                */
/*  count = returned UL count of operands in the linked list    */
/*  tpa_ptr = pointer to the Tparse data                */
/*                                  */
/*    This routine scans through the linked list of parsed operands */
/*    and returns a count of those found.               */
/*                                  */
/*----------------------------------------------------------------------*/

static unsigned long    GET_OPD_COUNT   (struct scl_tpa *tpa_ptr)

{
    /*
     *    Local data
     */
    struct   oph      *oph_ptr;     /* Pointer to OPH       */
    unsigned long     count = 0;        /* Count of operands found  */
    unsigned long     iret;         /* Status return        */
    char     *llb_pos;     /* Linked list position     */
    char     *llb_ctx;     /* Linked list context      */

    llb_pos = NULL;         /* Position at start of list    */

    for ( ; ; count++)
    {
        iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                            ,(char *) &oph_ptr ,llb_pos ,&llb_ctx
                            ,PSL_VM_MODE_NEXT);
        if (!(psl_test_status (iret))) return count;
        llb_pos = llb_ctx;      /* Update the context       */

        /*
         *    There should always be an OPD following an OPH so pick it up to
         *    complete the operand.
         */
        iret = PSL_VM_COPY ((*tpa_ptr).parsed_operand_llps
                            ,(char *) &oph_ptr ,llb_pos ,&llb_ctx
                            ,PSL_VM_MODE_NEXT);
        if (!(psl_test_status (iret))) return count;
        llb_pos = llb_ctx;      /* Update the context       */
    }
}
/*
 * 
 */

/*======================================================================*/
/*                                  */

/*                                  */
/*    GET_OPERAND_POS - Return LL position of the nth primary operand   */
/*    ===============                           */
/*                                  */
/*    Calling Format:                           */
/*  GET_OPERAND_POS (tpa_ptr ,number ,&llb_ptr)         */
/*                                  */
/*    Where:                                */
/*  tpa_ptr = pointer to the Tparse data                */
/*  number  = Number of primary operand to return.          */
/*  llb_ptr = Address of a pointer to the LLB position      */
/*                                  */
/*    This routine scans through the linked list of parsed operands */
/*    and returns the position of the nth primary operand in the list.  */
/*                                  */
/*----------------------------------------------------------------------*/

static void GET_OPERAND_POS     (struct scl_tpa *tpa_ptr
                                 ,       long    number
                                 ,       char    **llb_pos_pptr)

{
    /*
     *    Local data
     */
    struct   oph      *oph_ptr;     /* Pointer to OPH       */
    struct   opd      *opd_ptr;     /* Pointer to OPD       */
    long     count = 0;        /* Count of operands found  */
    unsigned long     isub;         /* Loop control         */
    char     *llb_pos;     /* Linked list position     */

    llb_pos = NULL;         /* Position at start of list    */
    *llb_pos_pptr = NULL;       /* Set position to return   */

    for (count = 0; count < number ;count++)
    {
        FORWARD_ONE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr ,&llb_pos);
        /* Skip 1 OPH/OPD pair (PRIMARY) */
        *llb_pos_pptr = llb_pos;    /* ...and save the position     */

        /*
         *    Certain operands have trailing operands which define further the primary
         *    operand. In order that we might identify the primary operands it is
         *    necessary to move forward past any trailing operands which are in fact
         *    associated with the primary operand.
         *
         *    These are as follows:
         *
         *     VP$SCREEN        : 4 Further operands
         *
         */
        if ((*oph_ptr).addrl == ADDR_VP_SCREEN)
        {                   /* ...else if VP$SCREEN operand */
            for (isub = 0;  isub < 4; isub++)
            {
                FORWARD_ONE_OPERAND (tpa_ptr ,&oph_ptr ,&opd_ptr ,&llb_pos);
                /* Skip 1 OPH/OPD pair (PRIMARY) */
            }
        }
    }

    /*
     *    All done so return
     */
    return;
}
/*
 * 
 */
