/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 * scl_set_routines.c:
 * ===================
 *
 * This module contains TPA action routines used to set data in various data
 * structures after an item has been parsed. The data is taken from the data
 * 'saved' during parsing, e.g. by the 'save' action routines.
 */

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines */
#include <trc_rtndef.h>         /* Tracing Routines */
#include <vtest_rtndef.h>       /* Vtest Routines */
#include "scl_rtndef.h"         /* SCL Routines */
#include "scl_def_tbl_rtndef.h"     /* SCL definitions routines */
#include "scl_thr_tbl_rtndef.h"     /* SCL thread script routines */
#include "scl_env_tbl_rtndef.h"     /* SCL environment routines */
#include "scl_cod_tbl_rtndef.h"     /* SCL code  routines */
#include "scl_sub_tbl_rtndef.h"     /* SCL subexpression routines */

/*
 *  Structure definitions.
 */
#include <shrmsg.h>         /* PSL Messages */
#include <psldiag_prmdef.h>     /* Diagnostic Routine Symbols */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block */
#include <psl_trc_prmdef.h>     /* Tracing parameters */
#include <scpdef.h>         /* SCP Header record */
#include "scl_limits.h"         /* SCL limits */
#include "scl_tof.h"            /* SCL TOF data */
#include "scl_tlodef.h"         /* Test structures */
#include "scl_tbldef.h"         /* SCL Table Definitions */
#include <tofprmdef.h>          /* TOF parameters */
#include <vtest_prmdef.h>       /* VTEST Parameters */
#include "scl_cmd.h"            /* Command Line Data Structure  */
#include "scl_comdef.h"         /* Common Data */
#include "scl_tpacom.h"         /* Tparse common Data */

/************************************************************************/
/*                                  */
/*    SET_CODE_SIZE - Set size of code and other script sections    */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Pointer to TPARSE argument block            */
/*    Outputs:  Script sections offset and sizes set in header record   */
/*                                  */
/*    This routine sets the offsets and sizes of the value lists,       */
/*    timers, shareable variables and script parameters in the header   */
/*    record. In doing so it can calculate the remaining buffer size    */
/*    available for the code section.                   */
/*                                  */
/************************************************************************/

unsigned long   SET_CODE_SIZE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   scph    *header_ptr;       /* Pointer to header record.    */
    unsigned long    code_size;     /* remaining code size.     */
    struct   scl_cmd *scl_cmd_ptr;      /* Pointer to Command Line Struct */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;
    header_ptr = (*tpa_ptr).header_begps;

    /*
     *    Calculate and save the size (in 512 byte blocks) of the value lists
     */
    (*tpa_ptr).values_sizel = (unsigned long) (*tpa_ptr).values_nxtpc -
                              (unsigned long) (*tpa_ptr).values_begpc;
    /* Calculate size of value lists    */

    (*tpa_ptr).values_sizel = ((*tpa_ptr).values_sizel + 511) / 512;
    /* Number of pages of value lists   */

    /*
     *    Calculate and save the size of the timer definitions
     */
    (*tpa_ptr).timer_sizel = (*tpa_ptr).tim_cntl * sizeof (struct timtbl);
    /* Calculate size of timer lists    */
    (*tpa_ptr).timer_sizel = ((*tpa_ptr).timer_sizel + 511) / 512;
    /* Number of pages of timer lists   */

    /*
     *    Calculate and save the size of the shareable variable definitions
     */
    (*tpa_ptr).shrvar_sizel = (*tpa_ptr).shrvar_cntl * sizeof (struct shrvartbl);
    /* Calculate size of shareable lists */
    (*tpa_ptr).shrvar_sizel = ((*tpa_ptr).shrvar_sizel + 511) / 512;
    /* Number of pages of shareable variables */

    /*
     *    Calculate and save the size of the script parameters
     */
    (*tpa_ptr).parameter_sizel = (*tpa_ptr).scpprm_cntl *
                                 sizeof (struct prmtbl);
    /* Calculate size of parameter lists */
    (*tpa_ptr).parameter_sizel = ((*tpa_ptr).parameter_sizel + 511) / 512;
    /* Number of pages of parameters */

    /*
     *    Write the offsets of these buffers into the script header. The offsets
     *    are in bytes from the script header base address. Each section starts
     *    on a page boundary. The sections are (in order):
     *
     *  HEADER
     *  VALUE LISTS
     *  TIMER LISTS
     *  SHAREABLE VARIABLE LISTS
     *  SCRIPT PARAMETERS
     *      DEBUG VARIABLE SYMBOL TABLE
     *  CODE
     *      DEBUG LABEL SYMBOL TABLE
     */
    (*header_ptr).scph_timoff_l = 512 * ((*tpa_ptr).values_sizel + 1);
    /* Offset in bytes of timer     */
    /* Including header record. */

    (*header_ptr).scph_gbloff_l = (*header_ptr).scph_timoff_l
                                  + ((*tpa_ptr).timer_sizel * 512);
    /* Offset of shareable variables */
    /* ...following timer table */

    (*header_ptr).scph_prmoff_l = (*header_ptr).scph_gbloff_l
                                  + ((*tpa_ptr).shrvar_sizel * 512);
    /* Offset of script parameters  */
    /* ...following shareable variables */

    (*header_ptr).scph_codoff_l = (*header_ptr).scph_prmoff_l
                                  + ((*tpa_ptr).parameter_sizel * 512);
    /* Offset of Code  */
    /* ...following script parameters */
    /*
     *    We have a total of 500K for the SCP which is 512000 bytes. By
     *    subtracting the code offset from this figure we can determine how
     *    many bytes are available for the code section.  We also subtract the
     *    code offset from the start of the object buffer to give the effective
     *    SCP start address from which to calculate SCP offsets.
     */
    code_size = MAX_SCP_SIZE - (*header_ptr).scph_codoff_l;
    /* Calc space for code section  */
    (*tpa_ptr).code_endpc = (*tpa_ptr).code_begpc + code_size;
    /* ...and set end of code section */
    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    SET_DEF_CASE -   On  parsing a default case statement     */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:                                                       */
/*                                      */
/*    Outputs:   def_case_fgl             Default case flag set         */
/*                                  */
/*   This routine is called on parsing a DEFAULT_CASE statement.        */
/*   It sets a variable to indicate that we have parsed a default case  */
/*                                  */
/*======================================================================*/

unsigned long   SET_DEF_CASE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   swstack *sw_ptr;               /* Pointer to switch stack      */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    sw_ptr  = (*tpa_ptr).swstack_begps + (*tpa_ptr).sw_depthl - 1;
    /* Point to current entry   */
    /*
     * Set default case flag on  switch stack.
     */

    (*sw_ptr).def_case_fgl = 1;        /* Set default_case flag   */

    return PSL_NORMAL;             /* ...and return success */

}
/*
 * 
 */

/************************************************************************/

/*    SET_ERROR_FLAG - On finding error in switch statement     */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:                                                       */
/*                                      */
/*    Outputs:   sw_error_fgl             Default case flag set         */
/*                                  */
/*   This routine is called on finding error in switch  statement.      */
/*   It sets a variable to indicate that we have found an error         */
/*   This is checked against when encoding the switch statement         */
/*   If it is set nothing is encoded                    */
/*                                  */
/*======================================================================*/

unsigned long   SET_ERROR_FLAG (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     * Set error flag on  switch stack.
     */

    (*tpa_ptr).sw_error_fgl = 1;           /* Set error flag   */

    return PSL_NORMAL;             /* ...and return success */

}
/*
 * 
 */


/************************************************************************/

/*    SET_DO_ERRFGL - On finding error in DO statement          */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:                                                       */
/*                                      */
/*    Outputs:   do_error_fgl             Do error flag set             */
/*                                  */
/*   This routine is called on finding error in do statement.           */
/*   It sets a variable to indicate that we have found an error         */
/*   This is checked against when encoding the do statement             */
/*   If it is set nothing is encoded                    */
/*                                  */
/*======================================================================*/

unsigned long   SET_DO_ERRFGL (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     * Set error do error flag.
     */

    (*tpa_ptr).do_error_fgl = 1;           /* Set error flag   */

    return PSL_NORMAL;             /* ...and return success */

}
/*
 * 
 */

/************************************************************************/

/*                                  */
/*    SET_END_RECORD - Set end of a record definition.              */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   DEF_RECIDXL - Record Start index            */
/*      Variables symbol table: INDEX               */
/*                  REC_ITEM            */
/*                  START_REC_ARRAY         */
/*    Outputs:  None                            */
/*                                  */
/*    This routine is called on parsing an END RECORD statement.    */
/*                                  */
/*    Its purpose is to determine whether or not we have started a  */
/*    record and if so to calculate the length of the record and save   */
/*    it into the variables symbol table.               */
/*                                  */
/*    Creates and stores pointer to format definition array.        */
/*                                  */
/*======================================================================*/

unsigned long   SET_END_RECORD (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct shrvartbl *shrvar_ptr;       /* Shareable variable data pointer */

    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */
    struct   symtbl  *var_ptr;      /* Variable pointer     */
    struct   symtbl  *rec_item_ptr;     /* Record item pointer      */
    unsigned long    rec_size;      /* Size of record       */
    long    *rec_array;        /* Format array         */
    unsigned long    num_items;     /* Items in record      */
    unsigned long    size_rec_array;    /* Size of array required   */
    unsigned long    loop;          /* Array looping variable   */
    unsigned long    array_ind;     /* Array index          */

#ifdef  VMS
    struct   scl_cmd *cmd_ptr;      /* Pointer to Command Line Struct */
#endif


    unsigned long    alignment;     /* Longword alignment          */
    unsigned long    last_dtyp;     /* Previous record data type    */
    unsigned long    record_size;       /* Keep track of record size    */
    /* ensuring alignment okay      */


    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    shrvar_ptr = (*tpa_ptr).last_shrvar_recps;

#ifdef  VMS
    cmd_ptr = (*com_ptr).cmd_ptrps;
#endif

    /*
     *    If we have had an error in the record definition then skip the record
     */
    if ((*tpa_ptr).rec_error_flgl != 0) goto exit;
    /* Error in record      */

    /*
     *    Check for a record being processed
     */
    if ((*tpa_ptr).def_recidxl == 0)    /* If not processing a record   */
    {
        REPORT_ERROR (argblk_ptr, SCL_INVENDREC);
        /* Report error         */
    }
    else
    {

        /*
         * 
         * Code modified to create a format definitions array
         */

        /*
         *    We are processing a record. Retrieve the offset of the record and from
         *    that calculate the record length
         */
        var_ptr = (*tpa_ptr).variable_begps + (*tpa_ptr).def_recidxl;
        /* Pointer to record entry     */

        rec_item_ptr = (*tpa_ptr).variable_begps + (*tpa_ptr).def_recidxl + 1;
        /* Pointer to first item in record */

        /*
         *    Number of items in a record
         *    Variable count - record header index - 1
         *    Record header does not count as an item in the record hence -1
         */
        num_items = (*tpa_ptr).var_cntl - (*tpa_ptr).def_recidxl;
        if (num_items == 0) goto end_rec;
        /* Exit if record is empty */
        num_items -= 1;         /* ...decrement count       */

        /*
         * CALLOC enough space for an array of longwords.
         * These will be initialised to 0.
         *
         *  Num_items + 1 (NULL terminator)
         */

        size_rec_array = num_items + 1;
        /*
         * Store necessary info in record table
         *
         *  Index of record.
         *  Start of format array
         *  Number of items.
         */

        (*var_ptr).var_indexl = (*tpa_ptr).rec_cntl;

        (*var_ptr).var_start_rec_arraypl = (long *) CALLOC ( size_rec_array ,
                                           sizeof (long));

#ifdef PSL_TRACING
        if (PSL_TEST_TRACE (PTRC_USR_USER5))
        {
            PSL_OUTPUT_TRACE ("MALLOC: recarray - '%d * %d/%p' - SET_ENDREC\n",
                              size_rec_array, sizeof (long),
                              (*var_ptr).var_start_rec_arraypl);
        }
#endif  /* PSL_TRACING */


        rec_array = (*var_ptr).var_start_rec_arraypl;

        (*var_ptr).var_rec_iteml = num_items;

        /*
         * 
         * Modify record definition array to take account of
         * record alignment.
         */
        record_size = 0;                    /* Initialise record size */
        last_dtyp = 0;              /* Initialise data type   */
        alignment = 0;              /* Initialise alignment   */
        /*
         * 
         */

        /*
         *    Loop through and set up this array of format identifiers.
         *
         *    If record alignment has been specified then integers must always be
         *    on a longword boundary, therefore there may well be padding between
         *    characters and integers
         *
         *    The FDA must take account of this.
         *
         *  If INTEGER  
         *       Encode  -1
         *               Add padding to previous entry if necessary (unless no align)
         *
         *  Else If CHARACTER
         *       Encode the size of the character.
         *      End if
         *  Should be no other data types in a record.
         */

        for (loop = 0,array_ind = 0;
                loop < num_items;
                loop++)
        {
            if ((*rec_item_ptr).var_typel == DTYP_INTEGER)
            {

                /*
                 *    The next item in the arry is an integer. If we have enabled record
                 *    alignment then we should ensure that the longword is aligned on a longword
                 *    boundary. If we are on VMS however we may turn off alignment of records
                 *    and so would not need to add fillers here.
                 */
#ifdef  VMS
                if (((*cmd_ptr).align_recordsn)
                        &&  (last_dtyp == DTYP_CHARACTER))
#else
                if (last_dtyp == DTYP_CHARACTER)
#endif
                {
                    alignment = (record_size % sizeof(long));
                    /* Get any "remainder"      */

                    alignment = ((sizeof(long) - alignment) % sizeof(long));
                    /* Get longword shortfall   */
                    /*
                     *    Add alignment to previous character entry in record definition array
                     *    and current record_size.
                     */
                    rec_array[array_ind - 1] += alignment;

                    record_size += alignment;
                }

                rec_array[array_ind] = -1 ; /* Invert if integer            */

                record_size += (*rec_item_ptr).var_sizel;
                /* Update record size           */
                last_dtyp = DTYP_INTEGER;
                /* Save data type               */
            }
            else if((*rec_item_ptr).var_typel == DTYP_CHARACTER)
            {

                rec_array[array_ind] = (*rec_item_ptr).var_sizel ;

                record_size += (*rec_item_ptr).var_sizel;
                /* Update record size           */
                last_dtyp = DTYP_CHARACTER;
                /* Save data type               */
            }

            array_ind += 1;               /* Next array entry         */

            rec_item_ptr += 1;            /* Next item in record.     */
        }

        rec_array[size_rec_array - 1] = 0;
        /* ENSURE NULL TERMINATED ARRAY   */

        /*
         *   Record  size  requires scope dependent addresses
         */
        if ((*tpa_ptr).rec_addrl == ADDR_LOCAL)
        {
            rec_size = (*tpa_ptr).local_data_offl - (*var_ptr).var_offl;
            /* Calculate size of record */
        }

        else if (((*tpa_ptr).rec_addrl == ADDR_GLOBAL)
                 ||   ((*tpa_ptr).rec_addrl == ADDR_THREAD)
                 ||   ((*tpa_ptr).rec_addrl == ADDR_TESTWIDE))

        {
            rec_size = (*tpa_ptr).shrvar_rec_offl ;
            (*shrvar_ptr).shrvar_sizel = rec_size;
            /* Calculate size of record */
        }

        else if ((*tpa_ptr).rec_addrl == ADDR_SCRIPT)
        {
            rec_size = (*tpa_ptr).script_data_offl - (*var_ptr).var_offl;
            /* Calculate size of record */
        }

        /*
         *    The record size must be no greater than the maximum permitted size of
         *    a character variable. This is currently 65535 characters.
         */
        if (rec_size > MAX_VARIABLE_SIZE)
        {               /* If record too large truncate it */
            rec_size = MAX_VARIABLE_SIZE;
            REPORT_ERROR (argblk_ptr, SCL_RECTOOBIG);
            /* Report record too big error    */
        }

        (*var_ptr).var_sizel = rec_size; /* Store record size     */
    }

end_rec:
    /*
     * 
     * Must ensure alignment after a record has been processed.
     */
    if ((*tpa_ptr).rec_addrl == ADDR_LOCAL)
    {
        rec_size = (*tpa_ptr).local_data_offl - (*var_ptr).var_offl;
        /* Calculate size of record */

        (*tpa_ptr).local_data_offl += (*tpa_ptr).local_align_offl;
        /* Add alignment offset         */
        (*tpa_ptr).local_align_offl = 0;
        /* Reset alignment          */
    }

    if ((*tpa_ptr).rec_addrl == ADDR_SCRIPT)
    {

        (*tpa_ptr).script_data_offl += (*tpa_ptr).script_align_offl;
        /* Add alignment offset         */
        (*tpa_ptr).script_align_offl = 0;
        /* Reset alignment          */
    }
    /*
     * 
     */
exit:
    ;
    (*tpa_ptr).def_recidxl = 0;     /* Clear record index pointer     */
    (*tpa_ptr).rec_addrl = ADDR_LOCAL;  /* Clear address                  */
    (*tpa_ptr).def_addrl = ADDR_LOCAL;  /* Clear address                  */
    (*tpa_ptr).rec_error_flgl = 0;      /* Clear error flag       */


    return PSL_NORMAL;          /* ...and return success      */
}

/************************************************************************/
/*                                  */
/*    SET_HDR_CNT - Set HTTP Request HEADER Value Count to 1        */
/*    ===========                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  tpa_ptr->http_hdr_cntl set to 1                 */
/*                                  */
/*    This routine is called to set the value count for the HTTP    */
/*    request HEADER clause that has just been parsed to 1. It is   */
/*    called when a single value - rather than a value list - has   */
/*    been parsed.                          */
/*                                  */
/************************************************************************/
unsigned long SET_HDR_CNT (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *    Set the value count.
     */
    tpa_ptr->http_hdrval_cntl = 1;

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    SET_REC_HEADER - Set IN_REC_HEADER as specifying record scope     */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:                                                       */
/*      Variables symbol table                  */
/*    Outputs:  None                            */
/*                                  */
/*     This routine is called on parsing an RECORD statement.       */
/*     It sets a variable to indicate that we are specifying scope of   */
/*     record in record header.                                         */
/*     It is reset to zero after specifying record scope                */
/*                                                                      */
/*     This is used later in error checking routines.                   */
/*                                                                      */
/*     Errors occur if the scope specified for variables in the record  */
/*     are not the same as that specified in the record header.         */
/*                                                                      */
/*     The record header scope would be "checked" against the default   */
/*     rec_addrl which is local and hence cause error - the record      */
/*     header scope does not need checking in this way. IN_REC_HEADERL  */
/*     when set is used to exclude header scope from these checks.      */
/*                                  */
/*                                  */
/*======================================================================*/

unsigned long   SET_REC_HEADER (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     * Set "in_record"
     */
    (*tpa_ptr).in_rec_headl = 1;      /* Set as at header of record  */


    (*tpa_ptr).rec_cntl += 1;   /* Increment count of records    */

    return PSL_NORMAL;           /* ...and return success     */

}
/*
 * 
 */

/************************************************************************/

/*                                  */
/*    SET_IGNORE_DIFFERENCES                        */
/*    ======================                        */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   TPARSE argument block                               */
/*                                      */
/*    Outputs:   (*tpa_ptr).ignore_differencesn (Set if input = 1)      */
/*                                  */
/*    This routine is called to either enable or disable the duplicate  */
/*    definition of SCL variables/timer and/or files. If the input  */
/*    parameter is 1 we enable duplicates otherwise we disable them.    */
/*                                  */
/*    This is of most use when we are parsing an SQL mode script which  */
/*    may include data files generated by the SQL editor which may  */
/*    contain duplicate definitions.                    */
/*                                  */
/*======================================================================*/

unsigned long   SET_IGNORE_DIFFERENCES  (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    if ((*argblk_ptr).paraml == 1)
    {
        (*tpa_ptr).ignore_duplicatesn = PTRUE;
    }
    else
    {
        (*tpa_ptr).ignore_duplicatesn = PFALSE;
    }

    return PSL_NORMAL;                  /* ...and return success */
}

/************************************************************************/
/*                                  */
/*    SET_INPUT_RATE - Set Character input rate             */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   INPUT_DELAY = Input delay value, or flag, to        */
/*                    put in the script header record.      */
/*    Outputs:  None                            */
/*                                  */
/*    This routine sets the Character Input Delay field in the Script   */
/*    Header record. On entry input delay value is held in      */
/*    INPUT_DELAY. Note that if an error has occurred in parsing the    */
/*    input rate this value may be invalid.             */
/*                                  */
/************************************************************************/

unsigned long   SET_INPUT_RATE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct scph     *header_ptr;        /* Pointer to header record */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    header_ptr = (*tpa_ptr).header_begps;
    /* Pointer to header record */

    /*
     *    Copy the input rate out of temporary storage into the script header.
     */
    (*header_ptr).scph_inpdly_w = (short) (*tpa_ptr).input_delayl;
    return PSL_NORMAL;      /* ...and return success    */
}


/************************************************************************/
/*                                  */
/*    SET_PIPE_SYNC - Set pipe as Synchronous               */
/*    =============                         */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Pointer to TPARSE argument block            */
/*    Outputs:  Current pipe flagged as being synchronous       */
/*                                  */
/*    This routine sets the synchronous flag in the FAT entry for the   */
/*    current pipe.                             */
/*                                  */
/************************************************************************/

unsigned long   SET_PIPE_SYNC (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    (*tpa_ptr).def_optl = VOPT_PIPE_SYNC;
    /* Set synchronous flag     */
    return PSL_NORMAL;          /* ...and return success    */
}



/************************************************************************/
/*                                  */
/*  SET_PROT_CAT - Store Protection Category            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = SET_PROT_CAT (argblk_ptr)             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine saves the Protection Category .            */
/*                                  */
/************************************************************************/

unsigned long SET_PROT_CAT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct scph     *header_ptr;        /* Pointer to header record */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    header_ptr = (*tpa_ptr).header_begps;
    /* Pointer to header record */

    /*
     *  Initialize the mask.
     */
    (*tpa_ptr).prot_catl = (*argblk_ptr).paraml;

    switch ((*tpa_ptr).prot_catl)
    {
    case SPROT_SYSTEM_V:
        {
            (*header_ptr).scph_sprot_l |= 0xF;     /* Disallow all access */
            break;
        }
    case SPROT_OWNER_V:
        {
            (*header_ptr).scph_sprot_l |= 0xF0;    /* Disallow all access */
            break;
        }
    case SPROT_GROUP_V:
        {
            (*header_ptr).scph_sprot_l |= 0xF00;   /* Disallow all access */
            break;
        }
    case SPROT_WORLD_V:
        {
            (*header_ptr).scph_sprot_l |= 0xF000;  /* Disallow all access */
            break;
        }
    default:                /* Should never get here */
        {
            break;
        }
    }

    /*
     *  Protection not defaulted.
     */
    (*header_ptr).scph_prot_defaulted_n = PFALSE;

    return PSL_NORMAL;
}



/************************************************************************/
/*                                  */
/*  SET_PROT_ACC - Store Protection Access              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = SET_PROT_ACC (argblk_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine stores the Protection Access for later use.    */
/*                                  */
/************************************************************************/

unsigned long SET_PROT_ACC (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct scph     *header_ptr;        /* Pointer to header record */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    header_ptr = (*tpa_ptr).header_begps;
    /* Pointer to header record */

    /*
     *  Clear the bit corresponding to the specified access in the nibble
     *  for the current category.
     */
    switch ((*tpa_ptr).prot_catl)
    {
    case SPROT_SYSTEM_V:
        {
            switch ((*argblk_ptr).paraml)
            {
            case SPTP_RED_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~ARM$M_READ;
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~SPTP_RED_M;
#endif /* VMS */
                    break;
                }
            case SPTP_WRT_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~ARM$M_WRITE;
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~SPTP_WRT_M;
#endif /* VMS */
                    break;
                }
            case SPTP_EXE_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~ARM$M_EXECUTE;
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~SPTP_EXE_M;
#endif /* VMS */
                    break;
                }
            case SPTP_DEL_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~ARM$M_DELETE;
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~SPTP_DEL_M;
#endif /* VMS */
                    break;
                }
            default:            /* Can never get here */
                {
                    break;
                }
            }
            break;
        }
    case SPROT_OWNER_V:
        {
            switch ((*argblk_ptr).paraml)
            {
            case SPTP_RED_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(16 * ARM$M_READ);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(16 * SPTP_RED_M);
#endif /* VMS */
                    break;
                }
            case SPTP_WRT_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(16 * ARM$M_WRITE);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(16 * SPTP_WRT_M);
#endif /* VMS */
                    break;
                }
            case SPTP_EXE_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(16 * ARM$M_EXECUTE);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(16 * SPTP_EXE_M);
#endif /* VMS */
                    break;
                }
            case SPTP_DEL_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(16 * ARM$M_DELETE);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(16 * SPTP_DEL_M);
#endif /* VMS */
                    break;
                }
            default:            /* Can never get here */
                {
                    break;
                }
            }
            break;
        }
    case SPROT_GROUP_V:
        {
            switch ((*argblk_ptr).paraml)
            {
            case SPTP_RED_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(256 * ARM$M_READ);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(256 * SPTP_RED_M);
#endif /* VMS */
                    break;
                }
            case SPTP_WRT_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(256 * ARM$M_WRITE);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(256 * SPTP_WRT_M);
#endif /* VMS */
                    break;
                }
            case SPTP_EXE_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(256 * ARM$M_EXECUTE);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(256 * SPTP_EXE_M);
#endif /* VMS */
                    break;
                }
            case SPTP_DEL_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(256 * ARM$M_DELETE);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(256 * SPTP_DEL_M);
#endif /* VMS */
                    break;
                }
            default:            /* Can never get here */
                {
                    break;
                }
            }
            break;
        }
    case SPROT_WORLD_V:
        {
            switch ((*argblk_ptr).paraml)
            {
            case SPTP_RED_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(4096 * ARM$M_READ);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(4096 * SPTP_RED_M);
#endif /* VMS */
                    break;
                }
            case SPTP_WRT_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(4096 * ARM$M_WRITE);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(4096 * SPTP_WRT_M);
#endif /* VMS */
                    break;
                }
            case SPTP_EXE_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(4096 * ARM$M_EXECUTE);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(4096 * SPTP_EXE_M);
#endif /* VMS */
                    break;
                }
            case SPTP_DEL_V:
                {
#ifdef VMS
                    (*header_ptr).scph_sprot_l &= ~(4096 * ARM$M_DELETE);
#else /* VMS */
                    (*header_ptr).scph_sprot_l &= ~(4096 * SPTP_DEL_M);
#endif /* VMS */
                    break;
                }
            default:            /* Can never get here */
                {
                    break;
                }
            }
            break;
        }
    default:                /* Can never get here */
        {
            break;
        }
    }

    return 1;
}




/************************************************************************/
/*                                  */
/*    SET_SCRIPT_DESC - Set script description              */
/*    ===============                           */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Pointer to TPARSE argument block            */
/*    Outputs:  Script description set in header record         */
/*                                  */
/*    This routine sets the Script Description field in the Script  */
/*    Object ('tof') header record data buffer. Report an error if  */
/*    the script description is too long.               */
/*                                  */
/************************************************************************/
unsigned long SET_SCRIPT_DESC (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */
    struct scl_tofh *tofh_ptr;      /* Pointer to header record data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tofh_ptr = (*tpa_ptr).tofh_begps;

    /*
     *    Check to ensure that the description is no longer than 50 characters
     *    reporting an error if it is and saving it into the header record (as
     *    a null-terminated string) if it is not.
     */
    if ((*tpa_ptr).chr_strlenl > 50)
    {
        REPORT_ERROR (argblk_ptr, PSL_DESCRTOOLONG);
    }

    else
    {
        (void) memcpy ((*tofh_ptr).tofh_sdsc_c52,
                       (*tpa_ptr).chr_stringc65536, (*tpa_ptr).chr_strlenl);
        (*tofh_ptr).tofh_sdsc_c52 [(*tpa_ptr).chr_strlenl] = '\0';
    }

    return PSL_NORMAL;      /* ...and return success */
}

/************************************************************************/
/*                                  */
/*    SET_STRING_CHR - Set string character variable.           */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  tpa_ptr->string_chrc8 - Character command using     */
/*          ASCII mnemonic.                 */
/*                                  */
/*    This routine sets the character buffer to the character command   */
/*    (in ASCII mnemonic form) from the mnemonic string passed in the   */
/*    current token, e.g. a current token of "CR" will create a     */
/*    character command string of ~<CR>.                */
/*                                  */
/************************************************************************/
unsigned long SET_STRING_CHR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;   /* Common Data          */
    struct scl_tpa  *tpa_ptr;   /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *  Reconstruct the character command in the character buffer in the form ~<mmm>,
     *  where "mmm" is the mnemonic string held in the current token.
     */
    tpa_ptr->string_chrc8 [0] = (*com_ptr).cmnd_chr_valuec256 [0];
    tpa_ptr->string_chrc8 [1] = '<';
    memcpy ((void *) &tpa_ptr->string_chrc8 [2],
            (void *) argblk_ptr->token_ptrpc,
            argblk_ptr->tokencntl);
    tpa_ptr->string_chrc8 [argblk_ptr->tokencntl+2] = '>';
    tpa_ptr->string_chrc8 [argblk_ptr->tokencntl+3] = '\0';

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    SET_STRING_CMNDCHR - Set string character to "~~"         */
/*    ==================                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  tpa_ptr->string_chrc8 - Character command for       */
/*          command character literal.                  */
/*                                  */
/*    This routine sets the character buffer to the character command   */
/*    representing a lietarl character command character, i.e. "~~".    */
/*                                  */
/************************************************************************/
unsigned long SET_STRING_CMNDCHR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;   /* Common Data */
    struct scl_tpa  *tpa_ptr;   /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  Reconstruct the character command in the character buffer in the form ~~
     */
    tpa_ptr->string_chrc8 [0] = com_ptr->cmnd_chr_valuec256 [0];
    tpa_ptr->string_chrc8 [1] = com_ptr->cmnd_chr_valuec256 [0];
    tpa_ptr->string_chrc8 [2] = '\0';
    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/

/*    SET_STRING_CTRLCHR - Set string character to "~^"         */
/*    ==================                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  tpa_ptr->string_chrc8 - Character command for       */
/*          control character literal.                  */
/*                                  */
/*    This routine sets the character buffer to the character command   */
/*    representing a literal control character, i.e. "~^".      */
/*                                  */
/************************************************************************/
unsigned long SET_STRING_CTRLCHR (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;   /* Common Data */
    struct scl_tpa  *tpa_ptr;   /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *  Reconstruct the character command in the character buffer in the form ~^.
     */
    tpa_ptr->string_chrc8 [0] = com_ptr->cmnd_chr_valuec256 [0];
    tpa_ptr->string_chrc8 [1] = com_ptr->ctrl_chr_valuec256 [0];
    tpa_ptr->string_chrc8 [2] = '\0';
    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/
/*                                  */
/*    SET_TERM_TYPE - Set terminal type.                */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Terminal Type          */
/*    Outputs:  None                            */
/*                                  */
/*    This routine sets the Terminal Type field in the Script       */
/*    Header record to the value passed in the parameter field of the   */
/*    TPARSE argument block.                        */
/*                                  */
/************************************************************************/

unsigned long   SET_TERM_TYPE (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */
    struct scph     *header_ptr;        /* Pointer to header record */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    header_ptr = (*tpa_ptr).header_begps;
    /* Pointer to header record */

    /*
     *    Copy the script name out of temporary storage into the script header,
     *    ensuring that it is space filled.
     */
    (*header_ptr).scph_ttyp_b = (char) (*argblk_ptr).paraml;
    return PSL_NORMAL;      /* ...and return success    */
}

/************************************************************************/

/*    SET_THR_ANSWER - Set thread 'answerback' option           */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread 'answerback' message  */
/*    field in the thread definition work structure to the current  */
/*    token value (minus the quotes). However, it first validates the   */
/*    message reporting an error if it is invalid.          */
/*                                  */
/************************************************************************/
unsigned long SET_THR_ANSWER (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Routine return status */
    char       wrkbuf [128];   /* Work buffer */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Copy the value to our local buffer for checking.
     */
    (void) memcpy ((void *) wrkbuf,
                   (void *) (*tpa_ptr).chr_stringc65536,
                   (*tpa_ptr).chr_strlenl);
    wrkbuf [(*tpa_ptr).chr_strlenl] = '\0';
    /* Null terminate */

    /*
     *  Check the value and convert it to its storable format.
     */
    iret = VTEST_ANSWERBACK_CHK (wrkbuf,
                                 (*tdf_ptr).tdf_answerbackc32,
                                 sizeof ((*tdf_ptr).tdf_answerbackc32));
    if (!(psl_test_status (iret)))
    {
        REPORT_ERROR (argblk_ptr, iret);
        goto end;
    }

    /*
     *  Clear the default flag for this option.
     */
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_ANSWERBACK;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_AUTSYN - Set thread 'automatic synchronization' option    */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = Automatic synchronization flag      */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'automatic synchronization' flag */
/*    in the thread definition work structure to the value passed in    */
/*    the parameter field of the TPARSE argument block.         */
/*                                  */
/************************************************************************/
unsigned long SET_THR_AUTSYN (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'automatic synchronization' flag value.
     */
    (*tdf_ptr).tdf_auto_syncl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags2l &= ~TDF_THRDF2_AUTO_SYNC;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_BAUD - Set thread 'baud rate' option          */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = Baud rate (TOF_THRBDR_*)        */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'baud rate' field in the thread  */
/*    definition work structure to the value passed in the parameter    */
/*    field of the TPARSE argument block.               */
/*                                  */
/************************************************************************/
unsigned long SET_THR_BAUD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'baud rate' flag value.
     */
    (*tdf_ptr).tdf_baud_ratel = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_BAUD_RATE;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_CHRSIZ - Set thread 'character size' option       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = Character size value            */
/*        (CHARACTER_SIZE_7BIT, CHARACTER_SIZE_8BIT).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'character size' field in the thread */
/*    definition work structure to the value passed in the parameter    */
/*    field of the TPARSE argument block.               */
/*                                  */
/************************************************************************/
unsigned long SET_THR_CHRSIZ (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'character size' field value.
     */
    (*tdf_ptr).tdf_character_sizel = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_CHARACTER_SIZE;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_DETEXT - Set thread 'detach on exit' option       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = Detach on exit flag         */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'detach on exit' field in the thread */
/*    definition work structure to the value passed in the parameter    */
/*    field of the TPARSE argument block.               */
/*                                  */
/************************************************************************/
unsigned long SET_THR_DETEXT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'detach on exit' flag value.
     */
    (*tdf_ptr).tdf_detach_exitl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_DETACH_EXIT;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_DEVTYP - Set thread 'device type' option          */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = Device type (DEVICE_TYPE_*).        */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'device type' field in the thread    */
/*    definition work structure to the value passed in the parameter    */
/*    field of the TPARSE argument block.               */
/*                                  */
/************************************************************************/
unsigned long SET_THR_DEVTYP (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'device type' field.
     */
    (*tdf_ptr).tdf_device_typel = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_DEVICE_TYPE;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_EMULMOD - Set thread 'emulator model' option      */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'emulator model' field in the    */
/*    thread definition work structure to the current token value.  */
/*                                  */
/************************************************************************/
unsigned long SET_THR_EMULMOD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'emulator model' field.
     */
    (void) memcpy ((void *) (*tdf_ptr).tdf_emulator_modelc16,
                   (void *) (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);
    (*tdf_ptr).tdf_emulator_modelc16[(*argblk_ptr).tokencntl] = '\0';

    (*tdf_ptr).tdf_default_flags2l &= ~TDF_THRDF2_EMULMOD;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_EMULTYP - Set thread 'emulator type' option       */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'emulator type' field in the     */
/*    thread definition work structure to the current token value.  */
/*    To make the SCL parsing easier, the emulator type values      */
/*    passed to this routine contain hyphens (rather than spaces)   */
/*    as separators. The first hyphen in any value must now be      */
/*    replaced by a space before being stored.              */
/*                                  */
/************************************************************************/
unsigned long SET_THR_EMULTYP (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */
    char*      pChar;      /* Work variable */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'emulator type' field.
     */
    (void) memcpy ((void *) (*tdf_ptr).tdf_emulator_typec16,
                   (void *) (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);
    (*tdf_ptr).tdf_emulator_typec16[(*argblk_ptr).tokencntl] = '\0';

    /*
     *    A little bit of post-processing here as we have a defined set of 
     *    values that we must store and we have slightly changed these values to
     *    make the SCL code easier.
     *
     *    Replace the first hyphen character with a space.
     */
    pChar = strchr ((*tdf_ptr).tdf_emulator_typec16, '-');
    if (pChar != NULL) *pChar = ' ';

    (*tdf_ptr).tdf_default_flags2l &= ~TDF_THRDF2_EMULTYP;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_FKEY - Set thread 'function key controls' option      */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = Function Key Controls flag      */
/*        (CHARACTER_SIZE_7BIT, CHARACTER_SIZE_8BIT).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'function key controls' field in */
/*    the thread definition work structure to the value passed in the   */
/*    parameter field of the TPARSE argument block.         */
/*                                  */
/************************************************************************/
unsigned long SET_THR_FKEY (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'function key controls' field.
     */
    (*tdf_ptr).tdf_fkey_controll = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_FKEY_CONTROL;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_HNGDET - Set thread 'hangup on detach' option     */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = Hangup on detach flag           */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'hangup on detach' flag in the   */
/*    thread definition work structure to the value passed in the   */
/*    parameter field of the TPARSE argument block.         */
/*                                  */
/************************************************************************/
unsigned long SET_THR_HNGDET (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'hangup on detach' field.
     */
    (*tdf_ptr).tdf_hangup_detachl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_HANGUP_DETACH;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_INPTIM - Set thread 'input timing' option         */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = Input timing value (TOF_THRITM_*)   */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'input timing' field in the      */
/*    thread definition work structure to the value passed in the   */
/*    parameter field of the TPARSE argument block.         */
/*                                  */
/************************************************************************/
unsigned long SET_THR_INPTIM (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set 'input timing' field.
     */
    (*tdf_ptr).tdf_input_timingl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_INPUT_TIMING;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_INET - Set thread 'internet name/address' option      */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread 'internet name/address'   */
/*    field in the thread definition work structure to the current  */
/*    token value. However, it first validates the token length     */
/*    reporting an error if it is too long.             */
/*                                  */
/************************************************************************/
unsigned long SET_THR_INET (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */
    unsigned long       offset;     /* Source line offset */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*argblk_ptr).tokencntl  > LEN_THR_INET)
    {
        REPORT_ERROR (argblk_ptr, SCL_INETTOOLONG);
        goto end;
    }

    /*
     *  Set 'internet name/address' field - stripping off any enclosing quotes.
     */
    if ((*((*argblk_ptr).token_ptrpc) == '"') ||
            (*((*argblk_ptr).token_ptrpc) == '\''))
    {
        /*
         *    The character is a printing character. Pick up the specified character
         *    from the source line as input and save it.
         */
        offset = (long) ((*argblk_ptr).token_ptrpc - (*tpa_ptr).parsable_linec136) + 1;
        /* Current offset in source line */
        (void) memcpy ((void *) (*tdf_ptr).tdf_fqtcpip_addressc128,
                       (void *) &(*tpa_ptr).source_linec136 [offset],
                       (*argblk_ptr).tokencntl-2);
        (*tdf_ptr).tdf_fqtcpip_addressc128[(*argblk_ptr).tokencntl-2] = '\0';
    }
    else
    {
        (void) memcpy ((void *) (*tdf_ptr).tdf_fqtcpip_addressc128,
                       (void *) (*argblk_ptr).token_ptrpc,
                       (*argblk_ptr).tokencntl);
        (*tdf_ptr).tdf_fqtcpip_addressc128[(*argblk_ptr).tokencntl] = '\0';
    }

    (*tdf_ptr).tdf_default_flags2l &= ~TDF_THRDF2_TCPIPADDR;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_IOM - Set thread 'IOM enabled/disabled' option        */
/*    ===========                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = IOM enabled/disabled flag       */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'IOM enabled' field in the       */
/*    thread definition work structure to the value passed in the   */
/*    parameter field of the TPARSE argument block.         */
/*                                  */
/************************************************************************/
unsigned long SET_THR_IOM (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_enable_ioml = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags2l &= ~TDF_THRDF2_ENABLE_IOM;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_LOGIN - Set thread 'login' option             */
/*    =============                                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = Login flag (TOF_THRLOG_*)       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'login type' field in the        */
/*    thread definition work structure to the value passed in the   */
/*    parameter field of the TPARSE argument block. However, it first   */
/*    checks that the login type is valid for the platform on which it  */
/*    is being built reporting an error if it is not.           */
/*                                  */
/************************************************************************/
unsigned long SET_THR_LOGIN (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  IF WE ARE RUNNING ON WINDOWS IGNORE THE FOLLOWING CHECKS AS WE MAY BE
     *  COMPILING FOR SYNTAX CHECKING ONLY AS THE SCRIPT WILL BE RUN REMOTELY.
     */
#ifndef WIN32

    /*
     *  RLOGIN & TELNET are UNIX-Specific.
     */
#ifdef VMS
    if (((*argblk_ptr).paraml == TOF_THRLOG_RLOGIN) ||
            ((*argblk_ptr).paraml == TOF_THRLOG_TELNET))
    {
        REPORT_ERROR (argblk_ptr, PSL_INVKEYWRD);
        goto end;
    }
#endif

    /*
     *  SCRIPT is VMS-Specific.
     */
#ifndef VMS
    if ((*argblk_ptr).paraml == TOF_THRLOG_SCRIPT)
    {
        REPORT_ERROR (argblk_ptr, PSL_INVKEYWRD);
        goto end;
    }
#endif

#endif /* ifndef WIN32 */

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_login_typel = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_LOGIN_TYPE;

    /*
     *  end:
     *  Routine exit point.
     */
#ifndef WIN32
end:
#endif
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_NODE - Set thread 'node' option               */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread 'node'            */
/*    field in the thread definition work structure to the current  */
/*    token value. However, it first validates the token length     */
/*    reporting an error if it is too long.             */
/*                                  */
/************************************************************************/
unsigned long SET_THR_NODE (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*argblk_ptr).tokencntl  > LEN_THR_NODE)
    {
        REPORT_ERROR (argblk_ptr, SCL_NODETOOLONG);
        goto end;
    }

    /*
     *  Set 'node name' field.
     */
    (void) memcpy ((void *) (*tdf_ptr).tdf_ltnodec8,
                   (void *) (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);
    (*tdf_ptr).tdf_ltnodec8[(*argblk_ptr).tokencntl] = '\0';

    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_LTNODE;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_PAGE - Set thread 'page size' option          */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread 'page size'       */
/*    field in the thread definition work structure to the current  */
/*    token value. However, it first validates the token value      */
/*    reporting an error if it is out of range.             */
/*                                  */
/************************************************************************/
unsigned long SET_THR_PAGE (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is in range.
     */
    if (((*argblk_ptr).numberl < MIN_THR_PAGE) ||
            ((*argblk_ptr).numberl > MAX_THR_PAGE))
    {
        REPORT_ERROR (argblk_ptr, SCL_BADPAGE);
        goto end;
    }

    /*
     *  Set 'page size' field.
     */
    (*tdf_ptr).tdf_pagel = (*argblk_ptr).numberl;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_PAGE;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_PORT - Set thread 'port number' option            */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread 'port number'     */
/*    field in the thread definition work structure to the current  */
/*    token value. However, it first validates the token value      */
/*    reporting an error if it is out of range.             */
/*                                  */
/************************************************************************/
unsigned long SET_THR_PORT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is in range.
     */
    if (((*argblk_ptr).numberl < MIN_THR_PORT) ||
            ((*argblk_ptr).numberl > MAX_THR_PORT))
    {
        REPORT_ERROR (argblk_ptr, SCL_BADPORT);
        goto end;
    }

    /*
     *  Set 'port number' field.
     */
    (*tdf_ptr).tdf_tcpip_portl = (*argblk_ptr).numberl;
    (*tdf_ptr).tdf_default_flags2l &= ~TDF_THRDF2_TCPIP_PORT;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_PHYDEV - Set thread 'physical device' option      */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread 'physical device'     */
/*    field in the thread definition work structure to the current  */
/*    token value (minus the quotes). However, it first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_THR_PHYDEV (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */
#ifdef VMS
    long        jsub;      /* Loop control */
#endif /* VMS */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if (tpa_ptr->chr_strlenl > LEN_THR_PHYDEV)
    {
        REPORT_ERROR (argblk_ptr, SCL_PHYDEVTOOLONG);
        goto end;
    }

    /*
     *  Set physical device name field (excluding quotes).
     */
    (void) memcpy ((void *) (*tdf_ptr).tdf_physical_devicec36,
                   (void *) (*tpa_ptr).chr_stringc65536,
                   (*tpa_ptr).chr_strlenl);
    (*tdf_ptr).tdf_physical_devicec36[(*tpa_ptr).chr_strlenl] = '\0';

#ifdef VMS
    /*
     *  On VMS we convert it to uppercase.
     */
    for (jsub = 0; jsub < (*tpa_ptr).chr_strlenl; jsub++)
    {
        (*tdf_ptr).tdf_physical_devicec36 [jsub] =
            toupper ((*tdf_ptr).tdf_physical_devicec36 [jsub]);
    }
#endif /* VMS */

    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_PHYSICAL_DEVICE;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;
}


/************************************************************************/

/*    SET_THR_RECFIL - Set thread 'recording file' option       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread 'recording file'      */
/*    field in the thread definition work structure to the current  */
/*    token value (minus the quotes). However, it first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_THR_RECFIL (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */
#ifdef VMS
    long        jsub;      /* Loop control */
#endif /* VMS */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl > LEN_THR_RECFIL)
    {
        REPORT_ERROR (argblk_ptr, SCL_RECFILTOOLONG);
        goto end;
    }

    /*
     *  Set field in thread definition work record (excluding quotes).
     */
    (void) memcpy ((void *) (*tdf_ptr).tdf_recordfilec52,
                   (void *) (*tpa_ptr).chr_stringc65536,
                   (*tpa_ptr).chr_strlenl);
    (*tdf_ptr).tdf_recordfilec52[(*tpa_ptr).chr_strlenl] = '\0';

#ifdef VMS
    /*
     *  On VMS we convert it to uppercase.
     */
    for (jsub = 0; jsub < (*tpa_ptr).chr_strlenl; jsub++)
    {
        (*tdf_ptr).tdf_recordfilec52 [jsub] =
            toupper ((*tdf_ptr).tdf_recordfilec52 [jsub]);
    }
#endif /* VMS */

    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_RECORDFILE;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;
}


/************************************************************************/

/*    SET_THR_RECORD - Set thread 'record' option           */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'recording enabled/disabled' field   */
/*    in the thread definition work structure to the current token  */
/*    value.                                */
/*                                  */
/************************************************************************/
unsigned long SET_THR_RECORD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_recordl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_RECORD;

    return PSL_NORMAL;
}


/************************************************************************/

/*    SET_THR_RECOPT - Set thread 'recording options' option        */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread recording options     */
/*    field in the thread parameter definition work structure from the  */
/*    current token value. However, it first validates the value    */
/*    reporting an error if it is invalid.              */
/*                                  */
/************************************************************************/
unsigned long SET_THR_RECOPT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that this recording type has not already been specified.
     */
    if (((*tdf_ptr).tdf_record_typel & (*argblk_ptr).paraml) != 0)
    {
        REPORT_ERROR (argblk_ptr, PSL_DUPKEYWRD);
        goto end;
    }
    /*
     *  'Or' this option with the existing options in thread definition work record.
     */
    (*tdf_ptr).tdf_record_typel |= (*argblk_ptr).paraml;

    (*tdf_ptr).tdf_default_flags2l &= ~TDF_THRDF2_RECORD_OPTS;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_RECSTP - Set thread 'recording stop on detach' option */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = recording stop on detach flag       */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread 'recording stop on detach enabled'   */
/*    field in the thread definition work structure to the value passed */
/*    in the parameter field of the TPARSE argument block.      */
/*                                  */
/************************************************************************/
unsigned long SET_THR_RECSTP (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_stoprecdetl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_STOPRECDET;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_SRV - Set thread service option               */
/*    ===========                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread service name      */
/*    field in the thread definition work structure to the current  */
/*    token value. However, it first validates the token length     */
/*    reporting an error if it is too long.             */
/*                                  */
/************************************************************************/
unsigned long SET_THR_SRV (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*argblk_ptr).tokencntl  > LEN_THR_SRV)
    {
        REPORT_ERROR (argblk_ptr, SCL_SERVTOOLONG);
        goto end;
    }

    /*
     *  Set service name field.
     */
    (void) memcpy ((void *) (*tdf_ptr).tdf_ltservicec20,
                   (void *) (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);
    (*tdf_ptr).tdf_ltservicec20[(*argblk_ptr).tokencntl] = '\0';

    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_LTSERVICE;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_SOCKET - Set thread socket option             */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = socket connection flag          */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread socket connection enabled        */
/*    field in the thread definition work structure to the value passed */
/*    in the parameter field of the TPARSE argument block.      */
/*                                  */
/************************************************************************/
unsigned long SET_THR_SOCKET (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl     *thr_ptr;       /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_use_tcp_socketl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags2l &= ~TDF_THRDF2_USE_TCP_SOCKET;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_TRMID - Set thread terminal id option         */
/*    =============                                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = terminal id (TOF_THRTID_*)      */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread terminal id              */
/*    field in the thread definition work structure to the value passed */
/*    in the parameter field of the TPARSE argument block.      */
/*    It first validates the token value reporting an error if it is    */
/*    not valid for the current platform.               */
/*                                  */
/************************************************************************/
unsigned long SET_THR_TRMID (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  WYSE60 and HP terminal id's are UNIX-Specific.
     */
#ifdef VMS
    if (((*argblk_ptr).paraml == TOF_THRTID_WYSE60) ||
            ((*argblk_ptr).paraml == TOF_THRTID_HP))
    {
        REPORT_ERROR (argblk_ptr, PSL_INVKEYWRD);
        goto end;
    }
#endif

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_terminal_idl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_TERMINAL_ID;

    /*
     *  end:
     *  Routine exit point.
     */
#ifdef VMS
end:
#endif
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_TRMNAM - Set thread terminal name option          */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread terminal name field in the thread    */
/*    definition work structure to the current token value.     */
/*                                  */
/************************************************************************/
unsigned long SET_THR_TRMNAM (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set field in thread definition work record.
     */
    (void) memcpy ((void *) (*tdf_ptr).tdf_trm_namec20,
                   (void *) (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);
    (*tdf_ptr).tdf_trm_namec20[(*argblk_ptr).tokencntl] = '\0';
    /* Null-terminate */
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_TRM_NAME;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_TRMNAM_GEN - Set thread generated terminal name option    */
/*    ==================                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine clears the thread terminal name field in the thread  */
/*    definition work structure to indicate that the name is        */
/*    auto-generated.                           */
/*                                  */
/************************************************************************/
unsigned long SET_THR_TRMNAM_GEN (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set field in thread definition work record.
     */
    (void) strcpy ((*tdf_ptr).tdf_trm_namec20, "<AUTO-GENERATED>");
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_TRM_NAME;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_TRMINF - Set thread terminfo name option          */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread terminfo terminal name    */
/*    field in the thread definition work structure to the current  */
/*    token value (excluding quotes). It first validates the token  */
/*    length reporting an error if it is too long.          */
/*                                  */
/************************************************************************/
unsigned long SET_THR_TRMINF (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl > LEN_THR_TRMINF)
    {
        REPORT_ERROR (argblk_ptr, SCL_TRMINFTOOLONG);
        goto end;
    }

    /*
     *  Set field in thread definition work record.
     */
    (void) memcpy ((void *) (*tdf_ptr).tdf_terminfo_namec36,
                   (void *) (*tpa_ptr).chr_stringc65536,
                   (*tpa_ptr).chr_strlenl);
    (*tdf_ptr).tdf_terminfo_namec36[(*tpa_ptr).chr_strlenl] = '\0';

    (*tdf_ptr).tdf_default_flags2l &= ~TDF_THRDF2_TRMINFO;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_TRC - Set thread tracing option               */
/*    ===========                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = tracing flag                */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread tracing enabled          */
/*    field in the thread definition work structure to the value passed */
/*    in the parameter field of the TPARSE argument block.      */
/*                                  */
/************************************************************************/
unsigned long SET_THR_TRC (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_traceflagl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_TRACEFLAG;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_TRCFIL - Set thread trace file option         */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread trace file        */
/*    field in the thread definition work structure to the current  */
/*    token value (minus the quotes). However, it first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_THR_TRCFIL (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */
#ifdef VMS
    long        jsub;      /* Loop control */
#endif /* VMS */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl > LEN_THR_TRCFIL)
    {
        REPORT_ERROR (argblk_ptr, SCL_TRCFILTOOLONG);
        goto end;
    }

    /*
     *  Set field in thread definition work record (excluding quotes).
     */
    (void) memcpy ((void *) (*tdf_ptr).tdf_tracefilec52,
                   (void *) (*tpa_ptr).chr_stringc65536,
                   (*tpa_ptr).chr_strlenl);
    (*tdf_ptr).tdf_tracefilec52[(*tpa_ptr).chr_strlenl] = '\0';

#ifdef VMS
    /*
     *  On VMS convert it to uppercase.
     */
    for (jsub = 0; jsub < (*tpa_ptr).chr_strlenl; jsub++)
    {
        (*tdf_ptr).tdf_tracefilec52 [jsub] =
            toupper ((*tdf_ptr).tdf_tracefilec52 [jsub]);
    }
#endif /* VMS */

    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_TRACEFILE;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;
}


/************************************************************************/

/*    SET_THR_TRMTYP - Set thread terminal type option          */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = terminal type (TOF_THRTTP_*)    */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread terminal type            */
/*    field in the thread definition work structure to the value passed */
/*    in the parameter field of the TPARSE argument block.      */
/*    It first validates the token value reporting an error if it is    */
/*    not valid for the current platform.               */
/*                                  */
/************************************************************************/
unsigned long SET_THR_TRMTYP (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  PHYSICAL terminals are not supported on HP.
     */
#ifdef HP
    if ((*argblk_ptr).paraml == TOF_THRTTP_PHYS)
    {
        REPORT_ERROR (argblk_ptr, PSL_INVKEYWRD);
        goto end;
    }
#endif /* HP */

    /*
     *  REMOTE LAT, DECTERM & VWSTERM terminals are VMS-Specific.
     *
     *  IF WE ARE RUNNING ON WINDOWS IGNORE THESE TESTS AS WE MAY BE COMPILING
     *  FOR SYNTAX CHECKING ONLY AS THE TEST WILL BE RUN REMOTELY.
     */
#ifndef WIN32

#ifndef VMS
    if (((*argblk_ptr).paraml == TOF_THRTTP_REMLAT)  ||
            ((*argblk_ptr).paraml == TOF_THRTTP_DECTERM) ||
            ((*argblk_ptr).paraml == TOF_THRTTP_VWSTERM))
    {
        REPORT_ERROR (argblk_ptr, PSL_INVKEYWRD);
        goto end;
    }
#endif /* #ifndef VMS */

#endif /* #ifndef WIN32 */

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_terminal_typel = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_TERMINAL_TYPE;

    /*
     *  end:
     *  Routine exit point.
     */
#ifndef VMS
#ifndef WIN32
end:
#endif
#endif
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_TYPAHD - Set thread typeahead option          */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = typeahead flag              */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the thread typeahead enabled            */
/*    field in the thread definition work structure to the value passed */
/*    in the parameter field of the TPARSE argument block.      */
/*                                  */
/************************************************************************/
unsigned long SET_THR_TYPAHD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_typeaheadl = (*argblk_ptr).paraml;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_TYPEAHEAD;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_THR_USRNAM - Set thread username option           */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread username          */
/*    field in the thread definition work structure to the current  */
/*    token value (minus the quotes - if any). It first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_THR_USRNAM (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Handle both quoted and non-quoted usernames.
     *
     *  In each case check the username length before storing the username.
     */
    if ((*(*argblk_ptr).token_ptrpc == '\"') ||
            (*(*argblk_ptr).token_ptrpc == '\''))
    {                           /* Quoted Username */
        if ((*tpa_ptr).chr_strlenl > LEN_THR_USRNAM)
        {
            REPORT_ERROR (argblk_ptr, SCL_USRTOOLONG);
            goto end;
        }

        (void) memcpy ((void *) (*tdf_ptr).tdf_usernamec36,
                       (void *) (*tpa_ptr).chr_stringc65536,
                       (*tpa_ptr).chr_strlenl);
        (*tdf_ptr).tdf_usernamec36[(*tpa_ptr).chr_strlenl] = '\0';
    }
    else
    {                           /* UN-Quoted Username */
        if ((*argblk_ptr).tokencntl > LEN_THR_USRNAM)
        {
            REPORT_ERROR (argblk_ptr, SCL_USRTOOLONG);
            goto end;
        }

        (void) memcpy ((void *) (*tdf_ptr).tdf_usernamec36,
                       (void *) (*argblk_ptr).token_ptrpc,
                       (*argblk_ptr).tokencntl);
        (*tdf_ptr).tdf_usernamec36[(*argblk_ptr).tokencntl] = '\0';
    }

    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_USERNAME;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;
}


/************************************************************************/

/*    SET_THR_WIDTH - Set thread screen width option            */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the thread screen width      */
/*    field in the thread definition work structure to the current  */
/*    token value. However, it first validates the token value      */
/*    reporting an error if it is out of range.             */
/*                                  */
/************************************************************************/
unsigned long SET_THR_WIDTH (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   thrtbl    *thr_ptr;        /* Pointer to thread definition */
    struct   tdf        *tdf_ptr;       /* Pointer to terminal definition */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    thr_ptr = (*tpa_ptr).def_thr_wrkps;
    tdf_ptr = (*thr_ptr).thr_tdf_ptrps;

    /*
     *  Check that the token value is in range.
     */
    if (((*argblk_ptr).numberl < MIN_THR_WIDTH) ||
            ((*argblk_ptr).numberl > MAX_THR_WIDTH))
    {
        REPORT_ERROR (argblk_ptr, SCL_BADSCRWIDTH);
        goto end;
    }

    /*
     *  Set field in thread definition work record.
     */
    (*tdf_ptr).tdf_widthl = (*argblk_ptr).numberl;
    (*tdf_ptr).tdf_default_flags1l &= ~TDF_THRDF1_WIDTH;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_ABTONF - Set test 'abort on failure' option       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = abort on failure flag           */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the test abort on failure flag          */
/*    in the test parameter work structure to the value passed in   */
/*    the parameter field of the TPARSE argument block.         */
/*                                  */
/************************************************************************/
unsigned long SET_TST_ABTONF (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Set 'abort on failure' flag value.
     */
    (*prm_ptr).prm_abortfaill = (*argblk_ptr).paraml;
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_ABORTFAIL;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_ABTTST - Set test 'abort test' option         */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the test abort test          */
/*    field in the test definition work structure to the current    */
/*    token value (minus the quotes). However, it first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_TST_ABTTST (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl > LEN_TST_ABTTST)
    {
        REPORT_ERROR (argblk_ptr, SCL_ABRTTOOLONG);
        goto end;
    }

    /*
     *  Store the string
     */
    (void) memcpy ((void *) (*prm_ptr).prm_abrttestc84,
                   (void *) (*tpa_ptr).chr_stringc65536,
                   (*tpa_ptr).chr_strlenl);
    (*prm_ptr).prm_abrttestc84[(*tpa_ptr).chr_strlenl] = '\0';

    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_ABRTTEST;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_CMPTST - Set test 'complete test' option          */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the test complete test       */
/*    field in the test definition work structure to the current    */
/*    token value (minus the quotes). However, it first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_TST_CMPTST (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl > LEN_TST_CMPTST)
    {
        REPORT_ERROR (argblk_ptr, SCL_COMPTOOLONG);
        goto end;
    }

    /*
     *  Store the string
     */
    (void) memcpy ((void *) (*prm_ptr).prm_comptestc84,
                   (void *) ((*tpa_ptr).chr_stringc65536),
                   (*tpa_ptr).chr_strlenl);
    (*prm_ptr).prm_comptestc84[(*tpa_ptr).chr_strlenl] = '\0';

    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_COMPTEST;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_CNCTRM - Set test 'concurrent terminals' option       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the test concurrent terminals    */
/*    field in the test definition work structure to the current    */
/*    token value. However, it first validates the value reporting  */
/*    an error if it is invalid.                    */
/*                                  */
/************************************************************************/
unsigned long SET_TST_CNCTRM (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Check that the token value is in range.
     */
    if (((*argblk_ptr).numberl < MIN_TST_CNCTRM) ||
            ((*argblk_ptr).numberl > MAX_TST_CNCTRM))
    {
        REPORT_ERROR (argblk_ptr, SCL_BADCONTRM);
        goto end;
    }

    /*
     *  Store the value.
     */
    (*prm_ptr).prm_max_concurrentl = (*argblk_ptr).numberl;
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_MAX_CONCURRENT;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_DSC - Set test 'description' option           */
/*    ===========                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the test description         */
/*    field in the test definition work structure to the current    */
/*    token value (minus the quotes). However, it first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_TST_DSC (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   tlh        *tlh_ptr;       /* Pointer to test header */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    tlh_ptr = (*tst_ptr).tst_tlh_ptrps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl > LEN_TST_DSC)
    {
        REPORT_ERROR (argblk_ptr, PSL_DESCRTOOLONG);
        goto end;
    }

    /*
     *  Store the string.
     */
    (void) memcpy ((void *) (*tlh_ptr).tlh_testdescc50,
                   (void *) (*tpa_ptr).chr_stringc65536,
                   (*tpa_ptr).chr_strlenl);
    if ((*tpa_ptr).chr_strlenl < sizeof ((*tlh_ptr).tlh_testdescc50))
    {
        (*tlh_ptr).tlh_testdescc50[(*tpa_ptr).chr_strlenl] = '\0';
    }
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_DSC;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_DMPEXC - Set test 'dump on exception' option      */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = dump on exception flag          */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the test dump on exception flag         */
/*    in the test parameter work structure to the value passed in   */
/*    the parameter field of the TPARSE argument block.         */
/*                                  */
/************************************************************************/
unsigned long SET_TST_DMPEXC (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Set 'dump on exception' flag value.
     */
    (*prm_ptr).prm_dump_stylel = (*argblk_ptr).paraml;
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_DUMP_STYLE;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_FAILTC - Set test 'fail test-case' option         */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the test fail test-case      */
/*    field in the test definition work structure to the current    */
/*    token value (minus the quotes). However, it first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_TST_FAILTC (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl > LEN_TST_FAILTC)
    {
        REPORT_ERROR (argblk_ptr, SCL_FAILTOOLONG);
        goto end;
    }

    /*
     *  Store the string.
     */
    (void) memcpy ((void *) (*prm_ptr).prm_failtcc84,
                   (void *) (*tpa_ptr).chr_stringc65536,
                   (*tpa_ptr).chr_strlenl);
    (*prm_ptr).prm_failtcc84[(*tpa_ptr).chr_strlenl] = '\0';
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_FAILTC;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_HCDEV - Set test 'hardcopy device' option         */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the test hardcopy device     */
/*    field in the test definition work structure to the current    */
/*    token value (minus the quotes). However, it first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_TST_HCDEV (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */
#ifdef VMS
    long       jsub;       /* Loop control */
#endif /* VMS */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl > LEN_TST_HCDEV)
    {
        REPORT_ERROR (argblk_ptr, SCL_HCOPYTOOLONG);
        goto end;
    }

    /*
     *  Store the string.
     *  Note that we specifically allow a null string to be specified
     *  to erase the default value.
     */
    if ((*tpa_ptr).chr_strlenl == 0)            /* If empty string */
    {
        (*prm_ptr).prm_hardcopy_logc52 [0] = '\0';
        /* Clear default */
    }
    else
    {
        (void) memcpy ((void *) (*prm_ptr).prm_hardcopy_logc52,
                       (void *) (*tpa_ptr).chr_stringc65536,
                       (*tpa_ptr).chr_strlenl);
        (*prm_ptr).prm_hardcopy_logc52[(*tpa_ptr).chr_strlenl] = '\0';

#ifdef VMS
        /*
         *  On VMS we convert it to uppercase.
         */
        for (jsub = 0; jsub < (*tpa_ptr).chr_strlenl; jsub++)
        {
            (*prm_ptr).prm_hardcopy_logc52 [jsub] =
                toupper ((*prm_ptr).prm_hardcopy_logc52 [jsub]);
        }
#endif /* VMS */

    }

    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_HARDCOPY_LOG;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_IGNINF - Set test 'ignore informational' option       */
/*    ==============                                */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   tpaarg.paraml = ignore informational flag       */
/*        (LGN_NO = DISABLED, LGN_YES = ENABLED).       */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the test ignore informational flag      */
/*    in the test parameter work structure to the value passed in   */
/*    the parameter field of the TPARSE argument block.         */
/*                                  */
/************************************************************************/
unsigned long SET_TST_IGNINF (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Set 'ignore informational' flag value.
     */
    (*prm_ptr).prm_ignrinfl = (*argblk_ptr).paraml;
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_IGNRINF;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_INPRAT - Set test 'input rate' option         */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the test input rate          */
/*    field in the test definition work structure to the current    */
/*    token value. However, it first validates the value reporting an   */
/*    error if it is invalid.                       */
/*                                  */
/************************************************************************/
unsigned long SET_TST_INPRAT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */
    long       iret;       /* Routine return status */
    char       wrkbuf [128];   /* Work buffer */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Copy the token value to our local buffer for checking.
     */
    (void) memcpy ((void *) wrkbuf,
                   (void *) (*argblk_ptr).token_ptrpc,
                   (*argblk_ptr).tokencntl);

    wrkbuf [(*argblk_ptr).tokencntl] = '\0';
    /* Null terminate */
    /*
     *  Check that the value is valid.
     */
    iret = VTEST_INPRATE_CHK (wrkbuf);

    if (!(psl_test_status (iret)))
    {
        REPORT_ERROR (argblk_ptr, iret);
        goto end;
    }

    /*
     *  Store the value.
     */
    (void) strcpy ((*prm_ptr).prm_input_ratec8, wrkbuf);
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_INPUT_RATE;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_INPRAT_MAX - Set test 'input rate' option to 'maximum'    */
/*    ==================                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine sets the test input rate field in the test       */
/*    definition work structure to 'maximum'.               */
/*                                  */
/************************************************************************/
unsigned long SET_TST_INPRAT_MAX (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Set the input rate value to 'maximum'.
     */
    (*prm_ptr).prm_input_ratec8 [0] = '\0';
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_INPUT_RATE;

    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_PASSTC - Set test 'pass test-case' option         */
/*    ==============                            */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the test pass test-case      */
/*    field in the test definition work structure to the current    */
/*    token value (minus the quotes). However, it first validates the   */
/*    value reporting an error if it is invalid.            */
/*                                  */
/************************************************************************/
unsigned long SET_TST_PASSTC (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Check that the token value is not too long.
     */
    if ((*tpa_ptr).chr_strlenl > LEN_TST_PASSTC)
    {
        REPORT_ERROR (argblk_ptr, SCL_PASSTOOLONG);
        goto end;
    }

    /*
     *  Store the string.
     */
    (void) memcpy ((void *) (*prm_ptr).prm_passtcc84,
                   (void *) (*tpa_ptr).chr_stringc65536,
                   (*tpa_ptr).chr_strlenl);
    (*prm_ptr).prm_passtcc84[(*tpa_ptr).chr_strlenl] = '\0';
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_PASSTC;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_TST_WKSQ - Set test 'working set quota' option        */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged.                      */
/*    Inputs:   None.                           */
/*    Outputs:  None.                           */
/*                                  */
/*    This routine attempts to set the test working set quota       */
/*    field in the test definition work structure to the current    */
/*    token value. However, it first validates the value reporting  */
/*    an error if it is invalid.                    */
/*                                  */
/************************************************************************/
unsigned long SET_TST_WKSQ (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   scl_com    *com_ptr;       /* Common data */
    struct   scl_tpa    *tpa_ptr;       /* Pointer to TPARSE data */
    struct   tsttbl     *tst_ptr;       /* Pointer to test definition */
    struct   prm        *prm_ptr;       /* Pointer to test parameters */

    /*
     *  Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;
    tst_ptr = (*tpa_ptr).def_tst_wrkps;
    prm_ptr = (*tst_ptr).tst_prm_ptrps;

    /*
     *  Check that the token value is in range.
     */
    if ((*argblk_ptr).numberl <= 0)
    {
        REPORT_ERROR (argblk_ptr, SCL_BADWKSQ);
        goto end;
    }

    /*
     *  Store the value.
     */
    (*prm_ptr).prm_wsquotal = (*argblk_ptr).numberl;
    (*prm_ptr).prm_default_flagsl &= ~PRM_TSTDF_WSQUOTA;

    /*
     *  end:
     *  Routine exit point.
     */
end:
    return PSL_NORMAL;          /* Return success */
}


/************************************************************************/

/*    SET_UNKNOWN_STR - Set flag indicating unknown string wait         */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   PSLTPA$L_PARAM(AP) = Terminal Type          */
/*    Outputs:  None                            */
/*                                  */
/*    This routine sets a flag to indicate waiting for unknown string   */
/*                                  */
/************************************************************************/


unsigned long   SET_UNKNOWN_STR (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     * Set the flag = 1
     */

    (*tpa_ptr).unknown_strl = 1;

    return PSL_NORMAL;      /* ...and return success    */
}
/*
 * 
 */

/************************************************************************/

/*    SET_VAR_FILE - Save the file variable 'file name' string      */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Argument block                      */
/*    Outputs:  tpa_ptr->def_filnamc256 set with filename string    */
/*                                  */
/*    This routine sets the filename string for a file variable into    */
/*    the variables parsing data.                   */
/*                                  */
/*======================================================================*/
unsigned long SET_VAR_FILE (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if the filename string length is in bounds. If it is OK
     *    save the filename as a NULL terminated string, otherwise report an
     *    error.
     */
    if ((*tpa_ptr).chr_strlenl < sizeof ((*tpa_ptr).def_filnamc256))
    {
        (void) memcpy ((*tpa_ptr).def_filnamc256
                       ,(*tpa_ptr).chr_stringc65536
                       ,(*tpa_ptr).chr_strlenl);

        (*tpa_ptr).def_filnamc256[(*tpa_ptr).chr_strlenl] = '\0';
    }
    else
    {
        REPORT_ERROR (argblk_ptr, SCL_FILNAMTOOLONG);
    }

    return PSL_NORMAL;          /* ...and return success */
}


/************************************************************************/
/*                                  */
/*    SET_VAR_PROMPT - Save the variable prompt string.         */
/*    ===============                               */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   Argument block                      */
/*    Outputs:  Prompt string saved in DEF variables            */
/*      DEF_PROMPTC50 set with prompt string            */
/*      DEF_PLENGTHL = Length of prompt string (max 50)     */
/*                                  */
/*    This routine sets the prompt string for a variable into the   */
/*    variables parsing data.                       */
/*                                  */
/*======================================================================*/

unsigned long   SET_VAR_PROMPT (struct tpaarg *argblk_ptr)

{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data          */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Check to see if the prompt string is 50 characters or less and reject it
     *    if it is too long.
     */
    if ((*tpa_ptr).chr_strlenl > 50)    /* If prompt string > 50 chars  */
    {
        REPORT_ERROR (argblk_ptr, VTEST_PROMPTTOOLONG);
        /* Report prompt string too long */
    }

    else                /* ...else if string O.k.   */
    {

        /*
         *    Move the prompt string into the DEF variables until it is copied into
         *    tye variable symbol table.
         */
        (void) memset ((*tpa_ptr).def_promptc50
                       ,' '
                       ,sizeof ((*tpa_ptr).def_promptc50));
        /* Space fill the prompt buffer */

        (void) memcpy ((*tpa_ptr).def_promptc50
                       ,(*tpa_ptr).chr_stringc65536
                       ,(*tpa_ptr).chr_strlenl);
        /* ...before copying prompt string */

        (*tpa_ptr).def_plengthl = (*tpa_ptr).chr_strlenl;
    }

    return PSL_NORMAL;          /* ...and return success    */
}

/************************************************************************/

/*    SET_WAIT_UNIT                         */
/*    =============                             */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   TPARSE argument block                               */
/*          argblk_ptr->paraml = WAIT_SECONDS/MILLISECONDS.     */
/*    Outputs:  tpa_ptr->wait_unitl set from parameter.         */
/*                                  */
/*    This routine sets the script wait unit field in the SCL       */
/*    TPARSE structure to the value passed in the parameter field   */
/*    of the argument block.                        */
/*                                  */
/*======================================================================*/
unsigned long SET_WAIT_UNIT (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct   scl_com *com_ptr;      /* Common Data */
    struct   scl_tpa *tpa_ptr;      /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *    Set script wait unit.
     */
    tpa_ptr->wait_unitl = argblk_ptr->paraml;

    return PSL_NORMAL;                  /* ...and return success */
}

/************************************************************************/
/*                                  */
/*    SET_WITH_CNT - Set HTTP Request WITH Value Count to 1     */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  tpa_ptr->http_withval_cntl set to 1         */
/*                                  */
/*    This routine is called to set the value count for the HTTP    */
/*    request WITH clause that has just been parsed to 1. It is     */
/*    called when a single value - rather than a value list - has   */
/*    been parsed.                          */
/*                                  */
/************************************************************************/
unsigned long SET_WITH_CNT (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *    Set the value count.
     */
    tpa_ptr->http_withval_cntl = 1;

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    SET_WOUT_CNT - Set HTTP Request WITHOUT Value Count to 1      */
/*    ============                              */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   Always succeeds (PSL_NORMAL)                */
/*    Pointers: Unchanged                       */
/*    Inputs:   None.                           */
/*    Outputs:  tpa_ptr->http_woutval_cntl set to 1         */
/*                                  */
/*    This routine is called to set the value count for the HTTP    */
/*    request WITHOUT clause that has just been parsed to 1. It is  */
/*    called when a single value - rather than a value list - has   */
/*    been parsed.                          */
/*                                  */
/************************************************************************/
unsigned long SET_WOUT_CNT (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) argblk_ptr->user_ptrp;
    tpa_ptr = com_ptr->tpa_ptrps;

    /*
     *    Set the value count.
     */
    tpa_ptr->http_woutval_cntl = 1;

    return PSL_NORMAL;          /* Return success */
}

/************************************************************************/
/*                                  */
/*    SET_TC_GOTO - Flag that the testcase GOTO clause was specified    */
/*    ===========                           */
/*                                          */
/*    Calling Format:                           */
/*  iret = SET_TC_GOTO (argblk_ptr)                 */
/*                                  */
/*    Type: Action routine                      */
/*    Status:   PSL_NORMAL always                   */
/*    Pointers: Unchanged                       */
/*    Inputs:   None                                */
/*    Outputs:  tpa.testcase_flagsl                 */
/*                                  */
/************************************************************************/

unsigned long   SET_TC_GOTO (struct tpaarg *argblk_ptr)
{
    /*
     *    Local data
     */
    struct scl_com  *com_ptr;       /* Common Data          */
    struct scl_tpa  *tpa_ptr;       /* TPARSE Data          */

    /*
     *    Copy pointers for ease of reference.
     */
    com_ptr = (struct scl_com *) (*argblk_ptr).user_ptrp;
    tpa_ptr = (*com_ptr).tpa_ptrps;

    /*
     *    Output tracing information if required.
     */
#ifdef PSL_TRACING
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Setting flag for TEST-CASE GOTO\n");
    }
#endif

    /*
     *    If this is a START TEST-CASE command, and a test-case is already active
     *    return a failure status.
     */
    tpa_ptr->testcase_flagsl |= TCAS_GOTO_TC;

    /*
     *    Return success.
     */
    return PSL_NORMAL;
}

