/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// AtvfRecord.cpp
/**
 * Define the Methods for the AtvfRecord class
 **/

// Local includes
#include "AtvfRecord.h"

/**
 * Constructor for unknown records.
 **/
AtvfRecord::AtvfRecord(string* valP)
        :   _tokP(NULL),
        _valP(valP),
        _ruleP(NULL)
{}

/**
 * Constructor for known records.
 **/
AtvfRecord::AtvfRecord(const SharedPtr<AtvfRecRule>& ruleP, string* tokP)
        :   _tokP(tokP),
        _valP(NULL),
        _ruleP(ruleP)
{}

/**
 * Destructor
 **/
AtvfRecord::~AtvfRecord()
{
    delete _tokP;
    delete _valP;
}

/**
 * Accessor to tell if record is a known valid one, ie. has a rule.
 * Blank records are not known even though they are meaningful ???
 **/
bool AtvfRecord::isKnown() const
{
    return !_ruleP.IsNull();
}

/**
 * Accessor to find the string value of this record
 **/
string AtvfRecord::value() const
{
    string str;
    // If no value simply return empty string
    if(_valP == NULL) return str;
    if((_ruleP.IsNull()) || _ruleP->leadWSpace())
    {
        str = *_valP;
    }
    else if(_valP->length() != 0)
    {
        string::size_type pos = _valP->find_first_not_of(" \t");
        if(pos != string::npos)
        {
            str = _valP->substr(pos);
        }
    }
    return str;
}

/**
 * Accessor to get a pointer to the rule associated with this record
 **/
SharedPtr<AtvfRecRule> AtvfRecord::rule() const
{
    return _ruleP;
}

/**
 * Accessor to get the key that represents this record,
 * return invalid key if this record isn't known
 **/
AtvfKey AtvfRecord::key() const
{
    if(isKnown())
    {
        return _ruleP->key();
    }
    return AtvfKey("");
}

/**
 * Accessor to tell if record represents blank line
 **/
bool AtvfRecord::isBlank() const
{
    if(!isKnown() && // If record is not a known one and
            ((_valP == NULL) || // value doesn't exist or is all whitespace
             (_valP->find_first_not_of(" \t") == string::npos)))
    {
        return true; // It's blank
    }

    return false; // Otherwise it's not blank
}

/**
 * Accessor to output, to the provided stream, the contents of this
 * record in its correct file based format.
 **/
void AtvfRecord::save(ostream& out) const
{
    if(isKnown())
    {
        out << ((_tokP != NULL)?(*_tokP):(_ruleP->token()));
        if((_valP != NULL) && _ruleP->hasValue()) _ruleP->saveValue(*_valP, out);
    }
    else
    { // Unknown record type - just output value
        if(_valP != NULL) out << *_valP;
    }
    out << endl;
}

/**
 * Mutator to change the string value of this record.
 **/
void AtvfRecord::value(const string& newVal)
{
    if(_valP == NULL)
    {
        _valP = new string();
    }
    else
    {
        _valP->resize(0);
    }

    if((_ruleP.IsNull()) || _ruleP->leadWSpace())
    {
        _valP->assign(newVal);
    }
    else
    {
        string::size_type pos = newVal.find_first_not_of(" \t");
        if(pos != string::npos)
        {
            _valP->assign(newVal.substr(pos));
        }
    }
}

/**
 * Mutator to parse a value from an Input stream.
 * Use the rule as the record doesn't really know how to do this.
 * If its an unkown record type we ignore for now
 **/
void AtvfRecord::getValue(AtvfInput& in)
{
    if(isKnown())
    {
        delete _valP;
        _valP = _ruleP->parseInput(in);
    }
}

/**
 * Define operator << to allow AtvfRecord to be output.
 **/
ostream& operator<<(ostream& os, const AtvfRecord& r)
{
    if(r.isKnown())
    {
        os << r.rule();
        if((r.rule())->hasValue())
        {
            os << " - " << r.value();
        }
    }
    else if(r.isBlank())
    {
        os << "Blank";
    }
    else
    {
        os << "Unknown - " << r.value();
    }
    return os;
}
