/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CECOperand.h
// @module  CECOperand class | This module defines the interface to the
//          CECOperand class.
///////////////////////////////////////////////////////////////////////////////

#if !defined (CYR_CECOPERAND)
#define CYR_CECOPERAND

#if defined(WIN32)
# if _MSC_VER >= 1200   // MSC version 6 and above
#pragma warning(disable:4786)
# endif
#endif

#include <psl_macros.h>
#include <string>
#include <list>
#include <vector>
#include <stack>

// Operand addressing modes.
enum AddrMode {addr_none, addr_immediate, addr_operator, addr_variable,
               addr_function};

// Operand data types.
enum DataType {dtyp_none, dtyp_integer, dtyp_character};

// Codes representing operators.
enum Operator {opr_none, opr_not, opr_and, opr_or, opr_eq, opr_ne, opr_lt, opr_le,
               opr_gt, opr_ge, opr_contains, opr_not_contains, opr_add, opr_sub,
               opr_mul, opr_div, opr_mod};

// Codes representing functions.
enum FncCode {fnc_none, fnc_locate, fnc_extract};

///////////////////////////////////////////////////////////////////////////////
// @class   Operand | Intrnal class to hold operand as either a PSL_LONGWORD_T
//			or a string expression string in SCL format.
//
class CECOperand
{
    // @access Public Members.
public:
    // @cmember Constructor to create a 'NULL' operand.
    CECOperand();
    // @cmember Constructor to create an operand from a function code.
    CECOperand(const FncCode fnccod);
    // @cmember Constructor to create an operand from an operator.
    CECOperand(const Operator opr);
    // @cmember Constructor to create an operand from a string value.
    CECOperand(const std::string &strval, const AddrMode addr);
    // @cmember Constructor to create a variable operand.
    CECOperand(const std::string &strval, const DataType dtyp);
    // @cmember Constructor to create a variable operand.
    CECOperand(const std::string &strval, const DataType dtyp, const long vardims);
    // @cmember Constructor to create an operand from an integer value.
    CECOperand(const long intval);
    // @cmember CECOperand copy constructor.
    CECOperand(const CECOperand& opd);

    CECOperand& operator=(const CECOperand&);

    // @cmember Destructor.
    virtual ~CECOperand()   {}

    // @cmember Accessor function to retrieve address mode.
    AddrMode Addrm()        const   { return _addrM; }
    // @cmember Accessor function to retrieve character value as a string.
    const std::string &ChrVal()  const   { return _chrValP; }
    // @cmember Accessor function to retrieve data type.
    DataType DTyp()         const   { return _dTyp; }
    // @cmember Accessor function to retrieve function code.
    FncCode FncCod()        const   { return _fncCod; }
    // @cmember Accessor function to retrieve integer value.
    long IntVal()           const   { return _intVal; }
    // @cmember Return 'true' if character operand, 'false' otherwise.
    bool IsChr()            const   { return _dTyp == dtyp_character; }
    // @cmember Return 'true' if integer operand, 'false' otherwise.
    bool IsInt()            const   { return _dTyp == dtyp_integer; }
    // @cmember Accessor function to retrieve operator type.
    Operator Opr()          const   { return _opr; }
    // @cmember Get string listing operand content.
    std::string GetContent();
    // @cmember Accessor function to retrieve variable name as a string.
    const std::string &VarNam()  const   { return _varNamP; }
    // @cmember Accessor function to retrieve variable dimensions.
    long VarDims()          const   { return _varDims; }

    // @access Private Members.
private:
    // @cmember Addressing mode.
    AddrMode _addrM;
    // @cmember Data type.
    DataType _dTyp;
    // @cmember Value of immediate character operand as character array.
    // N.B.: This field MAY contain NULL characters - therefore we save its
    // length. A terminating NULL character is appended, BUT this is NOT
    // part of the data.
    std::string _chrValP;
    // @cmember Value of immediate integer operand.
    long _intVal;
    // @cmember Operator for operator operand.
    Operator _opr;
    // @cmember Name of variable for variable operand as character array.
    std::string _varNamP;
    // @cmember Number of array dimensions for variable operand.
    long _varDims;
    // @cmember Function code for function operand.
    FncCode _fncCod;
};





///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor to create a NULL operand object.
//
__inline CECOperand::CECOperand() :
        _addrM(addr_none), _dTyp(dtyp_none),
        _fncCod(fnc_none), _intVal(0), _opr(opr_none), _varDims(0)
{} // End CECOperand()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor to create an operand object from a function code.
//
// @parm    Function code.
//
__inline CECOperand::CECOperand(const FncCode fnccod) :
        _addrM(addr_function), _dTyp(dtyp_none),
        _fncCod(fnccod), _intVal(0), _opr(opr_none), _varDims(0)
{
    switch (_fncCod)
    {
    case fnc_extract:
        _dTyp = dtyp_character;
        break;
    case fnc_locate:
        _dTyp = dtyp_integer;
        break;
    default:
        _dTyp = dtyp_none;
    }
    return;
} // End CECOperand()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor to create an operand object from an operator.
//
// @parm    Operator.
//
__inline CECOperand::CECOperand(const Operator opr) :
        _addrM(addr_operator), _dTyp(dtyp_none),
        _fncCod(fnc_none), _intVal(0), _opr(opr), _varDims(0)
{} // End CECOperand()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor to create an operand object from an immediate integer
//			value.
//
// @parm    String from which operand to be created.
// @parm    Addressing mode for operand.
//

__inline CECOperand::CECOperand(const std::string &strval, const AddrMode addrm)
{
    if (addrm == addr_variable)		// Set variable operand specific data
    {
        _addrM = addrm;
        _dTyp = dtyp_none;
        _varNamP = strval;
    }
    else if (addrm == addr_immediate)	// Set Immediate operand specific data
    {
        _addrM = addrm;
        _dTyp = dtyp_character;
        _chrValP = strval;
    }
    else				// Set Null operand specific data
    {
        _addrM = addr_none;
        _dTyp = dtyp_none;
    }
    // Set remaining 'common' data.
    _intVal = 0;
    _opr = opr_none;
    _fncCod = fnc_none;
    _varDims = 0;
} // End CECOperand()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor to create a variable operand object from a string
//			supplying the variable name, the data type and the number of
//			array dimensions (zero if not an array).
//
// @parm    Variable name for new variable operand.
// @parm    Data type of new variable operand.
//
__inline CECOperand::CECOperand(const std::string &strval, const DataType dtyp, const long vardims) :
    _addrM(addr_variable), _dTyp(dtyp), _fncCod(fnc_none),
    _intVal(0), _opr(opr_none), _varDims(vardims),
    _varNamP(strval)
{} // End CECOperand()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor to create an operand object from an immediate integer
//			value.
//
// @parm    Integer value from which to create operand.
//
__inline CECOperand::CECOperand(const long intval) :
        _addrM(addr_immediate), _dTyp(dtyp_integer), _fncCod(fnc_none),
        _intVal(intval), _opr(opr_none), _varDims(0)
{} // End CECOperand()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   CECOperand copy constructor. This constructor creates an operand
//			object from another operand object. It does this by copying the
//			member data from the supplied object to the new object.
//
// @parm    CECOperand object from which to create operand.
//

__inline CECOperand::CECOperand(const CECOperand& opd) :
        _addrM(opd._addrM), _dTyp(opd._dTyp), _fncCod(opd._fncCod),
        _intVal(opd._intVal), _opr(opd._opr), _varDims(opd._varDims),
        _chrValP(opd._chrValP), _varNamP(opd._varNamP)
{}

__inline CECOperand& CECOperand::operator=(const CECOperand& opd)
{
    _addrM  = opd._addrM;
    _dTyp   = opd._dTyp;
    _chrValP= opd._chrValP;
    _intVal = opd._intVal;
    _opr    = opd._opr;
    _varNamP= opd._varNamP;
    _varDims= opd._varDims;
    _fncCod = opd._fncCod;

    return *this;
}

#endif // !defined(CYR_CECOPERAND)
