/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

////////////////////////////////////////////////////////////////////////////////
// @doc	    TInterRecords.cpp
// @module  TInterRecords |TRecords that require full integration with Architecture
///////////////////////////////////////////////////////////////////////////////

#ifdef WIN32
#   pragma warning(disable:4786)
#   pragma warning(disable:4503)
#endif
#include "TInterRecords.inl"

// *****************************************************************************
// * TIMERS
// *****************************************************************************

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Start Timer Record
// =====================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdStimer::execute(TContext& ctxt)
{

    ctxt.set_phase_text(PHASE_STIMER);

    string  l_name = strPrmVal(TParam::stName, ctxt);

    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("START TIMER: " + l_name, TRACEMSG);

    ctxt.StartTimer(l_name);

    ctxt.set_phase_text(TP_PHASE_NONE);
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for End Timer Record
// ===================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdEtimer::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ETIMER);

    string  l_name = strPrmVal(TParam::stName, ctxt);

    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("STOP TIMER: " + l_name, TRACEMSG);

    ctxt.StopTimer(l_name);
    ctxt.set_phase_text(TP_PHASE_NONE);
    return 0;
}
///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Acquire Mutex Record
// =======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc

#ifndef __DLLSA__
#ifndef TH_POOL_VER

int TRcrdAcmutex::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ACQ_MUTEX);
    unsigned	l_scope	    = intPrmVal(TParam::stScope, ctxt);
    string	s_name	    = strPrmVal(TParam::stName, ctxt);
    string	s_timeout   = strPrmVal(TParam::stTimeOut, ctxt);
    float i_timeout=atoi(s_timeout.c_str())/1000;//Granularity of seconds
    bool timed_out=false;
    bool error=false;
    bool can_time_out=(s_timeout=="")?false:true;
    time_t stime;//start time
    time_t ctime;//current time
    time(&stime);
    bool have_acquired=false;
    try
    {
        while(!have_acquired && !timed_out )//loop until acquired or timed out
        {
            have_acquired=ctxt.LockMutex(s_name,l_scope);
            time(&ctime);
            if(((ctime-stime)>i_timeout)&& can_time_out)timed_out=true;//timeout
            ctxt.delay(500);//wait a half second
        }

    }
    catch(...)
    {
        error=true;//have caught an error that is not a timeout
    }

    int dest=0;
    if(timed_out)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto then try errlab else Execute the Goto
        {
            labelIndex=PrepareGoto(this,TParam::stTmolab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            error=true;
        }
        if(!error)dest=ExecuteGoto(this,ctxt,TParam::stTmolab,&labelIndex);
    }
    if(error)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException("Unable to Acquire Mutex");
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;

}// end of TRcrdAcmutex::execute

#else   // if defined TH_POOL_VER

int TRcrdAcmutex::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ACQ_MUTEX);

    CAsyncVirtualUser& VU = static_cast<CAsyncVirtualUser&> (ctxt);
    BOOL bTimedOut  = FALSE;
    BOOL bError     = FALSE;

    if(VU.GetTrace() && VU.GetTrace()->IsTrace("full"))
    {
        GetTraceStream(VU.GetTrace()) << "TRcrdAcmutex: entering step " << VU.GetTRecordCallStep() << ".\r\n";
        ReleaseTraceStream();
    }

    switch(VU.GetTRecordCallStep())
    {
    case 0: // Initialise the values in the Virtual user structure.
        {
            string s_name = strPrmVal(TParam::stName, ctxt);
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("ACQUIRE MUTEX: Synchronizing on the " + s_name + " mutex.", TRACEMSG);

            time_t stime;
            time(&stime);

            VU.SetMutexStartWaitTime(stime);
        }
        // No break since should continue.

    case 1: // Must try or retry to acquire the mutex.
        {
            unsigned    l_scope	    = intPrmVal(TParam::stScope, ctxt);
            string	    s_name	    = strPrmVal(TParam::stName, ctxt);

            BOOL bAcquired = FALSE;

            try
            {
                bAcquired = VU.LockMutex(s_name, l_scope);
            }
            catch(...)
            {
                // Caught an error that is not a timeout.
                bError = TRUE;
            }

            if(bAcquired == FALSE)
            {
                time_t ctime;  // Current time
                time(&ctime);

                string  s_timeout = strPrmVal(TParam::stTimeOut, ctxt);
                int     i_timeout = atoi(s_timeout.c_str())/1000;   // Granularity of seconds

                if(s_timeout.empty()==TRUE || (s_timeout.empty()==FALSE && ctime-VU.GetMutexStartWaitTime()<i_timeout))
                {
                    // Wait 500 milli-seconds before trying again to lock the mutex.
                    VU.SetTRecordCallStep(1);   // Will re-enter the "case 1" section next time execute is called.
                    VU.StartEventTimer(500);
                    return TOF_RES_ASYNC_STARTED;
                }
                else
                {
                    bTimedOut = TRUE;

                    if(ctxt.GetVerbose(TRACEMSG))
                        ctxt.log("ACQUIRE MUTEX: Timed-out synchronizing on the " + s_name + " mutex.", TRACEMSG);
                }
            }
            else
            {
                if(ctxt.GetVerbose(TRACEMSG))
                    ctxt.log("ACQUIRE MUTEX: Successfully synchronized on the " + s_name + " mutex.", TRACEMSG);
            }

            // Managed to acquire the mutex or timed-out.

            int dest = TOF_RES_OK;
            if(bTimedOut)
            {
                int labelIndex;
                try // Prepare the goto if there is no goto then try errlab else Execute the Goto
                {
                    labelIndex = PrepareGoto(this, TParam::stTmolab, ctxt.Script(), TRUE);
                }
                catch(...)//no valid goto statement
                {
                    bError = TRUE;
                }

                if(bError == FALSE)
                    dest = ExecuteGoto(this, ctxt, TParam::stTmolab, &labelIndex);
            }

            if(bError)
            {
                int labelIndex;
                try
                {
                    // Prepare the goto. If there is no goto throw to ON error goto.
                    labelIndex = PrepareGoto(this, TParam::stErrlab, ctxt.Script(), TRUE);
                }
                catch(...)//no valid goto statement
                {
                    throw mdbParseException("Unable to Acquire Mutex");
                }

                dest = ExecuteGoto(this, ctxt, TParam::stErrlab, &labelIndex);
            }

            ctxt.set_phase_text(TP_PHASE_NONE);
            return dest;
        }

    default:
        // TODO: error.
        assert(0);
        return TOF_RES_FAILED;
    };
}// end of TRcrdAcmutex::execute

#endif  // #ifdef TH_POOL_VER


///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Release Mutex Record
// =======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdRlmutex::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_REL_MUTEX);
    unsigned int    l_scope	= intPrmVal(TParam::stScope, ctxt);
    string	    s_name	= strPrmVal(TParam::stName, ctxt);
    string	    s_errlab    = strPrmVal(TParam::stErrlab, ctxt);
    bool is_unlocked=false;
    try
    {
#ifndef TH_POOL_VER
        is_unlocked = ctxt.UnLockMutex(s_name,l_scope);
#else
        is_unlocked = static_cast<CAsyncVirtualUser&> (ctxt).UnLockMutex(s_name, l_scope);
#endif
    }
    catch(...)
    {
        //treat all errors as cannot unlock
        is_unlocked=false;
    }
    int dest=0;
    if(is_unlocked==false)
    {
        if(s_errlab.empty())
            throw mdbParseException("Unable to unlock Mutex");
        int _labelIndex = ctxt.GetLabelIndex(s_errlab);
        dest=goto_destination(_labelIndex,ctxt);//jump to error label
    }

    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;
}// end of TRcrdRlmutex::execute

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Set Semaphore
// ================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdStsemaphore::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_SET_SEMAPHORE);

    unsigned int	l_scope		= intPrmVal(TParam::stScope, ctxt);
    string		    s_name		= strPrmVal(TParam::stName, ctxt);

    bool is_set;

    {
        CECSemaphore semaphore(s_name,l_scope);
        is_set=false;
        try
        {
            is_set=semaphore.release();
        }
        catch(...)
        {
            is_set=false;
        }
    }   // End of semaphore scope.

    int dest=0;
    if(is_set==false)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException("Unable to Set sempahore");
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }

    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;
}// end of TRcrdStsemaphore::execute


///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Wtsemaphore Record
// =====================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdWtsemaphore::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_WAIT_SEMAPHORE);

#ifndef TH_POOL_VER
    unsigned int	l_scope		= intPrmVal(TParam::stScope, ctxt);
    string		s_name		= strPrmVal(TParam::stName, ctxt);
    string		s_timeout	= strPrmVal(TParam::stTimeOut, ctxt);
    float i_timeout=atoi(s_timeout.c_str())/1000;//Granularity of seconds
    bool timed_out=false;
    bool error=false;
    bool can_time_out=(s_timeout=="")?false:true;
    time_t stime;//start time
    time_t ctime;//current time
    time(&stime);
    CECSemaphore* semaphore=new CECSemaphore(s_name,l_scope);
    bool have_waited=false;
    try
    {
        while(!have_waited && !timed_out )//loop until acquired or timed out
        {
            have_waited=semaphore->tryacquire();//try and acquire semaphore, doesn't actually acquire it
            time(&ctime);
            if(((ctime-stime)>i_timeout)&& can_time_out)timed_out=true;//timeout
            ctxt.delay(500);//wait a half second
        }

    }
    catch(...)
    {
        error=true;//have caught an error that is not a timeout
    }

    delete semaphore;
    int dest=0;
    if(timed_out)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto then try errlab else Execute the Goto
        {
            labelIndex=PrepareGoto(this,TParam::stTmolab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            error=true;
        }
        if(!error)dest=ExecuteGoto(this,ctxt,TParam::stTmolab,&labelIndex);
    }
    if(error)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException("Timed out Waiting for Sempahore");
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;

#else   // TH_POOL_VER case.

    CAsyncVirtualUser& VU = static_cast<CAsyncVirtualUser&> (ctxt);

    BOOL bTimedOut  = FALSE;
    BOOL bError     = FALSE;

    switch(VU.GetTRecordCallStep())
    {
    case 0: // Initialise the values in the Virtual user structure.
        {
            time_t stime;
            time(&stime);

            VU.SetSemaphoreStartWaitTime(stime);
        }
        // No break since should continue.

    case 1:
        {
            // Try (again) to acquire the semaphore.
            unsigned 	l_scope	= intPrmVal(TParam::stScope, ctxt);
            string	    s_name  = strPrmVal(TParam::stName, ctxt);

            try
            {
                if(VU.TryAcquireSemaphore(s_name, l_scope) == FALSE)    // Try and acquire semaphore, doesn't actually acquire it.
                {
                    time_t ctime;
                    time(&ctime);

                    string  s_timeout = strPrmVal(TParam::stTimeOut, ctxt);
                    int     i_timeout = atoi(s_timeout.c_str())/1000; //Granularity of seconds

                    if(s_timeout.empty()==TRUE || (s_timeout.empty()==FALSE && ctime-VU.GetSemaphoreStartWaitTime()<i_timeout))
                    {
                        // Wait 500 milli-seconds before trying again to lock the mutex.
                        VU.SetTRecordCallStep(1);   // Will re-enter the "case 1" section next time execute is called.
                        VU.StartEventTimer(500);
                        return TOF_RES_ASYNC_STARTED;
                    }
                    else
                    {
                        bTimedOut = TRUE;
                    }
                }
            }
            catch(...)
            {
                bError = TRUE; // Have caught an error that is not a timeout.
            }

            // Managed to acquire the semaphore.


            int dest = 0;
            if(bTimedOut)
            {
                int labelIndex;
                try
                {
                    // Prepare the goto. If there is no goto then try errlab else Execute the Goto.
                    labelIndex = PrepareGoto(this, TParam::stTmolab, ctxt.Script(), TRUE);
                }
                catch(...)//no valid goto statement
                {
                    bError = TRUE;
                }

                if(!bError)
                    dest = ExecuteGoto(this, ctxt, TParam::stTmolab, &labelIndex);
            }

            if(bError)
            {
                int labelIndex;
                try
                {
                    // Prepare the goto. If there is no goto throw to ON error goto.
                    labelIndex = PrepareGoto(this, TParam::stErrlab, ctxt.Script(), TRUE);
                }
                catch(...)
                {
                    // No valid goto statement.
                    throw mdbParseException("Timed out Waiting for Sempahore");
                }
                dest = ExecuteGoto(this, ctxt, TParam::stErrlab, &labelIndex);
            }
            ctxt.set_phase_text(TP_PHASE_NONE);
            return dest;
        }

    default:
        {
            assert(0);
            return TOF_RES_FAILED;
        }
    };

    return TOF_RES_FAILED;
#endif
}// end of TRcrdWtsemaphore::execute

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Clsemaphore Record
// ==============================
///////////////////////////////////////////////////////////////////////////////
// @mfunc

// TODO: this is a blocking object. Try to find a way to make it asynchronous.
int TRcrdClsemaphore::execute(TContext& ctxt)
{
#ifndef CYRANO_TP_STANDALONE

    ctxt.set_phase_text(PHASE_CLR_SEMAPHORE);
    unsigned int        l_scope		= intPrmVal(TParam::stScope, ctxt);
    string		s_name		= strPrmVal(TParam::stName, ctxt);
    CECSemaphore* semaphore=new CECSemaphore(s_name,l_scope);
    bool is_cleared=false;
    try
    {
        is_cleared=semaphore->acquire();
    }
    catch(...)
    {
        is_cleared=false;
    }
    delete semaphore;
    int dest=0;
    if(is_cleared==false)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException("Unable to Clear Semaphore");
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;

#else

    return 0;

#endif
}// end of TRcrdClsemaphore::execute

#endif // __DLLSA__
