/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
///////////////////////////////////////////////////////////////////////////////
// @doc     CECListVar.cpp
// @module  CECListVar | This module contains the implementations of this
//	    class.
///////////////////////////////////////////////////////////////////////////////

#include "CECListVar.h"
#include "mdbparse.h"
#include "cxxutils.h"

CECListVar::CECListVar(string name,int size,int options,string values,unsigned long seed)
        :CECProcessVar(options,name),_seed(seed),_cur_seed(seed)
{
    _size=size;
    int ret=parse_for_values(values);
    if(ret!=0)
        throw mdbParseException("Invalid values for List variable");
    if(_options & VORAND )
        _cur_seed=CECVar::GetRandomSeed();
    _cur_pos=-1;
}
CECListVar::~CECListVar()
{}

string  CECListVar::get_value(const string& expression)
{
    if(_cur_pos==-1)//have never done a next
    {
        string msg="No Next or Generate performed on variable "+_name;
        throw mdbParseException(msg);
    }
    return CECProcessVar::get_value(expression);
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   Get next value from value list. When the NEXT command is first
//	    executed, it will retrieve the first value from the set of values
//	    associated with the variable.
//
//	    _cur_pos is initially set to -1 so we can error
//
void    CECListVar::next_value()
{
    _cur_pos++;
    if(_cur_pos>=value_list.size())_cur_pos=0;

    _cur_value=value_list[_cur_pos];
}
void    CECListVar::generate_value()
{
    srand(_cur_seed);//use current seed to randomize
    _cur_seed=rand();//set current seed to random for next srand()
    _cur_pos=_cur_seed % value_list.size();
    _cur_value=value_list[_cur_pos];
}
void    CECListVar::reset_seed(unsigned long seed)
{
    if(seed==0)seed=_seed;//use original seed if required
    _cur_seed=seed;//set seed up for next generate_value

    // And reset our array pointer to the start of the array
    _cur_pos = -1;
}
//@mfunc    parse the declaration record and check that it matches
//	    the definition of the variable so far, if so then parse the values
//	    and return as a vector of strings, this should only be called once per object
//	    i.e. in the constructor of Range and List type variables
//	    that take their values from the declaration, so it should be thread safe as long
//	    as a mutex lock is around the construction(not internal)
//@parm	    declaration string
//@rdesc    0 for success
int CECListVar::parse_for_values(string values)
{
    value_list.clear();
    if(values.empty())
        return 0;
    // Create parsing utility object containing
    // the declaration string.
    Cmdbparse   par(values);
    // Temporary string
    string      str;

    // Skip spaces, tabs, newlines when parsing declaration
    par.skip_whitespace(true);
    // Ignore case when comparing characters and strings
    par.case_sensitive(false);
    // Remove continuation characters which preceed end of
    // each line.
    par.remove_contin_chars("&-");
    // Check for value list start character
    // Clear source bits
    _options &= ~VSMASK;

    // Set source is value list bit
    _options |= VSLIST;

    // Process value list
    while(1)
    {
        // Integer type
        if(_options & VTINT)
        {
            int dummy;

            // Read next integer value from declaration
            if((par.get_integer(dummy, str)) == 1)
            {
                // Add value to value list
                value_list.push_back(str);
                // If followed by comma, expect another value
                if((par.expect(',')) == 1)
                    continue;
            }
            // Got no value, exit loop
            break;
        }
        else if(_options & VTCHAR)
        {
            // Read next delimited string from declaration. The
            // delimiter can vary depending on the string contents,
            // Check we have one of the allowed delimiters
            char delim = par.peek_next_char();
            if( (delim != STRING_VAL_DELIMC) && (delim != ALT_STRING_VAL_DELIMC) )
                return -1;
            if((par.get_delim_string(str, delim)) == 1)
            {
                // Add value to value list
                value_list.push_back(str);
                // If followed by comma, expect another value
                if((par.expect(',')) == 1)
                    continue;
            }
            break;
        }
    }
    return 0;

}
