/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_ASYNCREQUEST_H_
#define _OPENSTA_ASYNCREQUEST_H_

////////////////////////////////////////////////////////////////////////////////
// Asynchronous request definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "Iocp.hpp"
#include "AsyncDevice.hpp"
#include "Trace.hpp"

struct IAsyncRequestEvent
{
    virtual void OnNewRequest(DWORD dwRequestNumber)       = 0;
    virtual void OnRequestFinished(DWORD dwRequestNumber)  = 0;
    virtual void OnRequestCancelled(DWORD dwRequestNumber) = 0;
    virtual void OnRequestClosed(DWORD dwRequestNumber)    = 0;
};

// This class implements an asynchronous "request", which is a complete
// query in a given protocol, using an asynchronous I/O device. This is
// a base class. Actual requests are to be implemented in a derivative.
// Examples: a GET request in HTTP, using a TCP socket, an update of a
// record in a file (1 read + 1 write), using a file handle device.
// THREAD UNSAFE, except for CancelRequest(), which must be callable
// from a concurrent thread. However, CancelRequest() is not required to
// be synchronous. This allows "soft cancels".
class CAsyncRequest : public IocpEvent
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CAsyncRequest(CIOCP& iocp, IAsyncRequestEvent& requestEventListener, DWORD dwRequestNumber = 0, CTrace* pTrace = NULL)
        : m_dwState(0), m_pDevice(NULL), m_dwRequestNumber(dwRequestNumber),
          m_requestEventListener(requestEventListener), m_iocp(iocp), m_bCancel(FALSE),
          m_pTrace(pTrace)
    {}
    virtual ~CAsyncRequest() {}

    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------

    // Initialization. Override to create the device object.
    virtual void Init() = 0;

    // Launch a new request. This is a model for derivatives.
    // virtual DWORD InitiateRequest(LPCSTR /*zRequest*/)
    // {InterlockedExchange((LPLONG)&m_bCancel, (LONG)FALSE); ... work ...; return 0;}

    // Marks the request (if any pending) as "to be cancelled".
    virtual void CancelRequest()
    {InterlockedExchange((LPLONG)&m_bCancel, (LONG)TRUE);}

    //// --------------------------------------------------------------------------------
    //// IocpEvent methods. This base implementation does nothing.
    //// --------------------------------------------------------------------------------
    virtual void OnEvent      (DWORD /*dwNumberOfBytesTransferred*/, BOOL /*bIsIo*/, DWORD /*dwEventType*/) {}
    virtual void OnEventFailed(DWORD /*dwNumberOfBytesTransferred*/, BOOL /*bIsIo*/, DWORD /*dwEventType*/, DWORD /*dwError*/) {}

    DWORD GetRequestNumber()  const   { return m_dwRequestNumber; }

protected:
    //// --------------------------------------------------------------------------------
    //// Internal callbacks.
    //// This base implementation forwards the calls to the request event listener.
    //// --------------------------------------------------------------------------------
    virtual void OnNewRequest()       {m_requestEventListener.OnNewRequest(m_dwRequestNumber);}
    virtual void OnRequestFinished()  {m_requestEventListener.OnRequestFinished(m_dwRequestNumber);}
    virtual void OnRequestCancelled() {m_requestEventListener.OnRequestCancelled(m_dwRequestNumber);}
    virtual void OnRequestClosed()    {m_requestEventListener.OnRequestClosed(m_dwRequestNumber);}

    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------
    CIOCP&              m_iocp;                 // The associated IOCP.
    IAsyncRequestEvent& m_requestEventListener; // Object listening to request events.
    const DWORD         m_dwRequestNumber;      // Request slot number.
    DWORD               m_dwState;              // State used by the object to define the finite state automaton
                                                // of request actions. The initial state is defined as 0.
    CAsyncDevice       *m_pDevice;              // Pointer to an asynchronous device.
    BOOL                m_bCancel;              // If TRUE, the current request is to be cancelled when the current I/O completes.
    CTrace             *m_pTrace;               // The ostream trace support object.

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    void operator=(const CAsyncRequest&);
};

#endif  //  _OPENSTA_ASYNCREQUEST_H_
