/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_ASYNCREQUESTMANAGER_H_
#define _OPENSTA_ASYNCREQUESTMANAGER_H_

////////////////////////////////////////////////////////////////////////////////
// Asynchronous request manager definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include <vector>

#include "AsyncRequest.hpp"
#include "ThreadPoolController.hpp"
#include "Trace.hpp"
#include "Fastlock.hpp"
#include "cxxutils.h"

// This class implements a container of requests.
// THREAD SAFE, except Init(). Derivatives should also be made thread safe,
// except their Init() method.
class CAsyncRequestManager : public IAsyncRequestEvent
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CAsyncRequestManager()
        : m_dwPendingRequests(0), m_pController(NULL),
          m_dwManagerNumber(0), m_pTrace(NULL),
          m_sl(4000)
    {}
    virtual ~CAsyncRequestManager()
    {
        // Delete the requests in the derived class.
    }

    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------

    // Initialization. Creates the requests.
    // The derived overrides must call this base implementation first.
    virtual DWORD Init(CThreadPoolController& controller, DWORD dwManagerNumber, CTrace* pTrace = NULL)
    {
        m_pController = &controller;
        m_dwManagerNumber = dwManagerNumber;
        m_pTrace = pTrace;

        return 0;
    }

    // Asks all clients to cancel any pending query.
    virtual void CancelAllQueries()
    {
        if(m_pTrace)
            m_pTrace->TraceIf("requests", string("CANCELQUERIES: Cancelling all queries for ") + int_to_string((unsigned)m_dwManagerNumber), TRUE);

        CLock lock(m_sl);
        for(RequestsType::iterator it = m_requests.begin(); it != m_requests.end(); it++)
            (*it)->CancelRequest();
    }

    // Get the number of requests currently executing.
    DWORD GetPendingRequestCount()  const   {
                                                CLock lock(m_sl);
                                                return m_dwPendingRequests;
                                            }

    DWORD GetManagerNumber()        const   { return m_dwManagerNumber; }

    /* This is a model of how to program a request in a derived class.
    DWORD InitiateRequest(LPCSTR zRequest)
    {
        DWORD dwRequestNumber = 0;

        {
            CLock lock(m_sl);
            dwRequestNumber = GetAvailableRequest();
            if(dwRequestNumber != 0xFFFFFFFF)
                m_requests[dwRequestNumber]->InitiateRequest(zRequest);
            else
                dwRequestNumber = (DWORD)OPENSTA_E_NO_REQUEST_AVAILABLE;
        }
        return dwRequestNumber;
    }
    */

    //// --------------------------------------------------------------------------------
    //// IAsyncRequestEvent methods.
    //// --------------------------------------------------------------------------------
    virtual void OnNewRequest(DWORD dwRequestNumber)
    {
        CLock lock(m_sl);
        ++m_dwPendingRequests;
        m_freeRequests[dwRequestNumber] = FALSE;
    }
    virtual void OnRequestFinished(DWORD /*dwRequestNumber*/)
    {
        CLock lock(m_sl);
        if(--m_dwPendingRequests == 0)
            OnAllRequestsFinished();
    }
    virtual void OnRequestCancelled(DWORD /*dwRequestNumber*/)
    {
        CLock lock(m_sl);
        if(--m_dwPendingRequests == 0)
            OnAllRequestsCancelled();
    }
    virtual void OnRequestClosed(DWORD dwRequestNumber)
    {
        CLock lock(m_sl);
        m_freeRequests[dwRequestNumber] = TRUE;
    }

protected:
    //// --------------------------------------------------------------------------------
    //// Internal callbacks.
    //// --------------------------------------------------------------------------------
    virtual void OnAllRequestsFinished()  {m_pController->OnAllRequestsFinished(m_dwManagerNumber);}
    virtual void OnAllRequestsCancelled() {m_pController->OnAllRequestsCancelled(m_dwManagerNumber);}

    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------

    // Adds an entry to the request vector, and returns the new position.
    // This method is not protected against concurrent calls: it must be called
    // within the scope of such a protection.
    DWORD AddRequest(CAsyncRequest* pRequest)
    {
        m_requests.push_back(pRequest);
        m_freeRequests.push_back(TRUE);
        return m_requests.size() - 1;
    }

    // Gets the next available request number, or 0xFFFFFFFF if none is available.
    // This method is not protected against concurrent calls: it must be called
    // within the scope of such a protection.
    DWORD GetAvailableRequest()
    {
        DWORD dwRequestNumber = ULONG_MAX;
        DWORD dwSize = m_freeRequests.size();
        for(DWORD dw = 0; dw < dwSize; dw++)
        {
            if(m_freeRequests[dw])
            {
                dwRequestNumber = dw;
                m_freeRequests[dwRequestNumber] = FALSE;    // Reserve now.
                break;
            }
        }
        return dwRequestNumber;
    }

    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------
    typedef std::vector<CAsyncRequest*> RequestsType;
    mutable CFastLock       m_sl;                               // Method-level protection against concurrent accesses.
    CThreadPoolController*  m_pController;                      // The enclosing controller.
    RequestsType            m_requests;                         // Vector of request pointers.
    std::vector<BOOL>       m_freeRequests;                     // An entry set to TRUE means that the corresponding request
                                                                // is disconnected.
    DWORD                   m_dwPendingRequests;                // Number of pending requests.
    DWORD                   m_dwManagerNumber;                  // Request manager number.
    CTrace                 *m_pTrace;                           // The ostream trace support object.

private:
    void operator=(const CAsyncRequestManager&);
};

#endif  // _OPENSTA_ASYNCREQUESTMANAGER_H_
