/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "stdafx.h"
#include <assert.h>
#include "MemMappedFile.hpp"
#include <string>

using namespace std;

CMemMappedFile::CMemMappedFile()
: m_hFile(NULL), m_pView(NULL), m_dwCurrViewOffset(0), m_dwFileSize(0), m_dwViewSize(0)
{
}

CMemMappedFile::~CMemMappedFile()
{
    ReleaseMemMappedFile();
}

BOOL CMemMappedFile::OpenFile(const string &zFileName, DWORD dwViewSize, DWORD dwAccessMode, DWORD dwShareMode)
{
    SYSTEM_INFO sysInfo;
    GetSystemInfo(&sysInfo);

    if(dwViewSize == 0)
    {
        dwViewSize = sysInfo.dwAllocationGranularity;
    }
    else
    {
        // Make sure the specified dwViewSize is a multiple of sysInfo.dwAllocationGranularity.
        if(dwViewSize == 0 || dwViewSize % sysInfo.dwAllocationGranularity != 0)
        {
            // Error: bad parameter.
            return FALSE;
        }
    }

    m_dwViewSize = dwViewSize;
    m_dwAccessMode = dwAccessMode;

    return AllocMemMappedFile(zFileName, (dwAccessMode==GENERIC_READ)?GENERIC_READ:GENERIC_READ|GENERIC_WRITE, dwShareMode);
}

void CMemMappedFile::CloseFile()
{
    ReleaseMemMappedFile();
}

// --------------------------------------------------------------------------------
//
// Write the dwBuffLen first bytes found in zBuff, into the file starting at
// the current position.
//
// Because of MapViewOfFile requirements linked to the memory granularity, must 
// fill the rest of the current view, then move the view/enlarge the file, and 
// finally copy the remaining data to the new view.
//
// --------------------------------------------------------------------------------
BOOL CMemMappedFile::WriteData(LPCSTR zBuff, DWORD dwBuffLen)
{
    if((m_dwAccessMode & GENERIC_WRITE) == 0 || m_pView == NULL)
    {
        assert(0);
        return FALSE;
    }

    if(dwBuffLen == 0)
        return TRUE;

    if(m_dwCurrViewOffset + dwBuffLen > m_dwViewSize)
    {
        // Not enough room left in the current view. Will fill the remaining 
        // empty space, then move the view and finally copy the remaining data.

        // Fill the remaining of the current view.
        DWORD dwViewSizeLeft = m_dwViewSize - m_dwCurrViewOffset;
        CopyMemory(m_pView + m_dwCurrViewOffset, zBuff, dwViewSizeLeft);

        // Adjust the size of the file.
        m_dwFileSize += m_dwViewSize;

        if(MoveView(m_dwFileSize) == TRUE)
        {
            // Dump the remaining data.
            CopyMemory(m_pView, zBuff + dwViewSizeLeft, dwBuffLen - dwViewSizeLeft);
            m_dwCurrViewOffset = dwBuffLen - dwViewSizeLeft;
            return TRUE;
        }
        else
            return FALSE;
    }
    else
    {
        // Enough room left in the current view.
        CopyMemory(m_pView + m_dwCurrViewOffset, zBuff, dwBuffLen);
        m_dwCurrViewOffset += dwBuffLen;
        return TRUE;
    }
}

// --------------------------------------------------------------------------------
//
// Read dwBuffLen bytes into zBuff, starting at the current position.
// dwReadBytes is the number of bytes actually read.
// If no bytes could be read, then return FALSE.
//
// --------------------------------------------------------------------------------
BOOL CMemMappedFile::ReadData(LPSTR zBuff, DWORD dwBuffLen, DWORD &dwReadBytes)
{
    if((m_dwAccessMode & GENERIC_READ) == 0 || m_pView == NULL)
    {
        assert(0);
        dwReadBytes = 0;
        return FALSE;
    }

    // Map the whole file when reading. Might be worth moving the view through the file if file is large.
    if(m_dwCurrViewOffset + dwBuffLen > m_dwViewSize)
    {
        // Not enough room left in the current view. Will read the remaining 
        // data in current view, then move the view and finally read the remaining data.

        // Fill the remaining of the current view.
        DWORD dwViewSizeLeft = m_dwViewSize - m_dwCurrViewOffset;
        CopyMemory(zBuff, m_pView + m_dwCurrViewOffset, dwViewSizeLeft);

        // Adjust the size of the file.
        m_dwFileSize += m_dwViewSize;

        if(MoveView(m_dwFileSize) == TRUE)
        {
            // Dump the remaining data.
            CopyMemory(zBuff + dwViewSizeLeft, m_pView, dwBuffLen - dwViewSizeLeft);
            m_dwCurrViewOffset = dwBuffLen - dwViewSizeLeft;
            dwReadBytes = dwBuffLen;
            return TRUE;
        }
        else
        {
            dwReadBytes = dwViewSizeLeft;
            return FALSE;
        }
    }
    else
    {
        // Enough room left in the current view.
        DWORD dwBytesToCopy = __min(dwBuffLen, m_dwViewSize - m_dwCurrViewOffset);

        if(dwBytesToCopy == 0)
            return FALSE;

        CopyMemory(zBuff, m_pView + m_dwCurrViewOffset, dwBytesToCopy);
        m_dwCurrViewOffset += dwBytesToCopy;
        dwReadBytes = dwBytesToCopy;
        return TRUE;
    }
}

void CMemMappedFile::Rewind()
{
    BOOL bRes = Seek(0L);
    assert(bRes);
}

BOOL CMemMappedFile::Seek(DWORD dwPosInFile)
{
    DWORD dwNewViewBeg = (dwPosInFile / m_dwViewSize) * m_dwViewSize;
    BOOL bRes = MoveView(dwNewViewBeg);
    m_dwCurrViewOffset = dwPosInFile - dwNewViewBeg;

    return bRes;
}

// --------------------------------------------------------------------------------
//
// Input/Output from/to a file. Map it into memory.
// Assume that dwViewSize is carefully specified.
//
// --------------------------------------------------------------------------------
BOOL CMemMappedFile::AllocMemMappedFile(const string &zFileName, DWORD dwAccessMode, DWORD dwShareMode)
{
    m_dwCurrViewOffset = 0;

    m_hFile = CreateFile(   zFileName.c_str(), dwAccessMode, dwShareMode, NULL, 
                            (dwAccessMode==GENERIC_READ)?OPEN_EXISTING:CREATE_ALWAYS, 
                            FILE_ATTRIBUTE_NORMAL|FILE_FLAG_RANDOM_ACCESS, NULL);

    if(m_hFile != INVALID_HANDLE_VALUE)
    {
        if(MoveView(0) == FALSE)
        {
            CloseHandle(m_hFile);
            m_hFile = NULL;

            return FALSE;
        }
    }
    else
    {
        m_hFile = NULL;
        return FALSE;
    }

    if(dwAccessMode == GENERIC_READ)
    {
        m_dwViewSize = GetFileSize(m_hFile, NULL);
    }

    return TRUE;
}

// --------------------------------------------------------------------------------
//
// Eventually enlarge the file and move the view.
// dwStartPos must be a multiple of m_dwViewSize.
//
// --------------------------------------------------------------------------------
BOOL CMemMappedFile::MoveView(DWORD dwStartPos)
{
    if(m_pView)
    {
        UnmapViewOfFile(m_pView);
        m_pView = NULL;
    }

    HANDLE hMap = CreateFileMapping(m_hFile, NULL, (m_dwAccessMode==GENERIC_READ)?PAGE_READONLY:PAGE_READWRITE, 0, (m_dwAccessMode==GENERIC_READ)?0:dwStartPos + m_dwViewSize, NULL);
    if(hMap != NULL)
    {
        m_pView = (LPSTR)MapViewOfFile(hMap, (m_dwAccessMode==GENERIC_READ)?FILE_MAP_READ:FILE_MAP_WRITE, 0, dwStartPos, (m_dwAccessMode==GENERIC_READ)?0:m_dwViewSize);
        CloseHandle(hMap);

        if(m_pView == NULL)
        {
            CloseHandle(m_hFile);
            m_hFile = NULL;

            return FALSE;
        }
        else
            return TRUE;
    }
    else
    {
        CloseHandle(m_hFile);
        m_hFile = NULL;

        return FALSE;
    }
}

// --------------------------------------------------------------------------------
// --------------------------------------------------------------------------------
void CMemMappedFile::ReleaseMemMappedFile()
{
    if(m_pView && m_hFile)
    {
        // File output: adjust file size and close.
        UnmapViewOfFile(m_pView);
        SetFilePointer(m_hFile, m_dwFileSize + m_dwCurrViewOffset, NULL, FILE_BEGIN);
        SetEndOfFile(m_hFile);
        CloseHandle(m_hFile);
        m_pView = NULL;
        m_hFile = NULL;
    }
}

