/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "Tools.hpp"

#ifdef USE_PERF_DATA
#   include "PerfCounterKeys.hpp"
#endif

/* #########################################################################

    CDataBuff

   ######################################################################### */

// Move dwInc buffers further in the list.
WILL_INLINE CDataBuff *CDataBuff::operator+(DWORD dwInc)
{
    CDataBuff *cursor = this;

    for(; dwInc>0 && cursor!=NULL ; dwInc--)
    {
        cursor = cursor->GetNextBuff();
    }

    return cursor;
}

WILL_INLINE std::ostream &operator << (std::ostream& os, const CDataBuff &dataBuff)
{
    os.write(dataBuff.GetDataPtr(), dataBuff.GetUsedDataSize());

    return os;
}


/* #########################################################################

    CGlobalDataBuffHandler

   ######################################################################### */

WILL_INLINE CGlobalDataBuffHandler::CGlobalDataBuffHandler()
    : m_dwCommittedMem(0), m_freeDataBuffsList(NULL), m_lastFreeDataBuffer(NULL), m_dwFreeBuffsCount(0), m_dwAllocatedBuffsCount(0)
{
}

WILL_INLINE CGlobalDataBuffHandler::~CGlobalDataBuffHandler()
{
    assert(m_dwAllocatedBuffsCount == m_dwFreeBuffsCount);  // Make sure all the buffers were released.
}


// ----------------------------------------------------------------------------
// Initialise static data in CGlobalDataBuffHandler.
// Includes enquiring for system page size and allocating the virtual
// memory buffer.
// Return FALSE if the virtual memory buffer could not be allocated.
// ----------------------------------------------------------------------------

WILL_INLINE BOOL CGlobalDataBuffHandler::Init(DWORD dwMaxDataBuff, DWORD dwGrowingBuffCount)
{
    assert(m_pFl == NULL);
    if(m_pFl == NULL)
        m_pFl = new CFastLock();

    if(m_pFl == NULL)   return FALSE;

    s_dwMaxBuffCount        = dwMaxDataBuff;
    s_dwGrowingBuffCount    = dwGrowingBuffCount;

#ifdef USE_PERF_DATA
    s_pRecvBuffAllocatedMem    = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_RECV_BUFF_ALLOCATED_MEM);
    (*s_pRecvBuffAllocatedMem) = 0L;
    s_pRecvBuffUsedMem         = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_RECV_BUFF_USED_MEM);
    (*s_pRecvBuffUsedMem)      = 0L;
    s_pDataStoredCounterRef    = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_DATA_STORED);
    (*s_pDataStoredCounterRef) = 0L;
#endif

    SYSTEM_INFO sysInfo;

    GetSystemInfo(&sysInfo);

    s_dwMaxPhysBuffLength       = sysInfo.dwPageSize;
    s_dwDataBuffExtraInfoSize   = sizeof(CDataBuff *) + sizeof(DWORD);
    s_dwDataBuffUsableSize      = s_dwMaxPhysBuffLength-s_dwDataBuffExtraInfoSize;

    s_dwBytesGrownBy = s_dwMaxPhysBuffLength * s_dwGrowingBuffCount;

    return AllocMemoryBuffer();
}

WILL_INLINE void CGlobalDataBuffHandler::UnInit()
{
    if(s_pvVirtualMemBuff != NULL)
    {
#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
        InterlockedExchange(CGlobalDataBuffHandler::s_pRecvBuffAllocatedMem, 0L);
#   else
        (*CGlobalDataBuffHandler::s_pRecvBuffAllocatedMem) = 0L;
#   endif
#endif
        if(VirtualFree(s_pvVirtualMemBuff, 0, MEM_RELEASE) == FALSE)
        {
            DWORD dwErr = GetLastError();
            assert(0);
        }
    }
}

// ----------------------------------------------------------------------------

// Returns a list of dwBuffCount free buffers allocated from the global list.
// Returns FALSE if all the buffers could not be allocated.

// [IN]     dwBuffCount:    number of buffers requested.
// [OUT]    pNewList:       pointer to the first item in the allocated list.
// [OUT]    pLastItem:      pointer to the last item in the allocated list.

// ----------------------------------------------------------------------------

WILL_INLINE BOOL CGlobalDataBuffHandler::RequestBuffs(DWORD dwBuffCount, CDataBuff *&pNewList, CDataBuff *&pLastItem)
{
    assert(dwBuffCount != 0U);

    CLock lock(m_pFl);

    if(m_dwFreeBuffsCount <= dwBuffCount)
    {
        // Commit enough memory to hold the required buffers rounded up to the growing size.
        if(CommitMemoryBuffer((((dwBuffCount-m_dwFreeBuffsCount)/s_dwGrowingBuffCount)+1L)*s_dwGrowingBuffCount) == FALSE)
            return FALSE;
    }

    // The free buffers are taken from the begining of the free list.
    pNewList    = m_freeDataBuffsList;
    pLastItem   = (*m_freeDataBuffsList)+(dwBuffCount-1L);

    m_freeDataBuffsList = pLastItem->GetNextBuff();
    m_dwFreeBuffsCount -= dwBuffCount;

#ifdef USE_PERF_DATA
    (*CGlobalDataBuffHandler::s_pRecvBuffUsedMem) += dwBuffCount*s_dwMaxPhysBuffLength; // No need to interlock since m_pFl is already locked.
#endif

    pLastItem->SetNextBuff(NULL);

    return TRUE;
}


// ----------------------------------------------------------------------------
//  ReturnBuffs
//
//  Return buffers to the global free buffers list.
//  That version must find the last buffer in the list and evaluate the new 
//  items count.
// ----------------------------------------------------------------------------

WILL_INLINE void CGlobalDataBuffHandler::ReturnBuffs(CDataBuff *pFirstBuff)
{
    assert(pFirstBuff != NULL);

    DWORD dwReturnedBuffsCount = 0U;
    CDataBuff *pLastBuff = pFirstBuff;

    while(pLastBuff->GetNextBuff() != NULL)
    {
        pLastBuff = pLastBuff->GetNextBuff();
        dwReturnedBuffsCount++;
    }

    ReturnBuffs(pFirstBuff, pLastBuff, dwReturnedBuffsCount);
}

// ----------------------------------------------------------------------------
//  ReturnBuffs
//
//  Return buffers to the global free buffers list.
//  That version is more efficient than the previous one since all the info on
//  the list is provided.
// ----------------------------------------------------------------------------

WILL_INLINE void CGlobalDataBuffHandler::ReturnBuffs(CDataBuff *pFirstBuff, CDataBuff *pLastBuff, DWORD dwRetBuffCount)
{
    assert(pFirstBuff != NULL && pLastBuff != NULL);

    CLock lock(m_pFl);

    m_lastFreeDataBuffer->SetNextBuff(pFirstBuff);

    m_lastFreeDataBuffer = pLastBuff;
    m_dwFreeBuffsCount += dwRetBuffCount;

#ifdef USE_PERF_DATA
    (*s_pRecvBuffUsedMem) -= (LONG)s_dwMaxPhysBuffLength*dwRetBuffCount;    // No need to use interlocks since m_pFl is locked.
#endif
}


// ----------------------------------------------------------------------------
//
// Create the global memory buffer as a Virtual Memory area.
//
// ----------------------------------------------------------------------------

WILL_INLINE BOOL CGlobalDataBuffHandler::AllocMemoryBuffer()
{
    s_pvVirtualMemBuff = VirtualAlloc(NULL, s_dwMaxBuffCount*s_dwMaxPhysBuffLength, MEM_RESERVE, PAGE_READWRITE);
    return s_pvVirtualMemBuff != NULL;
}


// ----------------------------------------------------------------------------
//
// Reserve several pages in the virtual memory pointed to by s_pvVirtualMemBuff.
//
// [IN] dwNbBuffRequired: Number of memory pages to allocate.
//
// ----------------------------------------------------------------------------

WILL_INLINE BOOL CGlobalDataBuffHandler::CommitMemoryBuffer(DWORD dwNbBuffRequired)
{
    dwNbBuffRequired = __min(dwNbBuffRequired, s_dwMaxBuffCount-m_dwAllocatedBuffsCount);

    DWORD dwToCommitSize = dwNbBuffRequired*s_dwMaxPhysBuffLength;
    LPVOID pNewBuff = VirtualAlloc((LPVOID)((CHAR*)s_pvVirtualMemBuff + m_dwCommittedMem), dwToCommitSize, MEM_COMMIT, PAGE_READWRITE);

    if(pNewBuff == NULL)
    {
        return FALSE;
    }

    m_dwAllocatedBuffsCount += dwNbBuffRequired;
    m_dwFreeBuffsCount      += dwNbBuffRequired;
    m_dwCommittedMem        += dwToCommitSize;

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
    InterlockedExchangeAdd(CGlobalDataBuffHandler::s_pRecvBuffAllocatedMem, dwToCommitSize);
#   else
    (*CGlobalDataBuffHandler::s_pRecvBuffAllocatedMem) += dwToCommitSize;
#   endif
#endif

    // Add these free buffers to m_freeDataBuffsList.
    if(m_freeDataBuffsList == NULL)
    {
        // We are allocating data buffers for the first time.
        m_freeDataBuffsList = (CDataBuff*)s_pvVirtualMemBuff;
    }
    else
    {
        assert(m_lastFreeDataBuffer != NULL);

        m_lastFreeDataBuffer->SetNextBuff((CDataBuff*)pNewBuff);
    }

    // Map m_lastFreeDataBuffer to the last buffer allocated (end of the committed memory minus the size of a CDataBuff.
    m_lastFreeDataBuffer = (CDataBuff*)((CHAR*)pNewBuff + dwToCommitSize - s_dwMaxPhysBuffLength);

    // Init the buffers.
    CHAR *pBuff = (CHAR *)pNewBuff;
    for(DWORD dw=0 ; dw<dwNbBuffRequired-1 ; dw++)
    {
        ((CDataBuff*)(pBuff + s_dwMaxPhysBuffLength*dw))->Init((CDataBuff*)(pBuff + s_dwMaxPhysBuffLength*(dw+1)));
    }

    ((CDataBuff*)(pBuff + s_dwMaxPhysBuffLength*dw))->Init(NULL);

    return TRUE;
}


/* #########################################################################

    CDataBuffList

   ######################################################################### */

WILL_INLINE CDataBuffList::CDataBuffList()
    : m_pDataBuffList(NULL), m_pLastItem(NULL), m_dwItemsCount(0)
{}


// ----------------------------------------------------------------------------
//
// Add dwNbItems data-buffers, allocated from the global buffers list.
//
// [IN] dwNbItems: Number of data-buffers to allocate.
// ----------------------------------------------------------------------------

WILL_INLINE BOOL CDataBuffList::AllocateBuffs(DWORD dwNbItems)
{
    CDataBuff *pLastBuff = m_pLastItem;
    CDataBuff *pNewBuffs;
    
    BOOL bRes = m_globalDataBuffHandler.RequestBuffs(dwNbItems, pNewBuffs, m_pLastItem);
    assert(bRes == TRUE);

    if(pLastBuff == NULL)
    {
        assert(m_pDataBuffList == NULL);
        m_pDataBuffList = pNewBuffs;
    }
    else
        pLastBuff->SetNextBuff(pNewBuffs);

    m_dwItemsCount += dwNbItems;

    return bRes;
}


// ----------------------------------------------------------------------------
//
// Release the whole data buff list.
// Send back the items to the global buffer list to make them available again.
//
// ----------------------------------------------------------------------------

WILL_INLINE void CDataBuffList::ReleaseBuffList()
{
    if(m_pDataBuffList != NULL)
    {
        m_globalDataBuffHandler.ReturnBuffs(m_pDataBuffList, m_pLastItem, m_dwItemsCount);
    }

    m_pDataBuffList = m_pLastItem = NULL;
    m_dwItemsCount = 0U;
}

// ----------------------------------------------------------------------------
//
// Append a buffer list to the end of the current list.
// Note that the two lists share the same buffers!!
//
// ----------------------------------------------------------------------------

WILL_INLINE const CDataBuffList &CDataBuffList::operator+=(const CDataBuffList &buffList)
{
    assert(m_pLastItem != NULL);

    m_pLastItem->SetNextBuff(buffList.m_pDataBuffList);
    m_pLastItem = buffList.m_pLastItem;
    m_dwItemsCount += buffList.m_dwItemsCount;

    return *this;
}


// ----------------------------------------------------------------------------
//
// Stores the first dwNbItemsToAdd buffers to the current list, and return a 
// list containing the remaining buffers. buffList should not be used anymore.
// Argument passed as buffList may refer to the same object as the one passed 
// as unusedBuffList.
//
// ----------------------------------------------------------------------------

WILL_INLINE void CDataBuffList::AddBuffList(const CDataBuffList &srcList, CDataBuff *pLastBuff, DWORD dwNbItemsToAdd, CDataBuffList &unusedBuffList)
{
    if(dwNbItemsToAdd == 0U)
    {
        // None of the buffers where used. 'Copy' srcList into unusedBuffList.
        unusedBuffList.m_pDataBuffList  = srcList.m_pDataBuffList;
        unusedBuffList.m_pLastItem      = srcList.m_pLastItem;
        unusedBuffList.m_dwItemsCount   = srcList.m_dwItemsCount;
    }
    else
    {
        // 'Add' items to this.
        if(m_pDataBuffList == NULL)
        {
            // We are an empty list.
            m_pDataBuffList = srcList.m_pDataBuffList;
        }
        else
        {
            m_pLastItem->SetNextBuff(srcList.m_pDataBuffList);
        }

        m_pLastItem = pLastBuff;
        m_dwItemsCount += dwNbItemsToAdd;

        // Build the unusedBuffList list.
        unusedBuffList.m_pDataBuffList = pLastBuff->GetNextBuff();

        if(unusedBuffList.m_pDataBuffList == NULL)
        {
            // All the list was added. Return an empty one.
            unusedBuffList.m_pLastItem = NULL;
            unusedBuffList.m_dwItemsCount = 0U;
        }
        else
        {
            unusedBuffList.m_pLastItem = srcList.m_pLastItem;
            unusedBuffList.m_dwItemsCount = srcList.m_dwItemsCount - dwNbItemsToAdd;
        }

        m_pLastItem->SetNextBuff(NULL);
    }
}

WILL_INLINE std::ostream &operator << (std::ostream& os, const CDataBuffList &dataBuffList)
{
    const CDataBuff *cursor = dataBuffList.GetBuffList();

    while(cursor != NULL)
    {
        os << *cursor;
        cursor = cursor->GetNextBuff();
    }

    return os;
}


/* #########################################################################

CDataBuffListChunkAccessor

  Accessor that specializes in parsing chunk data.
  The iterator must first be moved right after the header since the chunk 
  accessor expects the first element to be a chunk size string.

  The chunk accessor should not be used to parse data 'on the fly'.

  Define the specialisation of the methods in CDataBuffListBaseAccessor
  that behave differently when used with chunks.

######################################################################### */

template <>
WILL_INLINE CDataBuffListBaseAccessor<CdBLCA>::CDataBuffListBaseAccessor<CdBLCA>(CDataBuffList *pDataBuffList)
    : m_dwCurrentChunkSizeLeft(0), m_dwPosInCurrDataBuff(0), m_pDataBuffList(pDataBuffList)
{
    if(pDataBuffList == NULL)
    {
        m_pCurrDataBuff = NULL;
    }
    else
    {
        m_pCurrDataBuff = m_pDataBuffList->GetBuffList();

        // Adjust chunk information.
        ckReadChunkSize(FALSE);
    }
}

// Warning: requires that dBLA points at the begining of a chunk (ie. the caracter sequence defining the chunk size).
template <>
WILL_INLINE void CDataBuffListBaseAccessor<CdBLCA>::MoveIterTo(const CDataBuffListBaseAccessor<CdBLA> &dBLA)
{
    m_pDataBuffList             = dBLA.m_pDataBuffList;
    m_pCurrDataBuff             = dBLA.m_pCurrDataBuff;
    m_dwPosInCurrDataBuff       = dBLA.m_dwPosInCurrDataBuff;
    BOOL bRes = ckReadChunkSize(FALSE);
    assert(bRes);
}


__inline BOOL AdjustEndOfBuff(const CDataBuff *&pCurrDataBuff, DWORD &dwPosInCurrDataBuff)
{
    if(dwPosInCurrDataBuff == pCurrDataBuff->GetUsedDataSize())
    {
        // Move to the next buffer.
        if(pCurrDataBuff->GetNextBuff() == NULL)
            return FALSE;

        pCurrDataBuff = pCurrDataBuff->GetNextBuff();
        dwPosInCurrDataBuff = 0U;
    }

    return TRUE;
}


// ----------------------------------------------------------------------------
//
// Try moving the iterator dwItems items further. Returns the number of items 
// it skipped actually.
// The three cases are: move in the current buffer; must move to another buffer; 
// must move to the next buffer, but not received yet.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE DWORD CDataBuffListBaseAccessor<CdBLCA>::Skip(DWORD dwItems)
{
    DWORD dwToSkip = dwItems;
    DWORD dwCurBuffUsedSize, dwCurrBuffSizeLeft;

    while(dwToSkip > 0L)
    {
        // Skip either dwToSkip characters, or till the end of the current chunk, 
        // or till the end of current buffer, or any combination of these.
        dwCurBuffUsedSize   = m_pCurrDataBuff->GetUsedDataSize();
        dwCurrBuffSizeLeft  = dwCurBuffUsedSize - m_dwPosInCurrDataBuff;

        if(dwToSkip > m_dwCurrentChunkSizeLeft)
        {
            // Must skip current chunk.
            if(m_dwCurrentChunkSizeLeft > dwCurrBuffSizeLeft)
            {
                // Meet end of buffer first.
                dwToSkip                    -= dwCurrBuffSizeLeft;
                m_dwCurrentChunkSizeLeft    -= dwCurrBuffSizeLeft;

                m_dwPosInCurrDataBuff       = dwCurBuffUsedSize;
            }
            else
            {
                // Meet end of current chunk first.
                dwToSkip                    -= m_dwCurrentChunkSizeLeft;
                m_dwPosInCurrDataBuff       += m_dwCurrentChunkSizeLeft;

                m_dwCurrentChunkSizeLeft    = 0U;
            }
        }
        else
        {
            // Move in current chunk.
            if(dwToSkip > dwCurrBuffSizeLeft)
            {
                // Meet end of buffer first.
                dwToSkip                    -= dwCurrBuffSizeLeft;
                m_dwCurrentChunkSizeLeft    -= dwCurrBuffSizeLeft;

                m_dwPosInCurrDataBuff       = dwCurBuffUsedSize;
            }
            else
            {
                // Found the contiguous last dwToSkip chars.
                m_dwPosInCurrDataBuff       += dwToSkip;
                m_dwCurrentChunkSizeLeft    -= dwToSkip;

                dwToSkip                    = 0U;
            }
        }

        // Check if reached the end of the current buffer.
        if(m_dwPosInCurrDataBuff == dwCurBuffUsedSize)
        {
            // The current buffer is either empty or we must use the next buffer.
            if(m_pCurrDataBuff->GetNextBuff() != NULL)
            {
                m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
                m_dwPosInCurrDataBuff = 0U;
            }
            else
            {
                // No other buffers available. Cannot skip all the items.
                return dwItems - dwToSkip;
            }
        }

        // Check if reached the end of the current chunk.
        if(m_dwCurrentChunkSizeLeft == 0U)
        {
            if(ckReadChunkSize() == FALSE)
            {
                // Reached the end of the data.
                return dwItems - dwToSkip;
            }
        }
    }

    return dwItems;
}


// ----------------------------------------------------------------------------
//
// Compute the number of items 'first' should be moved forward in order to find 
// 'second'. Returns ULONG_MAX if could not find 'second' or if an error 
// occurred.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE DWORD CDataBuffListBaseAccessor<CdBLCA>::ComputeDistance(const CDataBuffListBaseAccessor<CdBLCA> &first, const CDataBuffListBaseAccessor<CdBLCA> &second) const
{
    assert(first.m_pDataBuffList->GetBuffList() != NULL && second.m_pDataBuffList->GetBuffList() != NULL);
    assert(first.m_pDataBuffList->GetBuffList() == second.m_pDataBuffList->GetBuffList());

    // Copy the accessors to avoid modifying them in the code below.
    CDataBuffListChunkAccessor iter = (CDataBuffListChunkAccessor&)first;

    DWORD dwDist = 0U;

    // Move to the right buffer, parsing the chunk data. Note that a chunk could be spread over several buffers.
    while(iter.m_pCurrDataBuff != second.m_pCurrDataBuff)
    {
        // Move to the next raw buffer, updating dwDist.

        DWORD dwBytesLeftInBuff = iter.m_pCurrDataBuff->GetUsedDataSize() - iter.m_dwPosInCurrDataBuff;

        if(iter.m_dwCurrentChunkSizeLeft >= dwBytesLeftInBuff)
        {
            // Find the end of current buffer first. Move the iterator to the next buffer.
            dwDist += dwBytesLeftInBuff;

            iter.m_dwCurrentChunkSizeLeft -= dwBytesLeftInBuff; // iter.m_dwCurrentChunkSizeLeft might be 0 now.

            if(iter.m_pCurrDataBuff->GetNextBuff() == NULL)
                return ULONG_MAX;

            iter.m_pCurrDataBuff = iter.m_pCurrDataBuff->GetNextBuff();
            iter.m_dwPosInCurrDataBuff = 0U;
        }
        else
        {
            // End of chunk comes before end of buffer.
            dwDist += iter.m_dwCurrentChunkSizeLeft;

            iter.m_dwPosInCurrDataBuff += iter.m_dwCurrentChunkSizeLeft;    // iter.m_dwPosInCurrDataBuff can't be GetUsedDataSize.
            iter.m_dwCurrentChunkSizeLeft = 0U;
        }

        if(iter.m_dwCurrentChunkSizeLeft == 0U)
        {
            BOOL bRes = iter.ckReadChunkSize();
            assert(bRes);
        }
    }

    // Move to the right chunk in the buffer pointed to by 'end'.
    while(iter.m_dwPosInCurrDataBuff + iter.m_dwCurrentChunkSizeLeft < second.m_dwPosInCurrDataBuff)
    {
        assert(iter.m_dwPosInCurrDataBuff + iter.m_dwCurrentChunkSizeLeft < iter.m_pCurrDataBuff->GetUsedDataSize());

        dwDist += iter.m_dwCurrentChunkSizeLeft;
        iter.m_dwPosInCurrDataBuff += iter.m_dwCurrentChunkSizeLeft;
        iter.m_dwCurrentChunkSizeLeft = 0U;
        BOOL bRes = iter.ckReadChunkSize();
        assert(bRes);
    }

    // Move to the right position in the found chunk.
    dwDist += second.m_dwPosInCurrDataBuff - iter.m_dwPosInCurrDataBuff;

    return dwDist;
}


// ----------------------------------------------------------------------------
//
// Read the (hexadecimal) chunk size and update the position variables.
//
// ----------------------------------------------------------------------------

__inline BOOL ckLocalReadChunkSize(const CDataBuff *&pCurrDataBuff, DWORD &dwPosInCurrDataBuff, DWORD &dwCurrentChunkSizeLeft, BOOL bSkipLastCkCRLF = TRUE)
{
    // Adjust the pointers.
    assert(pCurrDataBuff);

    CHAR curChar;

    dwCurrentChunkSizeLeft = 0U;

    if(bSkipLastCkCRLF == TRUE)
    {
        // Called when just parsed another chunk. Means we have to skip last chunk's CRLF.
        curChar = pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff];
        if(curChar == '\r')
        {
            // Next char must be '\n'.
            ++dwPosInCurrDataBuff;

            if(AdjustEndOfBuff(pCurrDataBuff, dwPosInCurrDataBuff) == FALSE)
                return FALSE;
        }

        assert(pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff] == '\n');

        // Skip the '\n'.
        ++dwPosInCurrDataBuff;

        if(AdjustEndOfBuff(pCurrDataBuff, dwPosInCurrDataBuff) == FALSE)
            return FALSE;
    }

    while(TRUE)
    {
        curChar = pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff];

        if(IsAlpha(curChar) == FALSE && IsDigit(curChar) == FALSE)
        {
            // Reached the end of the value string. The value was successfully parsed.
            break;
        }

        dwCurrentChunkSizeLeft *= 16U;
        if(curChar >= 'a')
            dwCurrentChunkSizeLeft += curChar - 'a' + 10U;
        else
            if(curChar >= 'A')
                dwCurrentChunkSizeLeft += curChar - 'A' + 10U;
            else
                dwCurrentChunkSizeLeft += curChar - '0';

        ++dwPosInCurrDataBuff;

        if(AdjustEndOfBuff(pCurrDataBuff, dwPosInCurrDataBuff) == FALSE)
            return FALSE;
    }

    // Skip everything (including chunk extension) till the CRLF (LF) sequence.
    do
    {
        // Read current char.
        curChar = pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff];

        // Move cursor to the next char.
        ++dwPosInCurrDataBuff;

        if(AdjustEndOfBuff(pCurrDataBuff, dwPosInCurrDataBuff) == FALSE)
            return FALSE;
    }
    while(curChar != '\n');

    // Eventually skip the trailer (skip all the LF till CRLF is first in line).
    if(dwCurrentChunkSizeLeft == 0U)
    {
        BOOL bEmptyLine = TRUE;
        do
        {
            bEmptyLine = TRUE;

            do
            {
                curChar = pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff];

                if(IsEOL(curChar) == FALSE)
                    bEmptyLine = FALSE;

                ++dwPosInCurrDataBuff;
                if(AdjustEndOfBuff(pCurrDataBuff, dwPosInCurrDataBuff) == FALSE && bEmptyLine == FALSE)
                    return FALSE;
            }
            while(curChar != '\n');
        }
        while(bEmptyLine == FALSE);
    }

    return TRUE;
}


// ----------------------------------------------------------------------------
//
// Read and skip the chunk size *this is pointing to.
// If not enough data available, then return FALSE.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::ckReadChunkSize(BOOL bSkipLastCkCRLF/*=TRUE*/)
{
    return ckLocalReadChunkSize(m_pCurrDataBuff, m_dwPosInCurrDataBuff, m_dwCurrentChunkSizeLeft, bSkipLastCkCRLF);
}


// ----------------------------------------------------------------------------
//
// Move the cursor to the next chunk.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::ckMoveNextChunk()
{
    // Search for the right buffer (a chunk may occupy several buffers).

    DWORD dwCurrBuffSizeLeft = m_pCurrDataBuff->GetUsedDataSize() - m_dwPosInCurrDataBuff;

    while(m_dwCurrentChunkSizeLeft >= dwCurrBuffSizeLeft)
    {
        // Meets end of buffer before end of chunk. Move to the end of the current buffer.
        m_dwCurrentChunkSizeLeft -= dwCurrBuffSizeLeft;

        if(m_pCurrDataBuff->GetNextBuff() == FALSE)
            return FALSE;

        m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
        m_dwPosInCurrDataBuff = 0U;

        dwCurrBuffSizeLeft = m_pCurrDataBuff->GetUsedDataSize();
    }

    // Move to the end of the chunk.
    m_dwPosInCurrDataBuff += m_dwCurrentChunkSizeLeft;

    return ckReadChunkSize();
}


// ----------------------------------------------------------------------------
//
// Copy dwCharCount chars into destStr starting from current pos.
// Removes chunk information.
// Return FALSE if couldn't find enough characters.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::BuildString(std::string &destStr, DWORD dwCharCount) const
{
    destStr.erase(); destStr.reserve(dwCharCount);

    const CDataBuff * pDataBuffCurs = m_pCurrDataBuff;
    const CHAR * pAvDataBuff;
    DWORD dwCopyBytes;
    DWORD dwCurrBuffSizeLeft= m_pCurrDataBuff->GetUsedDataSize() - m_dwPosInCurrDataBuff;
    DWORD dwChunkSizeLeft   = m_dwCurrentChunkSizeLeft;

    while(TRUE)
    {
        // Copy till the end of dwCharCount or end of buffer or end of chunk, whatever comes first.
        dwCopyBytes = __min(__min(dwCharCount, dwCurrBuffSizeLeft), dwChunkSizeLeft);
        pAvDataBuff = &(pDataBuffCurs->GetDataPtr()[pDataBuffCurs->GetUsedDataSize() - dwCurrBuffSizeLeft]);
        destStr.append(pAvDataBuff, dwCopyBytes);

        dwCharCount -= dwCopyBytes;

        if(dwCharCount == 0U)
            return TRUE;
        else
        {
            // Move to the next buffer or chunk depending on the one we stopped at.
            dwChunkSizeLeft     -= dwCopyBytes;
            dwCurrBuffSizeLeft  -= dwCopyBytes;

            // Ajust values.
            if(dwCurrBuffSizeLeft == 0U)
            {
                // Move to the next buffer.
                pDataBuffCurs = pDataBuffCurs->GetNextBuff();

                if(pDataBuffCurs == NULL)
                    return FALSE;

                dwCurrBuffSizeLeft = pDataBuffCurs->GetUsedDataSize();
            }

            if(dwChunkSizeLeft == 0U)
            {
                // Move to the next chunk.

                // Read new chunk size.
                DWORD dwPosInBuff = pDataBuffCurs->GetUsedDataSize() - dwCurrBuffSizeLeft;
                if(ckLocalReadChunkSize(pDataBuffCurs, dwPosInBuff, dwChunkSizeLeft) == FALSE)
                    return FALSE;
                dwCurrBuffSizeLeft = pDataBuffCurs->GetUsedDataSize() - dwPosInBuff;
            }
        }
    }
}

// ----------------------------------------------------------------------------
//
// Move to the next char.
// Return FALSE if no value could be accessed.
//
// ----------------------------------------------------------------------------

__inline BOOL ckLocalMoveNextChar(const CDataBuff *&pCurrDataBuff, DWORD &dwPosInCurrDataBuff, DWORD &dwCurrentChunkSizeLeft)
{
    // Move to the next char.
    ++dwPosInCurrDataBuff;
    assert(dwCurrentChunkSizeLeft != 0U);
    --dwCurrentChunkSizeLeft;

    // Eventually move to the next data buffer.
    if(dwPosInCurrDataBuff == pCurrDataBuff->GetUsedDataSize())
    {
        // No data left to read in the current buffer. Try to move to the next.
        if(pCurrDataBuff->GetNextBuff() == NULL)
            return FALSE;

        pCurrDataBuff = pCurrDataBuff->GetNextBuff();
        dwPosInCurrDataBuff = 0U;
    }

    // Eventually move to the next chunk.
    if(dwCurrentChunkSizeLeft == 0U)
    {
        // Moving to the next chunk. Read the chunk size.
        if(ckLocalReadChunkSize(pCurrDataBuff, dwPosInCurrDataBuff, dwCurrentChunkSizeLeft) == FALSE)
            return FALSE;
    }

    return TRUE;
}


// ----------------------------------------------------------------------------
//
// Read the next char in buffs without moving the iterator.
// Return FALSE if no value could be read.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::NextChar(CHAR &nextChar) const
{
    const CDataBuff * pCurrDataBuff     = m_pCurrDataBuff;
    DWORD       dwPosInCurrDataBuff     = m_dwPosInCurrDataBuff;
    DWORD       dwCurrentChunkSizeLeft  = m_dwCurrentChunkSizeLeft;

    if(ckLocalMoveNextChar(pCurrDataBuff, dwPosInCurrDataBuff, dwCurrentChunkSizeLeft) == FALSE)
        return FALSE;

    // Read the following char.
    nextChar = pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff];

    return TRUE;
}

// ----------------------------------------------------------------------------
//
// Move to the next char or return FALSE if reached the end of the buffers list.
// m_dwPosInCurrDataBuff may then point to a position exceeding the current buff size
// and m_dwCurrentChunkSizeLeft may evaluate to zero.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::MoveNextChar()
{
    return ckLocalMoveNextChar(m_pCurrDataBuff, m_dwPosInCurrDataBuff, m_dwCurrentChunkSizeLeft);
}


// ----------------------------------------------------------------------------
//
// Depending on 'bCaseSensitive', performs a case-sensitive or not-sensitive 
// search of the 'dwStrLen'-size string 'str' in buffers delimitted by 'this' 
// and 'IterStopAt'.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::ContainString(const CHAR *str, DWORD dwStrLen, const CDataBuffListChunkAccessor &iterStopAt, BOOL bCaseSensitive)   const
{
    CDataBuffListChunkAccessor baseIter = *this;     // The one used for come-backs.
    CDataBuffListChunkAccessor movingIter = *this;   // The one that iterates the buffs.
    DWORD dwMatchedChars = 0U;
    DWORD dwToMatchChars;
    int iCmpRes = 1;

    while(dwMatchedChars < dwStrLen)
    {
        // Compare up to dwMatchedChars or end of current buff or end of current chunk, whichever comes first.
        dwToMatchChars = __min(__min(dwStrLen-dwMatchedChars, movingIter.m_dwPosInCurrDataBuff-movingIter.m_pCurrDataBuff->GetUsedDataSize()), movingIter.m_dwCurrentChunkSizeLeft);

        // Check if we reach iterStopAt.
        if(movingIter.m_pCurrDataBuff == iterStopAt.m_pCurrDataBuff)
        {
            if(movingIter.m_dwPosInCurrDataBuff + dwToMatchChars > iterStopAt.m_dwPosInCurrDataBuff)
            {
                return FALSE;
            }
        }

        if(bCaseSensitive)
            iCmpRes = memcmp(&movingIter.m_pCurrDataBuff->GetDataPtr()[movingIter.m_dwPosInCurrDataBuff], &str[dwMatchedChars], dwToMatchChars);
        else
            iCmpRes = _memicmp(&movingIter.m_pCurrDataBuff->GetDataPtr()[movingIter.m_dwPosInCurrDataBuff], &str[dwMatchedChars], dwToMatchChars);

        if(iCmpRes == 0U)
        {
            // Matched! Two issues now: either comparison is over or we're moving to the next buffer to find the remaining of str.
            dwMatchedChars += dwToMatchChars;

            if(dwMatchedChars == dwStrLen)
            {
                // String entirely found.
                return TRUE;
            }
            else
            {
                // Must look in the next buffer and/or the next chunk for the remaining of the string.
                movingIter.m_dwCurrentChunkSizeLeft -= dwToMatchChars;
                movingIter.m_dwPosInCurrDataBuff    += dwToMatchChars;

                if(AdjustEndOfBuff(movingIter.m_pCurrDataBuff, movingIter.m_dwPosInCurrDataBuff) == FALSE)
                    return FALSE;

                if(movingIter.m_dwCurrentChunkSizeLeft == 0U)
                {
                    if(movingIter.ckReadChunkSize() == FALSE)
                        return FALSE;
                }
            }
        }
        else
        {
            // Found a difference. Try again starting with the next base-character.
            if(baseIter.MoveNextChar() == FALSE)
                return FALSE;
            movingIter = baseIter;
            dwMatchedChars = 0U;
        }
    }

    return dwMatchedChars == dwStrLen;
}

// ----------------------------------------------------------------------------
//
// Finds or skips the str string.
// If not found, returns FALSE and set *this so that it is possible to continue the search if buffers are
// added. That way, if the begining of the searched string is found but chars are missing, the next 
// search will start right at that position.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLCA>::FindSkipStr(LPCSTR str, DWORD dwStrLen, EFindSkip findSkip, BOOL bCaseSensitive)
{
    CDataBuffListChunkAccessor baseIter = *this;     // The one used for come-backs.
    DWORD dwMatchedChars = 0U;
    DWORD dwToMatchChars;
    int iCmpRes;

    while(TRUE)
    {
        if(m_dwCurrentChunkSizeLeft == 0U)
            return FALSE;

        dwToMatchChars = __min(__min(dwStrLen-dwMatchedChars, m_dwPosInCurrDataBuff-m_pCurrDataBuff->GetUsedDataSize()), m_dwCurrentChunkSizeLeft);

        if(bCaseSensitive)
            iCmpRes = memcmp(&(m_pCurrDataBuff->GetDataPtr()[m_dwPosInCurrDataBuff]), &(str[dwMatchedChars]), dwToMatchChars);
        else
            iCmpRes = _memicmp(&(m_pCurrDataBuff->GetDataPtr()[m_dwPosInCurrDataBuff]), &(str[dwMatchedChars]), dwToMatchChars);

        if(iCmpRes == 0U)
        {
            // Matched! Two issues now: either comparison is over or we're moving to the next buffer to find the remaining of str.
            dwMatchedChars += dwToMatchChars;

            if(dwMatchedChars == dwStrLen)
            {
                // String entirely found.
                if(findSkip == FIND)
                {
                    // Were searching for the begining of the string.
                    operator=(baseIter);
                    return TRUE;
                }
                else
                {
                    // Must skip the found string.
                    m_dwCurrentChunkSizeLeft -= dwToMatchChars;
                    m_dwPosInCurrDataBuff    += dwToMatchChars;

                    // Adjust iterator.
                    if(AdjustEndOfBuff(m_pCurrDataBuff, m_dwPosInCurrDataBuff) == FALSE)
                    {
                        operator=(baseIter);    // Move back.
                        return FALSE;
                    }

                    if(m_dwCurrentChunkSizeLeft == 0U)
                    {
                        if(ckReadChunkSize() == FALSE)
                        {
                            operator=(baseIter);    // Move back.
                            return FALSE;
                        }
                    }

                    return TRUE;
                }
            }
            else    // Found only the begining of the string.
            {
                // Must look in the next buffer and/or the next chunk for the remaining of the string.
                assert(m_dwCurrentChunkSizeLeft != 0U);
                --m_dwCurrentChunkSizeLeft;
                ++m_dwPosInCurrDataBuff;

                if(AdjustEndOfBuff(m_pCurrDataBuff, m_dwPosInCurrDataBuff) == FALSE)
                {
                    operator=(baseIter);    // Move back.
                    return FALSE;
                }

                if(m_dwCurrentChunkSizeLeft == 0U)
                {
                    if(ckReadChunkSize() == FALSE)
                    {
                        operator=(baseIter);    // Move back.
                        return FALSE;
                    }
                }
            }
        }
        else
        {
            // Found a difference. Try again starting with the next base-character.
            if(baseIter.MoveNextChar() == FALSE)
            {
                operator=(baseIter);    // Move back.
                return FALSE;
            }
            operator=(baseIter);
            dwMatchedChars = 0U;
        }
    }
}



/* #########################################################################

CDataBuffListAccessor

  Define the specialisation of the methods in CDataBuffListBaseAccessor
  for normal buffers.

######################################################################### */


template <>
WILL_INLINE CDataBuffListBaseAccessor<CdBLA>::CDataBuffListBaseAccessor<CdBLA>(CDataBuffList *pDataBuffList)
        : m_dwPosInCurrDataBuff(0), m_pDataBuffList(pDataBuffList)
{
    if(pDataBuffList != NULL)
    {
        m_pCurrDataBuff = m_pDataBuffList->GetBuffList();
    }
    // Otherwise, will have to call AdjustPointers() when the data is available.
}

template <>
WILL_INLINE void CDataBuffListBaseAccessor<CdBLA>::MoveIterTo(const CDataBuffListBaseAccessor<CdBLA> &dBLA)
{
    m_pDataBuffList             = dBLA.m_pDataBuffList;
    m_pCurrDataBuff             = dBLA.m_pCurrDataBuff;
    m_dwPosInCurrDataBuff       = dBLA.m_dwPosInCurrDataBuff;
    m_dwCurrentChunkSizeLeft    = 0U;
}

// Reset the iterator.
template <>
WILL_INLINE void CDataBuffListBaseAccessor<CdBLA>::Rewind()
{
    assert(m_pDataBuffList != NULL);

    m_pCurrDataBuff = m_pDataBuffList->GetBuffList();
    m_dwPosInCurrDataBuff = 0U;
}


// ----------------------------------------------------------------------------
//
// Adjust the iterator position.
// Set the iterator to point at the data contained in the CDataBuffList.
//
// Note that we should not continue using that iterator until AdjustIterator 
// returns TRUE.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::AdjustIterator()
{
    assert(m_pDataBuffList != NULL);

    if(m_pCurrDataBuff == NULL)
    {
        // The iterator was created before we received the first data buffers.
        m_pCurrDataBuff = m_pDataBuffList->GetBuffList();
        m_dwPosInCurrDataBuff = 0U;
    }

    if(m_pCurrDataBuff == NULL)
        return FALSE;

    if(AdjustEndOfBuff(m_pCurrDataBuff, m_dwPosInCurrDataBuff) == FALSE)
    {
        return FALSE;
    }

    return TRUE;
}


// Try to move the iterator dwItems items further. Returns the number of items it skipped actually
// (may be less than dwItems if not enough data is available).
// The three cases are: move in the current buffer; must move to another buffer; must move to the 
// next buffer, but not received yet.

template <>
WILL_INLINE DWORD CDataBuffListBaseAccessor<CdBLA>::Skip(DWORD dwItems)
{
    assert(m_pDataBuffList != NULL);

    DWORD dwToSkip = dwItems;
    DWORD dwCurBuffUsedSize;

    while(dwToSkip > 0)
    {
        dwCurBuffUsedSize = m_pCurrDataBuff->GetUsedDataSize();

        if(dwToSkip >= dwCurBuffUsedSize - m_dwPosInCurrDataBuff)
        {
            // Current buffer not wide enough. Must skip till one of the next buffers.
            dwToSkip -= dwCurBuffUsedSize - m_dwPosInCurrDataBuff;

            // Reached the end of the current buffer. Move the cursor to the next buffer.
            // The current buffer is either empty or we must use the next buffer.
            if(m_pCurrDataBuff->GetNextBuff() != NULL)
            {
                m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
                m_dwPosInCurrDataBuff = 0U;
            }
            else
            {
                // No other buffers available.
                m_dwPosInCurrDataBuff = dwCurBuffUsedSize;
                return dwItems - dwToSkip;
            }
        }
        else
        {
            // Move cursor in the current buffer.
            m_dwPosInCurrDataBuff += dwToSkip;
            dwToSkip = 0U;
        }
    }

    return dwItems;
}


/*
// Return TRUE if the current object represents a position "before" the other object.

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::IsBefore(const CDataBuffListBaseAccessor<CdBLA> &other) const
{
    assert(m_pCurrDataBuff != NULL && other.m_pCurrDataBuff != NULL);

    // Walk forward from the current buffer to the other buffer or NULL.
    for(const CDataBuff *pOtherDataBuff = m_pCurrDataBuff;
        pOtherDataBuff != NULL && pOtherDataBuff != other.m_pCurrDataBuff;
        pOtherDataBuff = pOtherDataBuff->GetNextBuff())
            ;

    return
        (pOtherDataBuff == m_pCurrDataBuff ?
            (m_dwPosInCurrDataBuff < other.m_dwPosInCurrDataBuff)   // Same buffer: compare positions.
        :
            (pOtherDataBuff == other.m_pCurrDataBuff)               // The other buffer is further in the sequence.
        );
}
*/

// Compute the number of items 'first' should be moved forward in order to find 'second'.
// Returns ULONG_MAX if could not find 'second' or if an error occurred.

template <>
WILL_INLINE DWORD CDataBuffListBaseAccessor<CdBLA>::ComputeDistance(const CDataBuffListBaseAccessor<CdBLA> &first, const CDataBuffListBaseAccessor<CdBLA> &second) const
{
    assert(first.m_pDataBuffList->GetBuffList() == second.m_pDataBuffList->GetBuffList());

    const CDataBuff *pFirstDataBuff = first.m_pCurrDataBuff;
    const CDataBuff *pSecondDataBuff = second.m_pCurrDataBuff;

    if(pFirstDataBuff == pSecondDataBuff)
    {
        // Pointers to the same buffers.
        if(second.m_dwPosInCurrDataBuff < first.m_dwPosInCurrDataBuff)
            return ULONG_MAX;
        else
            return second.m_dwPosInCurrDataBuff-first.m_dwPosInCurrDataBuff;
    }
    else
    {
        // Search in the buffers following *this.

        DWORD dwRes = first.m_pCurrDataBuff->GetUsedDataSize()-first.m_dwPosInCurrDataBuff;

        pFirstDataBuff = pFirstDataBuff->GetNextBuff();
        if(pFirstDataBuff == NULL)
            return ULONG_MAX;

        while(pFirstDataBuff != second.m_pCurrDataBuff)
        {
            dwRes += pFirstDataBuff->GetUsedDataSize();
            pFirstDataBuff = pFirstDataBuff->GetNextBuff();

            if(pFirstDataBuff == NULL)
                return ULONG_MAX;
        }

        return dwRes + second.m_dwPosInCurrDataBuff;
    }
}


// Copy dwCharCount chars into destStr starting from current pos.
// Return FALSE if couldn't find enough characters.

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::BuildString(std::string &destStr, DWORD dwCharCount) const
{
    destStr.erase(); destStr.reserve(dwCharCount);

    const CDataBuff *pDataBuffCurs = m_pCurrDataBuff;
    const CHAR      *pAvDataBuff = &(pDataBuffCurs->GetDataPtr()[m_dwPosInCurrDataBuff]);
    DWORD dwCopyBytes;
    DWORD dwDataAvailable = m_pCurrDataBuff->GetUsedDataSize() - m_dwPosInCurrDataBuff;

    while(TRUE)
    {
        dwCopyBytes = __min(dwCharCount, dwDataAvailable);
        destStr.append(pAvDataBuff, dwCopyBytes);

        dwCharCount -= dwCopyBytes;

        if(dwCharCount == 0U)
            return TRUE;
        else
        {
            pDataBuffCurs = pDataBuffCurs->GetNextBuff();

            if(pDataBuffCurs == NULL)
                return FALSE;

            pAvDataBuff     = pDataBuffCurs->GetDataPtr();
            assert(pAvDataBuff != NULL);
            dwDataAvailable = pDataBuffCurs->GetUsedDataSize();
        }
    }
}


// Copy dwCharCount chars into destStr starting from current pos.
// Return FALSE if couldn't find enough characters.
// Won't add a \0 to the end of string.

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::BuildString(LPSTR destStr, DWORD dwCharCount)   const
{
    const CDataBuff *pDataBuffCurs = m_pCurrDataBuff;
    const CHAR      *pAvDataBuff = &(pDataBuffCurs->GetDataPtr()[m_dwPosInCurrDataBuff]);
    DWORD   dwCopyBytes;
    DWORD   dwDataAvailable = m_pCurrDataBuff->GetUsedDataSize() - m_dwPosInCurrDataBuff;
    LPSTR   pCurrStrPos = destStr;

    while(TRUE)
    {
        dwCopyBytes = __min(dwCharCount, dwDataAvailable);
        CopyMemory(pCurrStrPos, pAvDataBuff, dwCopyBytes);
        pCurrStrPos += dwCopyBytes;

        dwCharCount -= dwCopyBytes;

        if(dwCharCount == 0U)
            return TRUE;
        else
        {
            pDataBuffCurs = pDataBuffCurs->GetNextBuff();

            if(pDataBuffCurs == NULL)
                return FALSE;

            pAvDataBuff     = pDataBuffCurs->GetDataPtr();
            assert(pAvDataBuff != NULL);
            dwDataAvailable = pDataBuffCurs->GetUsedDataSize();
        }
    }
}

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::ReadNextWord(std::string &destStr) const
{
    destStr.erase();

    const CDataBuff *pDataBuffCurs = m_pCurrDataBuff;
    DWORD dwPosInCurrDataBuff = m_dwPosInCurrDataBuff;

    CHAR cCurChar;

    while(TRUE)
    {
        cCurChar = pDataBuffCurs->GetDataPtr()[dwPosInCurrDataBuff];
        if(IsWS(cCurChar) == FALSE && IsEOL(cCurChar) == FALSE)
        {
            destStr += cCurChar;

            dwPosInCurrDataBuff++;
            if(dwPosInCurrDataBuff == pDataBuffCurs->GetUsedDataSize())
                pDataBuffCurs = pDataBuffCurs->GetNextBuff();

            if(pDataBuffCurs == NULL)
                return FALSE;
        }
        else
        {
            return (dwPosInCurrDataBuff != m_dwPosInCurrDataBuff);
        }
    }
}

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::ReadNextWord(DWORD &dwDest) const
{
    dwDest = 0U;

    const CDataBuff *pDataBuffCurs = m_pCurrDataBuff;
    DWORD   dwPosInCurrDataBuff = m_dwPosInCurrDataBuff;

    CHAR cCurChar;

    while(TRUE)
    {
        cCurChar = pDataBuffCurs->GetDataPtr()[dwPosInCurrDataBuff];
        if(IsDigit(cCurChar) == TRUE)
        {
            dwDest *= 10U;
            dwDest += cCurChar - '0';

            dwPosInCurrDataBuff++;
            if(dwPosInCurrDataBuff == pDataBuffCurs->GetUsedDataSize())
                pDataBuffCurs = pDataBuffCurs->GetNextBuff();

            if(pDataBuffCurs == NULL)
                return FALSE;
        }
        else
        {
            return (dwPosInCurrDataBuff != m_dwPosInCurrDataBuff);
        }
    }
}


// Read the next char in buffs.
// Return FALSE if no value could be read.

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::NextChar(CHAR &nextChar) const
{
    assert(m_pCurrDataBuff);

    const CDataBuff * pCurrDataBuff = m_pCurrDataBuff;
    DWORD       dwPosInCurrDataBuff = m_dwPosInCurrDataBuff;

    dwPosInCurrDataBuff++;

    // Make sure the following char is available.
    if(dwPosInCurrDataBuff == pCurrDataBuff->GetUsedDataSize())
    {
        // No data left to read in the current buffer. Try to move to the next.
        if(pCurrDataBuff->GetNextBuff() != NULL)
        {
            pCurrDataBuff = pCurrDataBuff->GetNextBuff();
            dwPosInCurrDataBuff = 0U;
        }
        else
        {
            return FALSE;
        }
    }

    // Read the following char.
    nextChar = pCurrDataBuff->GetDataPtr()[dwPosInCurrDataBuff];

    return TRUE;
}


// Move to the next char or does not move and return FALSE if reached the end 
// of the buffers list.

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::MoveNextChar()
{
    assert(m_pCurrDataBuff != NULL);

    m_dwPosInCurrDataBuff++;

    if(AdjustEndOfBuff(m_pCurrDataBuff, m_dwPosInCurrDataBuff) == FALSE)
        return FALSE;

    return TRUE;
}


// Depending on bCaseSensitive, performs a case-sensitive or insensitive search of str in buffers delimitted by this and IterStopAt.
// TODO: merge with FindSkipStr ?
template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::ContainString(const CHAR *str, DWORD dwStrLen, const CDataBuffListAccessor &iterStopAt, BOOL bCaseSensitive) const
{
    CDataBuffListAccessor baseIter = *this;     // The one used for come-backs.
    CDataBuffListAccessor movingIter = *this;   // The one that iterates the buffs.
    DWORD dwMatchedChars = 0U;
    DWORD dwToMatchChars;
    int iCmpRes;

    while(TRUE)
    {
        dwToMatchChars = __min(dwStrLen, movingIter.m_dwPosInCurrDataBuff-movingIter.m_pCurrDataBuff->GetUsedDataSize());

        // Check if we reach iterStopAt.
        if(movingIter.m_pCurrDataBuff == iterStopAt.m_pCurrDataBuff)
        {
            if(movingIter.m_dwPosInCurrDataBuff + dwToMatchChars > iterStopAt.m_dwPosInCurrDataBuff)
            {
                return FALSE;
            }
        }

        if(bCaseSensitive)
            iCmpRes = memcmp(&movingIter.m_pCurrDataBuff->GetDataPtr()[movingIter.m_dwPosInCurrDataBuff], &str[dwMatchedChars], dwToMatchChars);
        else
            iCmpRes = _memicmp(&movingIter.m_pCurrDataBuff->GetDataPtr()[movingIter.m_dwPosInCurrDataBuff], &str[dwMatchedChars], dwToMatchChars);

        if(iCmpRes == 0)
        {
            // Matched! Two issues now: either comparison is over or we're moving to the next buffer to find the remaining of str.
            dwMatchedChars += dwToMatchChars;

            if(dwMatchedChars == dwStrLen)
            {
                // String entirely found.
                return TRUE;
            }
            else
            {
                // Must look in the next buffer for the remaining of the string.
                if(movingIter.MoveNextBuff() == FALSE)
                    return FALSE;
            }
        }
        else
        {
            // Found a difference.
            if(baseIter.MoveNextChar() == FALSE)
                return FALSE;
            movingIter = baseIter;
            dwMatchedChars = 0U;
        }
    }
}

// Finds or skips the str string.
// If not found, returns FALSE and set *this so that it is possible to continue the search if buffers are
// added. That way, if the begining of the searched string is found but chars are missing, the next 
// search will start right at that position.
template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::FindSkipStr(LPCSTR str, DWORD dwStrLen, EFindSkip findSkip, BOOL bCaseSensitive)
{
    CDataBuffListAccessor baseIter = *this;     // The one used for come-backs.
    DWORD dwMatchedChars = 0U;
    DWORD dwToMatchChars;
    int iCmpRes;

    while(TRUE)
    {
        dwToMatchChars = __min(dwStrLen-dwMatchedChars, m_dwPosInCurrDataBuff-m_pCurrDataBuff->GetUsedDataSize());

        if(bCaseSensitive)
            iCmpRes = memcmp(&(m_pCurrDataBuff->GetDataPtr()[m_dwPosInCurrDataBuff]), &(str[dwMatchedChars]), dwToMatchChars);
        else
            iCmpRes = _memicmp(&(m_pCurrDataBuff->GetDataPtr()[m_dwPosInCurrDataBuff]), &(str[dwMatchedChars]), dwToMatchChars);

        if(iCmpRes == 0)
        {
            // Matched! Two issues now: either comparison is over or we're moving to the next buffer to find the remaining of str.
            dwMatchedChars += dwToMatchChars;

            if(dwMatchedChars == dwStrLen)
            {
                // String entirely found.
                if(findSkip == FIND)
                {
                    // Move back to the begining of the found string.
                    operator=(baseIter);
                    return TRUE;
                }
                else
                {
                    // Must skip the found string.
                    m_dwPosInCurrDataBuff += dwStrLen;
                    return TRUE;
                }
            }
            else
            {
                // Must look in the next buffer for the remaining of the string.
                if(MoveNextBuff() == FALSE)
                    return FALSE;
            }
        }
        else
        {
            // Found a difference. Try with the next character.
            if(baseIter.MoveNextChar() == FALSE)
                return FALSE;
            operator=(baseIter);
            dwMatchedChars = 0U;
        }
    }
}

template <>
WILL_INLINE BOOL CDataBuffListBaseAccessor<CdBLA>::MoveNextBuff()
{
    if(m_pCurrDataBuff->GetNextBuff() == NULL)
    {
        // Caller must wait for more data before continuing.
        return FALSE;
    }
    else
    {
        m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();
        m_dwPosInCurrDataBuff = 0U;

        return TRUE;
    }
}

// ----------------------------------------------------------------------------
//
// Defined only for standard CDataBuffListAccessor.
// Move cursor past the end of the data buffers.
//
// ----------------------------------------------------------------------------

template <>
WILL_INLINE void CDataBuffListBaseAccessor<CdBLA>::SkipEnd()
{
    while(m_pCurrDataBuff->GetNextBuff() != NULL)
        m_pCurrDataBuff = m_pCurrDataBuff->GetNextBuff();

    m_dwPosInCurrDataBuff = m_pCurrDataBuff->GetUsedDataSize();
}


/* #########################################################################

    CDataBuffHandler

   ######################################################################### */

WILL_INLINE WSABUF* CDataBuffHandler::AllocWsaBuffList()
{
    // Reserve enough data buffers, at least WSABUF_COUNT to avoid accessing 
    // too often the global buffer pool and to leverage memory fragmentation.

    if(m_dwRemainingBuffsInMappedList < WSABUF_COUNT)
    {
        if(m_wsaMappedBuffList.AllocateBuffs(WSABUF_COUNT) == FALSE)
            return NULL;
    }

    // Map the data buffers to the WSA list.
    CDataBuff *pCurDataBuff = m_wsaMappedBuffList.GetBuffList();
    for(DWORD dw=0 ; dw<WSABUF_COUNT ; dw++)
    {
        assert(pCurDataBuff);
        m_wsaBuffArray[dw].buf = pCurDataBuff->GetDataPtr();
        m_wsaBuffArray[dw].len = GetDataBuffUsableSize();
        pCurDataBuff = pCurDataBuff->GetNextBuff();         // Nb: last call to MoveNextBuff will return FALSE 
                                                            // since reached the end of the list.
    }

    return m_wsaBuffArray;
}

// Store the buffers where data was stored and keep a list of the empty ones 
// for use by the next AllocWsaBuffList. Remaining buffs are stored in 
// m_wsaMappedBuffList.
WILL_INLINE void CDataBuffHandler::ReturnWsaBuffList(DWORD dwBytesStored)
{
    // Update the size of the WSA buffers into the matching data buffers.
    DWORD dwBuffLen;
    DWORD dwRelevantBuffCount = 0U;

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
        InterlockedExchangeAdd(CGlobalDataBuffHandler::s_pDataStoredCounterRef, dwBytesStored);
#   else
        (*CGlobalDataBuffHandler::s_pDataStoredCounterRef) += dwBytesStored;
#   endif
#endif

    CDataBuff *pLastUsedBuff = NULL;

    // Update the used memory in each CDataBuff involved.
    CDataBuffListAccessor dbListAcc(&m_wsaMappedBuffList);
    while(dwBytesStored != 0U)
    {
        dwBuffLen = __min(dwBytesStored, GetDataBuffUsableSize());

        dbListAcc.GetCurrDataBuff()->SetUsedDataSize(dwBuffLen);
        dwBytesStored -= dwBuffLen;

        if(dwBytesStored == 0U)
            pLastUsedBuff = dbListAcc.GetCurrDataBuff();

        dbListAcc.MoveNextBuff();
        dwRelevantBuffCount++;
    }

    // Store the buffers containing data, and keep the empty buffers for the next AllocWsaBuffList.
    m_dataBuffList.AddBuffList(m_wsaMappedBuffList, pLastUsedBuff, dwRelevantBuffCount, m_wsaMappedBuffList);

    assert(WSABUF_COUNT >= dwRelevantBuffCount);

    m_dwRemainingBuffsInMappedList = WSABUF_COUNT-dwRelevantBuffCount;
}

WILL_INLINE void CDataBuffHandler::ClearBuffer()
{
#ifdef USE_PERF_DATA
    CDataBuffListAccessor dbListAcc(&m_dataBuffList);
    while(dbListAcc.GetCurrDataBuff() != NULL)
    {
#   ifdef LOCK_COUNTERS
        InterlockedExchangeAdd(CGlobalDataBuffHandler::s_pDataStoredCounterRef, -(LONG)dbListAcc.GetCurrDataBuff()->GetUsedDataSize());
#   else
        (*CGlobalDataBuffHandler::s_pDataStoredCounterRef) -= (LONG)dbListAcc.GetCurrDataBuff()->GetUsedDataSize();
#   endif
        if(dbListAcc.MoveNextBuff() == FALSE)
            break;
    }
#endif

    m_dataBuffList.ReleaseBuffList();
    m_wsaMappedBuffList.ReleaseBuffList();
    m_dwRemainingBuffsInMappedList = 0U;
}

WILL_INLINE std::ostream &operator <<(std::ostream &os, CDataBuffHandler &dbHdl)
{
    os << dbHdl.GetWholeBuffList();
    return os;
}
