/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
#ifndef __NODE_ADDRESS__
#define __NODE_ADDRESS__

#include "TagParsing.hpp"
#include <string>
#include <vector>
#include <list>
#include <sstream>


///////////////////////////////////////////////////////////////////////////////////////
// Address string format:
// ---------------------
//
// <tag_list>:<addresstype>:[<attribute_name>](<childnumber>)
//
// '<' and '>' delimit a symbolic address element,
// '[' and ']' delimit an optional part.
// addresstype: one of "TEXT", "SCRIPT", "COMMENT", "ANONYMOUS ATTRIBUTE" and "ATTRIBUTE".
// attribute_name: the name of an attribute (when not anonymous), like "HTTP-EQUIV".
// childnumber: the child index number of the attribute/text/comment/script. For attributes,
//              this number is relative to the parent tag.
// tag_list: a slash-separated list of nodes. each node being:
//  <tag_label>(<tag_childnumber>)
// tag_childnumber: the child index number of the tag.
//
// where tag_label is the label of an HTML tag, as "SELECT", or "HEAD".
//
// Sample address:
//
// HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)
///////////////////////////////////////////////////////////////////////////////////////

struct AddressNode
{
    AddressNode() : iChild_(0), eTag_(TAG_NONE) {}
    AddressNode(const AddressNode& an) : iChild_(an.iChild_), eTag_(an.eTag_) {}
    AddressNode(int iChild, Tags eTag) : iChild_(iChild), eTag_(eTag) {}
    void operator=(const AddressNode& an) {iChild_ = an.iChild_; eTag_ = an.eTag_;}
    BOOL operator==(const AddressNode& an) const {return iChild_ == an.iChild_ && eTag_ == an.eTag_;}

    int     iChild_;                    // The child index number of the current node.
    Tags    eTag_;                      // If lToken_ is a tag, this contains its type.
};
typedef std::vector<AddressNode> TagHierarchy;
struct Address
{
    enum AddressType
    {
        ADDRESS_NONE,                   // For empty addresses.
        ADDRESS_TEXT,                   // A text.
        ADDRESS_SCRIPT,                 // A script.
        ADDRESS_COMMENT,                // A comment.
        ADDRESS_ANONYMOUS_ATTRIBUTE,    // An anonymous attribute.
        ADDRESS_ATTRIBUTE,              // A named attribute.
        ADDRESS_LAST
    };

    BOOL IsValidAddress() const {return addrType_ != ADDRESS_NONE;}
    void Parse(std::string& strAddr);
    int GetChildNumber() const;
    void SetChildNumber(AddressType addressType, int iChild);
    Address() : addrType_(ADDRESS_NONE), iCurChild_(-1), iCurChildAttribute_(-1) {}
    Address(const Address& a) : tagHierarchy_(a.tagHierarchy_), addrType_(a.addrType_), iCurChild_(a.iCurChild_), iCurChildAttribute_(a.iCurChildAttribute_) {}
    Address(const char* zStringAddress);
    void operator=(const Address& a) {tagHierarchy_ = a.tagHierarchy_; addrType_ = a.addrType_; iCurChild_ = a.iCurChild_; iCurChildAttribute_ = a.iCurChildAttribute_; strAtt_ = a.strAtt_;}
    BOOL operator==(const Address& a) const {return tagHierarchy_ == a.tagHierarchy_ && addrType_ == a.addrType_ && GetChildNumber() == a.GetChildNumber() && strAtt_ == a.strAtt_;}
    void PushNode(int iChild, Tags eTag) {tagHierarchy_.push_back(AddressNode(iChild,eTag));}
    void PushNode(const AddressNode& node) {tagHierarchy_.push_back(node);}
    static const char* AddressTypeToString(AddressType addrType);
    static AddressType StringToAddressType(const std::string& strAddrType);
    std::string ToString() const;
#ifdef _DEBUG
    void Dump(std::ostream& os) const;
#endif

    TagHierarchy  tagHierarchy_;        // Stack containing the tag context.
    AddressType   addrType_;            // The kind of information this address refers to.
    int           iCurChild_;
    int           iCurChildAttribute_;

    // Field used only when the address refers to a named attribute, i.e., addrType_ == ADDRESS_ATTRIBUTE.
    std::string   strAtt_;              // Attribute name.
};

// Inline definitions.
inline int Address::GetChildNumber() const
{
    int iChild;
    switch(addrType_)
    {
        case ADDRESS_ANONYMOUS_ATTRIBUTE:
        case ADDRESS_ATTRIBUTE:
            iChild = iCurChildAttribute_;
            break;
        case ADDRESS_TEXT:
        case ADDRESS_SCRIPT:
        case ADDRESS_COMMENT:
        default:
            iChild = iCurChild_;
            break;
    }
    return iChild;
}

inline void Address::SetChildNumber(AddressType addressType, int iChild)
{
    switch(addressType)
    {
        case ADDRESS_ANONYMOUS_ATTRIBUTE:
        case ADDRESS_ATTRIBUTE:
            iCurChildAttribute_ = iChild;
            break;
        case ADDRESS_TEXT:
        case ADDRESS_SCRIPT:
        case ADDRESS_COMMENT:
        default:
            iCurChild_ = iChild;
            break;
    }
}

///////////////////////////////////////////////////////////////////////////////////////
// Address query string formats:
// ----------------------------
//
// 1. Address only
//    ------------
// This format specifies nothing more than an address. In this case, the whole HTML
// element must ne retrieved.
//
// 2. Position + length
//    -----------------
// <address_string>'['<position>, <length>']'
//
// address_string: address string to the format described above.
// position: the zero-based starting index of the substring to extract.
// length: the number of characters of the substring to extract.
//
// Semantics: extract the substring beginning at position <position>
// inclusive, and having <length> characters.
// NOTES:
// - If the position is invalid, then a null result is returned.
// - If length = 0 or is invalid, then all characters after <position> are taken.
// Sample: HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)[2,5]
//
// 3. Delimiters
//    ----------
// <address_string>'['<delimiter1>,<delimiter2>']'
// address_string: address string to the format described above.
// delimiter1,delimiter2: single-quoted character strings to be matched
//
// Semantics: extract the substring beginning at (and containing) delimiter1, and
// ending with (and containing) delimiter2.
// NOTES:
// - If delimiter1 could not be found, a null result is returned.
// - If delimiter2 could not be found, then all characters after delimiter1 are taken.
// Sample: HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)['document.cookie=',';']
//
// 4. Mixed notation
//    --------------
// <address_string>'['<delimiter1>,<length>']'
// or
// <address_string>'['<position>,<delimiter2>']'
// address_string: address string to the format described above.
// delimiter1,delimiter2: single-quoted character strings to be matched
// position: the zero-based starting index of the substring to extract.
// length: the number of characters of the substring to extract.
//
// NOTES:
// - If delimiter1 could not be found, a null result is returned.
// - If the position is invalid, then a null result is returned.
// - If delimiter2 could not be found, then all characters after delimiter1 are taken.
// - If length = 0 or is invalid, then all characters after <position> are taken.
// Samples: HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)['document.cookie=',3]
//          HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)[2,';']
//
// 5. Excluding boundaries
//    --------------------
// Assuming that the default behavior is to include boundaries, the user may desire
// to exclude a boundary, be it a delimiter or a position. The way to specify
// exclusion is to use an opposite square bracket, like in open intervals.
// NOTE: excluding a length has no effect.
// Samples: HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)]2,';']
//              -> starting position is 3
//          HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)[2,';'[
//              -> result ends before (but does not contain) the semicolon
//          HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)]2,';'[
//              -> double exclusion
//          HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)['abc',5[
//              -> equivalent to: HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)['abc',5]
//
// 6. Offset
//    ------
// Within the target HTML element value's text, one may want to use all of the above
// capabilities, but starting after a given position. This position is an offset from
// the beginning of the HTML element's text. It is represented by an additional
// parameter, before the 2 boundaries. The offset can be specified by a position or a
// character string.
// If the offset begins with >, then it is excluded.
// If the offset begins with >=, then it is included.
// NOTE: If the offset string or position could not be found, a null result is returned.
// Samples: HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)]>'// Cookie','document.cookie=',';']
//              -> the zero position is set to the first character after '// Cookie',
//                 and from that position, search the first occurrence of 'document.cookie=';
//                 extract the substring just after that occurrence, up to (and excluding)
//                 the first semicolon.
//          HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)]>='// Cookie','document.cookie=',';']
//              -> same as above, except that the zero position is now the first character
//                 of '// Cookie'.
//          HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)]>=50,'document.cookie=',';']
//              -> here, the zero position is position 50.
//
// 7. Case sensitivity
//    ----------------
// The extensions presented so far have are by default case sensitive: an exact match
// is required for character and string delimiters. However, we can specify a case
// insensitive behavior for any string by prefixing it with an I.
// Sample: HTML(0)/HEAD(0)/META(1):ATTRIBUTE:CONTENT(1)]>I'// Cookie',I'document.cookie=',';']
//              -> The strings preceded by I must be matched in a case insensitive way.
//
// 8. Special cases
//    -------------
// To specify a single quote inside a single-quote delimited character string, precede it
// with a backslash.
// To specify a double quote inside a single-quote delimited character string, which can
// itself reside in an SCL string literal, just follow the rules of SCL, and write ~<22>
// instead of " in the SCL literal.
//
// NOTE: the syntax used for address query strings may not contain additional whitespace.
///////////////////////////////////////////////////////////////////////////////////////

struct AddressQuery
{
    AddressQuery(const std::string& strVariable, const char* zStringAddressQuery);

    BOOL IsValidAddress() const {return address_.IsValidAddress();}

    //// Data.
    std::string     strVariable_;           // The name of the script variable associated to the address.
    Address         address_;               // An address within an HTML page.
    enum Representation
    {
        IS_UNUSED,
        IS_NUMBER,
        IS_STRING
    };
    struct ob
    {
        ob():bIncluded(TRUE),rep(IS_UNUSED),iNumber(0){}

        BOOL bIncluded;                     // Exclusion flag: is the offset data included in the result?
        Representation rep;                 // Representation of the offset/boundary.
        std::string str;                    // Valid if rep is IS_STRING. String defining the offset/boundary.
        int  iNumber;                       // Valid if rep is IS_NUMBER. Number defining the offset/boundary.
                                            // For offset_ and begin_, this is a position. For end_, it's a length.
        BOOL bCaseSensitive;                // Valid if rep is IS_STRING. Case sensitivity of str.
    } offset_, begin_, end_;                // Offset, beginning and end boundaries.

private:
    //// Helpers.
    BOOL ExtractString(std::istringstream& is, std::string& str);
};

typedef std::list<AddressQuery*> AddressQueryList;

#endif  // __NODE_ADDRESS__
