/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     MonMain
//
// @module  Test Monitor Main | This is the main control module for a Test
//          Monitor.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#include <process.h>
#endif

// Include files.

#include "OSTAORB/OSTAORB.h" 
#include "TPLog.h"

#include "SmpExecuter_i.h"
#include "OtaDefs.h"
#include "OtaStandalone.h"

#include <string>
#include <new.h>
#include <CECInetInfo.h>
#include <strstream>
#include "mdbexcept.h"
#include "mdbgetopt.h"
#include "cxxutils.h"
#include "mdbtypes2.h"
#include <mthread.hpp>

extern    "C"
{
#include <psl_macros.h>
#include <psl_rtndef.h>
}

#ifndef VMS
using     std::string;
#endif

// Constants.

#define     VERSION      "3.0.0.0"
#define     ODLFILE      "smp.odl"

#ifdef CYRANO_NTPERF_MONITOR
const string TPTEXT = "TestExecuter_ntp";    // Used for Tracing filename
#define TESTPROCESSIMPL NtpExecuterImpl

#elif defined(CYRANO_SNMP_MONITOR)
const string TPTEXT = "TestExecuter_smp";    // Used for Tracing filename
#define TESTPROCESSIMPL SmpExecuterImpl

#else
#error "Undefined TestExecuter/Monitor"
#endif

// Local functions.

int Run (CORBAServer * pServer, string exename);
void sqlr_cleanup(void);

// Stop cleaning up occurring more than once at a time
static CTMutex cleanup_mutex;

//Global Variables
//Static: for use only here
static Cmdbgetopt *cl = NULL;
static bool standalone;
TScriptRule *rule;      // Need global access to this - is there a better way?

// The general idea here is to have all TestProcessImpl Access done in this file
// hopefully removing the need to have it as a fully global variable
// This should encourage minimal access from separate threads
#define TESTPROCESS         TestExecuter
#define TPROCESS            TExecuter
#define TESTPROCESS_PTR     TestExecuter_ptr
#define TPROCESS_PTR        TExecuter_ptr
#define TESTPROCESS_NIL     TestExecuter::_nil()

CTSemaphore blocking_semaphore(0);

static TESTPROCESSIMPL *TPROCESS = NULL;

// But just in case we do need to access it elsewhere here is a _ptr ptr
TESTPROCESS_PTR TPROCESS_PTR = TESTPROCESS_NIL;

// Thread control object pointer.
string exedir = "";

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Add a message to the Monitor Trace Log.
//
// @parm    Message category (input).
// @parm    Message severity level (input).
// @parm    Message text (input).
//
void OtaAddTrace (string category, long level, string msg)
{
    if (standalone)
    {
        if (level <= TRACE_RESULTS)
            printf("%s\n", msg.c_str());
    }
    else
    {
        logr->AddTrace(category.c_str(), level, msg.c_str());
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Monitor Main.
//
int main (int argc, char *argv[])
{
    logr = new TPLog();

    // Set handler for out of memory conditions in new operator.
    SET_NEW_HANDLER (mdb_new_handler);

    // Maps malloc failure to new fail handler on WIN32 only.
    SET_NEW_MODE

    // Initialise CORBA for logging purposes
    CORBAServer * pServer;
    try
    {
        pServer = new CORBAServer(argc, argv);
    }
    catch (...)
    {
        OtaAddTrace(STARTUP, ERRORLEVEL,
                    "Caught Exception on startup, is Nameservice Running?");
        sqlr_cleanup();
        return 1;
    }

    // Catch exceptions for command line errors
    try
    {
        // Create command line arg object. A general option syntax error exception
        // (string type) can be thrown here.
        cl = new Cmdbgetopt (argc, argv, "TestExecuter: ");

        // Define valid command line options and value types
        cl->define_option("-q", CLO_STRINGVAL);
        cl->define_option("-t", CLO_STRINGVAL);
        cl->define_option("-v", CLO_INTVAL);
        cl->define_option("-1", CLO_NOVAL);
        cl->define_option("-d", CLO_NOVAL);

        // Check option validity (string exception thrown on fail).
        cl->validate();

        // Get pid for this TestProcess
        MDBPID tppid = getpid();

        // Check for -q and, optionally, -t options, indicating that we are running
        // a single query standalone.
        if (cl->option_supplied("-q"))
        {
            standalone = true;
            string    query = cl->get_value_for("-q");
            string    target = "";

            if (cl->option_supplied("-t"))
            {
                target = cl->get_value_for("-t");
            }

            StandaloneQuery(query, target);
            return 0;
        }

        standalone = false;

        try
        {
            //**pServer->Ready(true);
            // Initialise Global Logging Object
            int       loglevel = cl->get_int_value_for("-v");
            CECInetInfo info;
            string    HostName = info.GetIPAddressAsString();
            strstream RegisteredLogName;

            RegisteredLogName << TPTEXT << "_" << HostName << "_" << tppid << ends;
            logr->Setup(pServer, RegisteredLogName.str(), loglevel);
            RegisteredLogName.freeze(false);
        }
        catch (...)
        {
            OtaAddTrace(STARTUP, ERRORLEVEL,
                        "Caught Exception on startup, is Nameservice Running?");
            sqlr_cleanup();
            return 1;
        }

        OtaAddTrace (STARTUP, ERRORLEVEL, "Start Execution");

        // Now that the logging mechanisms are in place check if just a request for version info

        if (cl->option_supplied("-1"))
        {
            cout << TPTEXT << " version " << VERSION << " build "
                 << __DATE__ << " " << __TIME__ << endl;
            OtaAddTrace(STARTUP, TRACE_PARTIALDIAG,
                        "Version information requested");
            sqlr_cleanup();
            return 0;
        }
    }
    catch (string es)
    {
        // Detected error while checking args
        OtaAddTrace (STARTUP, ERRORLEVEL, es);
        sqlr_cleanup ();
        return 1;
    }
    catch (exception & e)
    {
        OtaAddTrace(STARTUP, ERRORLEVEL, e.what());
        sqlr_cleanup();
        return 1;
    }
    catch (...)
    {
        // Some other unexpected exception caught
        OtaAddTrace(STARTUP, ERRORLEVEL,
                    "Caught exception while parsing command options");
        sqlr_cleanup();
        return 1;
    }

    // Read in atvf(tof) file to create the Script Rules, and commence execution.
    const char * const catchmsg = "Caught Exception in MonMain: ";
    bool caught = false;
    try
    {
        string exename;

        if (argc > 0)       // If we have a executable name use in logging, else don't bother
        {
            exename = filename_base(string(argv[0]));
        }
        else
        {
            exename = "";
        }

        // Get directory of this executable, path used to invoke is argv[0].
        exedir = filename_dir(string(argv[0]));

        // Run the Monitor image.
        Run(pServer, exename);

        // All done - perform any cleanup required
        sqlr_cleanup();
    }

    OTACATCHALL

    if (caught)
    {
        sqlr_cleanup();
        return 1;
    }

    return 0;

} // End of MAIN

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Perform any necessary 'clean-up' prior to exiting the image.
//
void sqlr_cleanup (void)
{
    static bool bCleanedAlready = false;

    // Return if already cleaned
    if (bCleanedAlready) return;

    // Ensure we don't do this twice.
    bCleanedAlready = true;

    // Use guard to deal with exceptions and ensure that we are the only ones
    // doing the cleanup
    CTLockGuard<CTMutex> guard(cleanup_mutex);

    // Catch any exceptions during cleanup. This could happen if
    // a thread was hard killed, leaving some object in an undefined
    // state.
    if (logr) OtaAddTrace(SHUTDOWN, TRACE_FULLDIAG, "Clean up on Exiting ");

    try
    {
        if (TPROCESS)
        {
            if (logr) OtaAddTrace(SHUTDOWN, TRACE_FULLDIAG, "Clear down CORBA Object");

            // We need to do two releases here to ensure that the object's
            // reference count falls to zero and the destructor is called.
            // (not sure why!!)
            CORBA::release(TPROCESS_PTR);
            // CORBA::release(TPROCESS_PTR);
            
            if (logr) OtaAddTrace(SHUTDOWN, TRACE_FULLDIAG, "Dispose of Executer object");
            TPROCESS->_remove_ref();    
            TPROCESS=NULL;
        }

        if (logr) OtaAddTrace (SHUTDOWN, TRACE_FULLDIAG, "Clear down Static Info");

        if (cl)
        {
            delete    cl;
            cl=NULL;
        }
    }

    catch(CyrRepositoryException & e)
    {
        strstream logstr;

        logstr << e.what() << "," << e.where() << " Status: " << e.
        status() << ends;
        if (logr)
            OtaAddTrace(SHUTDOWN, ERRORLEVEL, logstr.str());
        logstr.freeze(false);
        exit(0);
    }

    catch(CORBA::Exception&)
    {
        if (logr)
            OtaAddTrace(SHUTDOWN, ERRORLEVEL,
                        "Caught CORBA::Exception in clean-up ");
        exit(0);
    }

    catch(exception & e)
    {
        string    msg =
            string("Caught exception in clean-up :") + string(e.what());

        if (logr)
            OtaAddTrace(SHUTDOWN, ERRORLEVEL, msg);
        exit(0);
    }

    catch(...)
    {
        if (logr)
            OtaAddTrace(SHUTDOWN, ERRORLEVEL,
                        "Caught exception in clean-up ");
        exit(0);
    }

    if (logr)
        OtaAddTrace(SHUTDOWN, ERRORLEVEL, "Clear out Logging");

    if (logr)
    {
        delete    logr;
        logr = NULL;
    }

} // End of sqlr_cleanup

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Add a message to the Monitor Trace Log.
//
// @rdesc   Return status.
//
// @parm    Pointer to CORBA Server object (input).
// @parm    Name of image to execute (input).
//
int Run(CORBAServer * pServer, string exename)
{
    OtaAddTrace(DBGSTD, DEBUGLEVEL, "Before TScript");

    // Create odlfile
    string odlfile;

    // Build odl file name
    odlfile = exedir + FN_PATHSEP + ODLFILE;
    string msg = "Using odl file: " + odlfile;

    OtaAddTrace(STARTUP, ERRORLEVEL, msg);

    rule = new TScriptRule(odlfile);

    OtaAddTrace(DBGSTD, DEBUGLEVEL, "After TScript");

    // Initialize the CORBA Initialise - pends Test Monitor
    //
    CECInetInfo info;
    string HostName = info.GetIPAddressAsString();

    // Set the name to be registered,
    char      strProcess[64];
    string    ObjectClass;

    // Test Process registration
    int       ProcessID = getpid();

    ObjectClass = TESTPROCESS::CLASS_NAME;

    sprintf(strProcess, "%s_%8.8lX", HostName.c_str(), ProcessID);  

    // Create the instance of the test Process CORBA object
    TPROCESS =
        new TESTPROCESSIMPL(pServer, string(strProcess), HostName, ProcessID,
                            exename);
    TPROCESS_PTR = TPROCESS->_this();

    // If we are running in debug mode then register as such.
    if (cl->option_supplied("-d"))
    {
        string    debug_reg =

            "DEBUG_" + HostName + "_" + strip_extension(exename);
        string    trcmsg = "Debug requested, register as " + debug_reg;

        OtaAddTrace(STARTUP, DEBUGLEVEL, trcmsg);
        pServer->NameObject(TPROCESS->_this(), ObjectClass.c_str(),  
                                debug_reg.c_str());
    }

    // The Test Monitors have no threads, therefore we simply wait until the
    // Test Manager closes us down.
    blocking_semaphore.lock();
    OtaAddTrace(SHUTDOWN, DEBUGLEVEL, "Shutdown invoked");

    // Cleanup and return.
    sqlr_cleanup();
    delete rule;
    delete pServer;

    return 0;
}
